      module kernel
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!****************************************************************
!  This file contains the VINE module to calculate the          *
!  kernel quantities for the SPH kernel in standard use         *
!                                                               *
!  See Monaghan and Lattanzio 1985 A&A 149 135-143              *
!                                                               *
!  Note: The kernel derivative here is divided by v.            *
!  The reason for this is that it is always multiplied          *
!  by direction vectors dx/r, dy/r, dz/r in SPH calculations.   *
!  In order to avoid lots of mess with divide by r's in the     *
!  sph code itself, we do it here with the v term instead.      *
!                                                               *
!  This kernel can be easily swapped for another one            *
!  by writing your own such that the subroutine names           *
!  are the same and their return values are the same.           *
!  Care must be taken when doing this however, because the      *
!  tree neighbor search routines currently assume that          *
!  neighbors are defined as particles within 2h of each         *
!  other. If your kernel is different, you will need to         *
!  define an 'hfac' that multiplies the h value calculated      *
!  by the neighbor search routines by the proper factor.        *
!  For example, if the kernel requires neighbors within         *
!  2.5h, you would need to define a factor hfac=25/16           *
!  (i.e. [5/4]**2) and modify the function hsum2 in treefuncs   *
!  to include this factor. You must also change the lines       *
!  in the tree.F routines getnaybs and its cousins that         *
!  store the h values in the distmas array to reflect this      *
!  change as well.                                              *
!                                                               *
!  This module contains a tabulated version of the kernel       *
!  and a calculated-as-needed version. The speed of each        *
!  will depend on your hardware, but do not appear to be        *
!  much different on the hardware I've look at it for           *
!  specifically. Points to watch out for regarding speed are    *
!  that the tabulated version will blow out your cache much     *
!  more extensively (and thereby slow down your run). The       *
!  calculated version might take a few more clock cycles per    *
!  iteration. That depends mostly on the speed of the real to   *
!  integer conversion, compared to the speed of a floating      *
!  point division and square root, since those will typically   *
!  be the most expensive parts of the kernel calculation, in    *
!  one case or the other.                                       *
!                                                               *
!  Only one version can be used at a time, and it is called     *
!  (in sph.F) using the name SPHkern. You must rename the       *
!  active version to this name in the use statement of the      *
!  sph module.                                                  *
!                                                               *
!  Some important variables used in these routines:             *
!    ntot             : Number of particles you want kernel     *
!                       quantities for                          *
!    nneigh(j)        : Number of neighbors of particle j       *
!    distmas(ir2,k,j) : squared distance between k and j        *
!    distmas(ih2,k,j) : squared smoothing length                *
!    WdW(1,k,j)       : SPH kernel value                        *
!    WdW(2,k,j)       : SPH kernel derivative value             *
!                                                               *
!  Written by Andy Nelson                                       *
!****************************************************************
#ifndef SPH
      W4kern module: SPH preprocessor flag not set
#endif

      use dimens
      use constants

      implicit none

      private
      public W0,v2max,SPHkern_ana,SPHkern_tab,SPHkerntablinit 

      save
!
! normalization:
!
#if NDIM==2   
      double precision, parameter::W0=10d0/(7d0*PI)
#else
      double precision, parameter::W0=one/PI
#endif
!
! Extent of kernel:
!
      double precision, parameter:: v2max=4d0
!
! Tabulated version of kernel, if desired:
!
      integer, parameter:: maxtabl=10000,maxtabl1=maxtabl+1
      double precision, parameter:: xmaxtabl=maxtabl
      double precision  dv2incr,dv2incr_inv

      double precision WdWtab(2,0:maxtabl1)


      contains
!--------------------------------------------------------------------
      subroutine SPHkern_ana(ichoice,nneigh,ntot,distmas,WdW) 
!
! ichoice=1: return the kernel and its derivative
! ichoice=2: return the kernel derivative only
!
      integer, intent(in):: ichoice
      integer, intent(in):: ntot,nneigh(ntot)
      double precision, intent(in ):: distmas(ndim3,maxbite,ntot)
      double precision, intent(out):: WdW(2,maxbite,ntot)

      integer j,k
      double precision hmeani,hmean2i,v,v2,v3,dif2,sum

      select case(ichoice)
      
      case(1)
         do j=1,ntot
            do k=1,nneigh(j)
               hmean2i=one/distmas(ih2,k,j)
               v2=min(v2max,distmas(ir2,k,j)*hmean2i)
               v=sqrt(v2)
               v3=v2*v
               if(v2.lt.one)then
                  sum=one - thrhalf*v2 + thrfourth*v3
                  WdW(1,k,j)=W0*sum
                  sum=-three + 2.25d0*v
                  WdW(2,k,j)=W0*sum
               else
                  dif2=max(zero,two - v)
                  sum=fourth*dif2*dif2*dif2
                  WdW(1,k,j)=W0*sum
                  sum=-thrfourth*v + three - three/v
                  WdW(2,k,j)=W0*sum
               endif
               WdW(1,k,j)=WdW(1,k,j)*hmean2i
               WdW(2,k,j)=WdW(2,k,j)*hmean2i*hmean2i
#if NDIM==3
               hmeani=sqrt(hmean2i)
               WdW(1,k,j)=WdW(1,k,j)*hmeani
               WdW(2,k,j)=WdW(2,k,j)*hmeani
#endif
            enddo
         enddo

      case(2)
         do j=1,ntot
            do k=1,nneigh(j)
               hmean2i=one/distmas(ih2,k,j)
               v2=min(v2max,distmas(ir2,k,j)*hmean2i)
               v=sqrt(v2)
               if(v2.lt.one)then
                  sum=-three + 2.25d0*v
                  WdW(2,k,j)=W0*sum
               else
                  sum=-thrfourth*v + three - three/v
                  WdW(2,k,j)=W0*sum
               endif
               WdW(2,k,j)=WdW(2,k,j)*hmean2i*hmean2i
#if NDIM==3
               hmeani=sqrt(hmean2i)
               WdW(2,k,j)=WdW(2,k,j)*hmeani
#endif
            enddo
         enddo

      case default 
         stop'invalid kernel option'
      end select

      return
      end subroutine SPHkern_ana
!--------------------------------------------------------------------
      subroutine SPHkerntablinit
!
! The table increments linearly in the square of v.
! Therefore interpolations are in the quantity v2, 
! and that means we save one sqrt in the interpolation.
!
      integer k
      double precision v,v2,v3,dif2,sum

      dv2incr=v2max/dble(maxtabl)
      dv2incr_inv=one/dv2incr

      do k=0,maxtabl1
         v2=dble(k)*dv2incr
         v=sqrt(v2)
         v3=v2*v
         if(v2.lt.one)then
            sum=one - thrhalf*v2 + thrfourth*v3
            WdWtab(1,k)=W0*sum
            sum=-three + 2.25d0*v
            WdWtab(2,k)=W0*sum
         else
            dif2=max(zero,two - v)
            sum=fourth*dif2*dif2*dif2
            WdWtab(1,k)=W0*sum
            sum=-thrfourth*v + three - three/v
            WdWtab(2,k)=W0*sum
         endif
      enddo

      return
      end subroutine SPHkerntablinit
!--------------------------------------------------------------------
      subroutine SPHkern_tab(ichoice,nneigh,ntot,distmas,WdW) 
!
! ichoice=1: return the kernel and its derivative
! ichoice=2: return the kernel derivative only
!
      integer, intent(in):: ichoice
      integer, intent(in):: ntot,nneigh(ntot)
      double precision, intent(in ):: distmas(ndim3,maxbite,ntot)
      double precision, intent(out):: WdW(2,maxbite,ntot)

      integer indx,indx1
      double precision dvsq,dwdvsq,d2wd2vsq

      integer j,k
      double precision hmeani,hmean2i,v2

      select case(ichoice)
      case(1)
         do j=1,ntot

            do k=1,nneigh(j)

               hmean2i=one/distmas(ih2,k,j)
               v2=min(v2max,distmas(ir2,k,j)*hmean2i)

               indx=min(v2*dv2incr_inv,xmaxtabl)
               indx1=indx+1
               dvsq=v2 - dble(indx)*dv2incr

               dwdvsq  =(WdWtab(1,indx1)-WdWtab(1,indx))*dv2incr_inv
               d2wd2vsq=(WdWtab(2,indx1)-WdWtab(2,indx))*dv2incr_inv

               WdW(1,k,j)=WdWtab(1,indx) + dwdvsq*dvsq
               WdW(2,k,j)=WdWtab(2,indx) + d2wd2vsq*dvsq

               WdW(1,k,j)=WdW(1,k,j)*hmean2i
               WdW(2,k,j)=WdW(2,k,j)*hmean2i*hmean2i
#if NDIM==3
               hmeani=sqrt(hmean2i)
               WdW(1,k,j)=WdW(1,k,j)*hmeani
               WdW(2,k,j)=WdW(2,k,j)*hmeani
#endif
            enddo
         enddo

      case(2)
         do j=1,ntot
            do k=1,nneigh(j)

               hmean2i=one/distmas(ih2,k,j)
               v2=min(v2max,distmas(ir2,k,j)*hmean2i)

               indx=min(v2*dv2incr_inv,xmaxtabl)
               indx1=indx+1
               dvsq=v2 - dble(indx)*dv2incr

               d2wd2vsq=(WdWtab(2,indx1)-WdWtab(2,indx))*dv2incr_inv

               WdW(2,k,j)=WdWtab(2,indx) + d2wd2vsq*dvsq

               WdW(2,k,j)=WdW(2,k,j)*hmean2i*hmean2i
#if NDIM==3
               hmeani=sqrt(hmean2i)
               WdW(2,k,j)=WdW(2,k,j)*hmeani
#endif
            enddo
         enddo

      case default
         stop'invalid kernel option'
      end select

      return
      end subroutine SPHkern_tab
!--------------------------------------------------------------------
      end module kernel
