      module derivcalc 
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!*********************************************************************
!  This file contains the VINE module that is the main driver for    *
!  calling various derivative calculation routines.                  *
!                                                                    *
!  Written by Andy Nelson, Markus Wetzstein and Thorsten Naab        *
!*********************************************************************
      use dimens
      use stubs
      use miscvars
      use clmpvars
      use perform
      use constants

      use tree, only:gravity,activate_tree,deactivate_tree

      implicit none

      private
      public deriv,renew_deriv

      contains
!-----------------------------------------------------------------
      subroutine deriv(nactive,nactive_s,nactive_n,nactive_c,nactivepm,
     &                 iactive,iactive_s,iactive_n,iactive_c,iactivepm,
     &                 itstepbin,itstepbinpm,dtincr,
     &                 npart,npart_sph,npoim,posmas,vel,u,h,rho,divv,
     &                 alpha,material,acc,poten,dudt,dhdt,alphadot,
     &                 posmaspm,hpm,accpm,potpm)
!******************************************************************
!                                                                 *
!  The driver for all right-hand-side terms                       *
!  for all differential equations.                                *
!                                                                 *
!  NOTE: If you add additional routines here, make absolutely     *
!  certain that the activate_tree/deactivate_tree pair remain     *
!  as the first and last calls. Any attempt to get data from      *
!  the tree otherwise, will fail. (e.g. getneigh in density       *
!  and sphaccel)                                                  *
!                                                                 *
!******************************************************************
      use treemake, only: rtree,rebuildcheck,resethnode
#if PM
      use pmcalcs, only: pmgrav
#endif
#if SPH
      use sph
      use energy
      use eostate
#endif
      integer, intent(in):: npart,npart_sph,npoim
      integer, intent(in):: nactive_s, iactive_s(npart_sph)
      integer, intent(in):: nactive_n, iactive_n(max(1,npart-npart_sph))
      integer, intent(in):: nactive  , iactive  (npart)
      integer, intent(in):: nactivepm, iactivepm(npoim)
      integer, intent(inout):: nactive_c, iactive_c(npart)

      integer, intent(in):: itstepbin(npart),itstepbinpm(npoim)
      double precision, intent(in   ):: dtincr

      double precision, intent(in   ):: posmas  (ndim1,npart)
      double precision, intent(in   ):: vel     (ndim ,npart)
      double precision, intent(in   ):: h             (npart)
      double precision, intent(in   ):: posmaspm(ndim1,npoim)
      double precision, intent(in   ):: hpm           (npoim)

      integer, intent(in):: material(npart_sph)
      double precision, intent(in   ):: alpha(npart_sph)
      double precision, intent(inout):: u(npart_sph) !inout for iso_eos

      double precision, intent(inout):: dhdt(npart_sph)
      double precision, intent(inout):: dudt(npart_sph)
      double precision, intent(inout):: alphadot(npart_sph)
      double precision, intent(inout):: rho(npart_sph),divv(npart_sph)

      double precision, intent(inout):: acc  (ndim,npart),poten(npart)
      double precision, intent(inout):: accpm(ndim,npoim),potpm(npoim)

      logical, save :: first =.true.
      integer i,j,imac0
      double precision ta

      double precision tme1,tme2
      character(len=20), parameter:: where =('deriv')
      if(idbg.ge.2)print *,where

!$omp parallel do private(j,i) 
!$omp&  shared(npart_sph,nactive,iactive)
!$omp&  shared(dudt,alphadot,dhdt,poten,acc)
      do j=1,nactive
         i=iactive(j)
         if(i.le.npart_sph)then
            dhdt    (i)=zero
            dudt    (i)=zero
            alphadot(i)=zero
         endif
         poten(  i)=zero
         acc  (1,i)=zero
         acc  (2,i)=zero
#if NDIM==3
         acc  (3,i)=zero
#endif
      enddo
      do j=1,nactivepm
         i=iactivepm(j)
         potpm(  i)=zero
         accpm(1,i)=zero
         accpm(2,i)=zero
#if NDIM==3
         accpm(3,i)=zero
#endif
      enddo
!
! Turn on active particles and clumps in the tree (always do this first)
!
      call activate_tree(igrape,numclumps,lstclumps,
     &                   nactive,iactive,nactive_c,iactive_c)
!
!====================================================
! Start Gravity related calculations and update tree.
!====================================================
!
! accelerations due to and on point masses
!
      if(npoim.ne.0)then
         call pmgrav(nactive,iactive,npart,nactivepm,iactivepm,npoim,
     &               posmas,acc,poten,posmaspm,hpm,accpm,potpm)
      endif

      if(nactive.gt.0)then   !i.e. do if not only PMs are active 
!
! update the tree information, and set to rebuild next time if needed
!
         imac0=imac ; ta=treeacc
         if(first .and. imac.eq.id_gadget)then
            imac0=-id_angle
            ta=HALF
            first=.false.
         endif

         call rtree(igrav,imac0,isoft,igrape,ifullextrap,irevise,
     &              npart,nactive,iactive,nactive_s,iactive_s,
     &              eps,ta,boxlims,posmas,h,numclumps,lstclumps,
     &              hclmp_is_exact)
         call rebuildcheck(hclmp_is_exact,clhfrac,numclumps,lstclumps,
     &                     rebuild)
!
! compute gravitational accelerations
!
         call gravity(igrav,ibctype,idirect,igrape,igrapelim,imac0,
     &                isoft,eps,gmasslim,nactive_c,iactive_c,
     &                nactive,iactive,npart,posmas,vel,acc,poten)

      endif

      if(nactive_s.eq.0)then
         call deactivate_tree(igrape,nactive,iactive,
     &                        nactive_c,iactive_c)
         return
      endif
!
! For some softening options (eg. fixed spline or Plummer), we need 
! to reset the smoothing lengths in the tree, so that the SPH
! particles get correct smoothing. Also, if there are nbody particles,
! in principle we should want to set their sizes to zero so that 
! the sph calculations will go faster by not having such big node
! sizes. In practice this is a negative because the reset costs more
! than the gain in sph speed. Your mileage may vary as they say and 
! if you want to try it, uncomment the alternate if here:
!
!      if(isoft.ne.id_varkern .or. npart.ne.npart_sph)then
      if(isoft.ne.id_varkern)then
         call resethnode(ifullextrap,npart,h,numclumps,lstclumps)
      endif

!=============================
! Start SPH only calculations.
!=============================

      if(itme.ge.1)call cpu_time(tme1)
!
! find the densities
!
      call density(nactive_c,iactive_c,
     &             npart,npart_sph,posmas,vel,h,rho,divv)
!
! compute the smoothing length derivative
!
      call hdot(nactive_s,iactive_s,nactive_n,iactive_n,
     &          itstepbin,npart,npart_sph,dtincr,h,divv,dhdt)
!
! compute the pressure and sound speed
!
      call eos(npart,npart_sph,npoim,rho,material,posmas,posmaspm,u)
!
! compute the viscosity coefficient derivative
!
      call viscdot(nactive_s,iactive_s,npart,npart_sph,h,
     &             alpha,divv,alphadot)
!
! compute the hydrodynamic accelerations
!
      call sphaccel(nactive_c,iactive_c,npart,npart_sph,vel,
     &              rho,alpha,acc)
!
! compute the energy derivative
!
      call udot(nactive_s,iactive_s,npart_sph,npoim,posmas,
     &          rho,u,posmaspm,dudt)

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_sph)=timing(itme_sph)+(tme2-tme1)
         ncalls(itme_sph)=ncalls(itme_sph)+1
         if(itme.ge.2)print *,perfnames(itme_sph),xcpus_inv*(tme2-tme1)
      endif
!
! Turn off active particles and clumps in the tree (always do this last)
!
      call deactivate_tree(igrape,nactive,iactive,nactive_c,iactive_c)

      return
      end subroutine deriv
!-----------------------------------------------------------------
      subroutine renew_deriv(do_it,required)
!***********************************************************************
!   Renew the derivative calculation, for any one of a                 *
!   variety of reasons. For example:                                   *
!     1) Bookkeeping to keep track of energy conservation              *
!        more exactly by accounting for 1/2 step offsets in            *
!        calculation of the gravitational potential vs end             *
!        of timestep in the leapfrog integrator).                      *
!        (NOT Required)                                                *
!     2) Recalculating the forces after a particle is                  *
!        removed from a simulation. (REQUIRED)                         *
!                                                                      *
!   There is a side effect of running this subroutine, because the     *
!   derivatives get updated in a place they wouldn't normally be.      *
!   One could argue that the update is beneficial in that they are     *
!   slightly better estimates for the integration going forward than   *
!   otherwise. On the other hand, one could argue that centered        *
!   derivatives are needed for order accuracy. Caveat emptor. This     *
!   may also make exact reproducibility a problem if you change output *
!   frequency.                                                         *
!                                                                      *
!***********************************************************************
      use partvars
      use integvars
      use indivtsvars
#if PERIODIC
      use ewald
#endif
      logical, intent(in):: do_it,required
      integer i


      if(.not.required .and. .not.do_it)then
         print *,'   WARNING WARNING WARNING WARNING WARNING'
         print *,'Diagnostic energy calculation has not syncronized' 
         print *,'the particle times for their derivatives. This is'
         print *,'beneficial for the calculation but ugly for plots of'
         print *,'energy, which show apparent violation of the'
         print *,'conservation of energy because the gravitational'
         print *,'potential for each particle comes from different'
         print *,'times.  Set do_it=.true. in the appropriate'
         print *,'renew_deriv caller to synchronize the derivative'
         print *,'times and get the gravitational potential for all'
         print *,'particles at the same time.'

         return
      endif

      if(do_it .and. indts.eq.1)then
         print *,'   WARNING WARNING WARNING WARNING WARNING'
         print *,'Synchronizing the derivative calculations of'
         print *,'all particles to the same time. This may upset'
         print *,'your integrator''s requirements for stability'
         print *,'and order accuracy'
      endif



!$omp parallel do default(none) private(i) 
!$omp& firstprivate(npart_sph) shared(npart,iactive)
!$omp&  shared(iactive_s,iactive_n)
      do i=1,npart
         iactive(i)=i
         if(i.le.npart_sph)then
            iactive_s(i)=i
         else
            iactive_n(i-npart_sph)=i
         endif
      enddo
      do i=1,npoim
         iactivepm(i)=i
      enddo

      nactivepm=npoim
      nactive  =npart
      nactive_s=npart_sph
      nactive_n=npart-npart_sph

      call activate_tree(igrape,numclumps,lstclumps,
     &                   nactive,iactive,nactive_c,iactive_c)
      if(ibctype.eq.id_perdbnd)then
         call checkboxlimits(npart,npoim,posmas,posmaspm)
      endif
      call deriv(nactive,nactive_s,nactive_n,nactive_c,nactivepm,
     &           iactive,iactive_s,iactive_n,iactive_c,iactivepm,
     &           itstepbin,itstepbinpm,dtincr,
     &           npart,npart_sph,npoim,posmas,vel,u,h,
     &           rho,divv,alpha,material,
     &           acc1,poten,dudt1,dhdt1,dalfdt1,
     &           posmaspm,hpm,acc1pm,potpm)
      call deactivate_tree(igrape,nactive,iactive,nactive_c,iactive_c)

      return
      end subroutine renew_deriv
!-----------------------------------------------------------------
      end module derivcalc
