      program vine
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!****************************************************************
!************************************************************
!************************************************************
!******************                    **********************
!******************    This code is    **********************
!******************    called VINE     **********************
!******************  which stands for  **********************
!******************                    **********************
!****************** V ery              **********************
!****************** I nteresting       **********************
!****************** N ame for an       **********************
!****************** E volution code    **********************
!******************                    **********************
!************************************************************
!************************************************************
!************************************************************
!                                                           *
!  This file contains the VINE module to drive the          *
!  integration of particles forward in time, using either   *
!  a one or two step integrator.                            *
!                                                           *
!************************************************************
!************************************************************
! CODE PEDIGREE                                             *
!************************************************************
!************************************************************
!                                                           *
! Originally based on the SPH code from                     *
!                                                           *
!  Willy Benz     Los Alamos National Laboratory 08/31/86   *
!                 Harvard College Observatory    03/10/87   *
!                                                           *
! though not much is left from that version, other than     *
! some basic outline.                                       *
!                                                           *
! Rewrites for speed of the tree and sph parts:             *
!                                                           *
! Andy Nelson    U of AZ, MPIA Heidelberg, U of Edinburgh   *
!                                                           *
! Further tree/sph rewrites, individual timesteps,          *
!  nbody, leapfrog, cosmology, grape, periodic bc's,        *
!  general code cleanup and integration to make them        *
!  all play nicely together written in 2000-2005 by         *
!                                                           *
! Andy Nelson       U of AZ, later MPIA Heidelberg,         *
!                     and later still U of Edinburgh        *
!                     and later still LANL                  *
! Markus Wetzstein  MPIA Heidelberg, later Universitaets-   *
!                                   Sternwarte Munich       *
! Thorsten Naab     MPIA Heidelberg, later Cambridge        *
!                                   University, England     *
!                                                           *
!************************************************************
      use dimens
      use partvars, only:npart
      use indivtsvars
      use miscvars
      use stubs
      use perform
      use io
      use constants

      implicit none

      integer idump,ii

      double precision tme1,tme2
      character(len=20), parameter:: where =('vine')
!
! Don't allow the number of processors to vary from initial
! setting
!
      call fix_numprocessors
!
! Initialize performance summary information
!
      call initperf
!
! Read all files and data needed to start the code running
!
      call lunit
      call options(idump)
      call getfname(idump,1,1)
      call readfile(idump,1)
!
! set all quantities needed for the run
!
      call preset(idump)
!
! Integrate the system, one dump step at a time
!
      ii=abs(idump)+1
      do idump=ii,100000
         if(itme.ge.1)call cpu_time(tme1)

#if defined(RK)
         call twostep(idump)
#elif defined(LF)
         call onestep(idump)
#endif
         call save(idump)

         if(itme.ge.1)then
            call cpu_time(tme2)
            print *,'Dump required wall clock time',
     &                             xcpus_inv*(tme2-tme1),' s'
            timing(itme_dumptime)=timing(itme_dumptime)+tme2-tme1
            ncalls(itme_dumptime)=ncalls(itme_dumptime)+1
            call perfsummary(npart)

         endif
      enddo

      stop'VINE finished normally'

!-----------------------------------------------------------------
      contains
      subroutine onestep(idump)
!************************************************************
!  This is a template for a one step integrator.            *
!  The current implmentation is specific to the             *
!  Leapfrog method, but should require very minimal         *
!  modification to be used for any other two step           *
!  integration method.                                      *
!                                                           *
!  It is possible to have individual (binned) time steps    *
!  for each particle.                                       *
!                                                           *
!************************************************************
      use partvars
      use clmpvars
      use derivcalc, only:deriv
      use indivts
      use integ
      use errorcrit
      use integvars
#if PERIODIC
      use ewald
#endif
#if COSMO
      use cosmo, only: cosmic_stepa
#endif
      integer, intent(in) :: idump

      integer, parameter:: i12step=1,ihalf=0,ifull=1

      integer ii

      logical, save :: first =.true.
      logical success,donefullstep
      integer nexthalf,notused
      character(len=20), parameter:: where =('onestep')
      if(idbg.ge.2)print *,where

!
! First time, we need to extrapolate some particle quantities 
! to the 1/2 step 
!
      if(first)then
         first=.false.
         call extrap_half(deltat,deltatpm)
      endif
!
! Integrate some subset of particles forward in time
!
! Start of main loop 
!
      do while( (iusebin.eq.0 .and. t.lt.tnext_out .and.        !unbinned ts
     &                     n_dtmax_steps.lt.next_outstep) .or.  !
     &          (iusebin.eq.1  .and.                            !binned ts
     &                     n_dtmax_steps.lt.next_outstep) .or.  !
     &          (iexpand.eq.1 .and. t.lt.tnext_out)       )     !cosmo

         call getnext(npart,npoim,nexthalf,notused)
         call getactive(ihalf,npart,npart_sph,npoim)

         now = nexthalf
         dtnext=half*dtincr*itstepbin(iactive(1))   !half of current timestep
         tnow= t + dtincr*dble(max(1,now))
!
! synchronize everything to the same time (nactivef* not used)
! do derivative calculation, and then integration.
!
         if(iexpand.eq.1)then
            call cosmic_stepa(ascale,dtnext,cosm_h0,omegam,omegal)
         endif
         call halfupdate(nactivef_s,nactivef_n,nactivef,nactivefpm,
     &                   nactive_s ,nactive_n ,nactive ,nactivepm ,
     &                   iactive_s ,iactive_n ,iactive ,iactivepm ,
     &                   itstepbin,itstepbinpm,dtincr,
     &                   deltat,deltatpm)
         if(ibctype.eq.id_perdbnd)then
            call checkboxlimits(npart,npoim,dumposmas,dumposmaspm)
         endif
         t_deriv=tnow
         call deriv(nactive,nactive_s,nactive_n,nactive_c,nactivepm,
     &              iactive,iactive_s,iactive_n,iactive_c,iactivepm,
     &              itstepbin,itstepbinpm,dtincr,
     &              npart,npart_sph,npoim,dumposmas,dumvel,dumu,dumh,
     &              rho,divv,dumalf,material,
     &              acc1,poten,dudt1,dhdt1,dalfdt1,
     &              dumposmaspm,dumhpm,acc1pm,potpm)

         tnow= tnow + dtnext
         call fullupdate(nactive_s,nactive_n,nactive,nactivepm,
     &                   iactive_s,iactive_n,iactive,iactivepm,
     &                   deltat,deltatpm)
         if(iexpand.eq.1)then
            call cosmic_stepa(ascale,dtnext,cosm_h0,omegam,omegal)
         endif
         if(ibctype.eq.id_perdbnd)then
            call checkboxlimits(npart,npoim,dumposmas,dumposmaspm)
         endif
!
! calculate errors, set new timestep and do leapfrog corrections
! on particles that changed their timestep.
!
         call errors(i12step,nactive_s,nactive_n,nactive,nactivepm,
     &                       iactive_s,iactive_n,iactive,iactivepm,
     &                       itstepbin,itstepbinpm,dtincr,
     &                       dtrat,dtratpm)
         call setnewdt(npart,npoim,success)
         call lf_correct  (nactive_s,nactive,nactivepm,
     &                     iactive_s,iactive,iactivepm,
     &                     itstepbin,itstepbinpm,dtincr,
     &                     deltat,deltatpm)
         call update_stepping  (i12step)

         if(success)then     !always true for indts.eq.1

            call copyback(nactive_s,nactive,nactivepm,
     &                    iactive_s,iactive,iactivepm)

            call update_time(npart,npoim,donefullstep)
            call afterstep(idump,ihalf,nactivepm,donefullstep)
         endif

      enddo   !t.lt.tnext_out


      return
      end subroutine onestep
!-----------------------------------------------------------------
      subroutine twostep(idump)
!************************************************************
!  This is a template for a two step integrator.            *
!  The current implmentation is specific to the             *
!  Runge-Kutta-Felhberg 1(2) method, but should             *
!  require very minimal modification to be used             *
!  for any other two step integration method.               *
!                                                           *
!  It is possible to have individual (binned) time steps    *
!  for each particle.                                       *
!                                                           *
!************************************************************
      use partvars
      use clmpvars
      use derivcalc, only:deriv
      use indivts
      use errorcrit
      use integ
      use integvars
#if PERIODIC
      use ewald
#endif
      integer, intent(in) :: idump

      integer, parameter :: i12step=2,ihalf=0,ifull=1

      logical success,donefullstep
      integer nexthalf,nextfull

      logical, save :: first =.true.

      character(len=20), parameter:: where =('twostep')
      if(idbg.ge.2)print *,where

!
! Main loop to integrate some subset of particles 
! either a half or a full step forward in time
!
      do while( (iusebin.eq.0 .and. t.lt.tnext_out .and.        !unbinned ts
     &                     n_dtmax_steps.lt.next_outstep) .or.  !
     &          (iusebin.eq.1  .and.                            !binned ts
     &                     n_dtmax_steps.lt.next_outstep) .or.  !
     &          (iexpand.eq.1 .and. t.lt.tnext_out)       )     !cosmo

         call getnext(npart,npoim,nexthalf,nextfull)
         now = min(nexthalf,nextfull)
         tnow= t + dtincr*dble(max(1,now))

         if(nexthalf.lt.nextfull)then
!
! Half step integration only (didn't find any needing full steps)
!
            call getactive (ihalf,npart,npart_sph,npoim)
            call halfupdate(nactivef_s,nactivef_n,nactivef,nactivefpm,
     &                      nactive_s ,nactive_n ,nactive ,nactivepm ,
     &                      iactive_s ,iactive_n ,iactive ,iactivepm ,
     &                      itstepbin,itstepbinpm,dtincr,
     &                      deltat,deltatpm)
            if(ibctype.eq.id_perdbnd)then
               call checkboxlimits(npart,npoim,dumposmas,dumposmaspm)
            endif
            t_deriv=tnow
            call deriv(nactive,nactive_s,nactive_n,nactive_c,nactivepm,
     &                 iactive,iactive_s,iactive_n,iactive_c,iactivepm,
     &                 itstepbin,itstepbinpm,dtincr,
     &                 npart,npart_sph,npoim,dumposmas,dumvel,dumu,dumh,
     &                 rho,divv,dumalf,material,
     &                 acc2,poten,dudt2,dhdt2,dalfdt2,
     &                 dumposmaspm,dumhpm,acc2pm,potpm)

            call copyrho(npart_sph,nactivef_s,nactive_s,iactive_s,
     &                   rho,rhohalf,divv,divvhalf)
            call update_stepping(i12step)
         else
!
! Half step integration for particles needing half steps and
!   full step integration for particles needing full steps.
!
            call getactive (ifull,npart,npart_sph,npoim)
            call halfupdate(nactivef_s,nactivef_n,nactivef,nactivefpm,
     &                      nactive_s ,nactive_n ,nactive ,nactivepm ,
     &                      iactive_s ,iactive_n ,iactive ,iactivepm ,
     &                      itstepbin,itstepbinpm,dtincr,
     &                      deltat,deltatpm)
            call fullupdate(nactivef_s,nactivef_n,nactivef,nactivefpm,
     &                      iactive_s ,iactive_n ,iactive ,iactivepm ,
     &                      deltat,deltatpm)
            if(ibctype.eq.id_perdbnd)then
               call checkboxlimits(npart,npoim,dumposmas,dumposmaspm)
            endif

            t_deriv=tnow
            call deriv(nactive,nactive_s,nactive_n,nactive_c,nactivepm,
     &                 iactive,iactive_s,iactive_n,iactive_c,iactivepm,
     &                 itstepbin,itstepbinpm,dtincr,
     &                 npart,npart_sph,npoim,dumposmas,dumvel,dumu,dumh,
     &                 rho,divv,dumalf,material,
     &                 acc2,poten,dudt2,dhdt2,dalfdt2,
     &                 dumposmaspm,dumhpm,acc2pm,potpm)
            call copyrho(npart_sph,nactivef_s,nactive_s,iactive_s,
     &                   rho,rhohalf,divv,divvhalf)
            call errors(i12step,
     &                  nactivef_s,nactivef_n,nactivef,nactivefpm,
     &                  iactive_s ,iactive_n ,iactive ,iactivepm ,
     &                  itstepbin, itstepbinpm,dtincr,
     &                  dtrat,dtratpm)
            call setnewdt(npart,npoim,success)
            call update_stepping(i12step)

            if(success)then

               call copyback(nactivef_s,nactivef,nactivefpm,
     &                       iactive_s ,iactive ,iactivepm)

               call update_time(npart,npoim,donefullstep)
               call afterstep(idump,ifull,nactivefpm,donefullstep)
            endif
         endif  !half or full step

      enddo

      return
      end subroutine twostep
!-----------------------------------------------------------------
      subroutine preset(idump)
!************************************************************
!                                                           *
!  this routine makes sure that everything not entered as   *
!  an option and needed for the main integration loops      *
!  is defined before integration begins.                    *
!                                                           *
!************************************************************
      use partvars
      use clmpvars
      use integvars
      use indivts
      use vineutils
      use derivcalc, only:deriv
      use treemake , only: mtree
      use analyze  , only: centmass, insync
#ifdef SPH
      use sph, only: SPHkerntablinit
#endif
      integer, intent(in):: idump

      integer i,j
      integer id

      character(len=20), parameter:: where =('preset')
      if(idbg.ge.2)print *,where

! SPH:
!    initialize the kernel table. Assumes that you have
!    selected the tabulated version of the kernel in the first 
!    place, otherwise the table is never accessed.
!
      if(npart_sph.ne.0)then
         call SPHkerntablinit
      endif
!
! Determine centers of mass and momentum
!
      call centmass
!
! Set initial size for boundary box, if there is a box
!
      call setup_boundaries
!
! Build the tree for the first time, then reorder the particles
! for speed optimizations.
!
      call mtree(isoft,igrape,npart,npart_sph,sepmax,bsepmax,eps,
     &           boxlims,posmas,h,numclumps,lstclumps)
      call orderparticles
      nbuild=0
!
! Initialize the definition of 'now', set all particles to active
!  and set their timesteps
! 
      now = 0
      tnow= t
      call init_active(npart,npart_sph,npoim)
      call initdt(npart,npoim)
      call report(npart,npoim)
!
! Initialize the derivatives for all particles
! 
      call deriv(nactive,nactive_s,nactive_n,nactive_c,nactivepm,
     &           iactive,iactive_s,iactive_n,iactive_c,iactivepm,
     &           itstepbin,itstepbinpm,dtincr,
     &           npart,npart_sph,npoim,posmas,vel,u,h,
     &           rho,divv,alpha,material,
     &           acc1,poten,dudt1,dhdt1,dalfdt1,
     &           posmaspm,hpm,acc1pm,potpm)
      call copyrho(npart_sph,nactivef_s,nactive_s,iactive_s,
     &             rho,rhosave,divv,divvsave)
!
! set restart and analysis dump times and reference times
! from which later dump times are calibrated. 
! Last 3 values in these calls are outputs
!
      call setbase_time(readfromchkpnt,override,dtmax ,dt_out,
     &                  tlast_out,tnext_out,t,idump      ,
     &                  t0_out,idump0_out)
      call setbase_time(.false.       ,.false. ,dt_ana,dt_ana,
     &                  tlast_ana,tnext_ana,t,n_ana_steps,
     &                  t0_ana,idump0_ana)

      call setnextout_time(nout_max,idump0_out,n_dtmax_steps,
     &                     idump      ,t0_out,t,dt_out,
     &                     tlast_out,tnext_out,next_outstep)

      call setnextana_time(nana_max,idump0_ana,n_dtmax_steps,
     &                     n_ana_steps,t0_ana,t,dt_ana,dt_ana,
     &                     tlast_ana,tnext_ana,next_anastep)

!
! Do a initial diagnostic calculation of the total energies
!
      if(idbg.ge.1)call insync(0)

!
! Do any other on the fly analysis routines
!
      if(.not.readfromchkpnt)then

         !call your_routine(   )

      endif

      print *,' '
      print *,'End of Initialization'
      print *,' '
      print *,'==================================================='
      print *,'==================================================='

      if(idbg.ge.2)print *,'done ',where

      return
      end subroutine preset
!-----------------------------------------------------------------
      subroutine setup_boundaries
!**********************************************************************
!                                                                     *
! Set initial size for boundary box.                                  *
!   if free boundaries, set the box size to be huge,                  *
!      so that if they are ever used they will do no harm             *
!      or at least be very obvious in their behavior.                 *
!   if periodic or reflecting boundaries, set the box size            *
!      as follows:                                                    *
!                                                                     *
!       boxlims has the positions of the four or six                  *
!          sides of the box (2 or 3d):                                *
!        boxlims(1,1)  low  x boundary (eg x=-1)                      *
!        boxlims(2,1)  high x boundary (eg x=+1)                      *
!        boxlims(1,2)  low  y boundary (eg y=-1)                      *
!        boxlims(2,2)  high y boundary (eg y=+1)                      *
!        boxlims(1,3)  low  z boundary (eg z=-1)                      *
!        boxlims(2,3)  high z boundary (eg z=+1)                      *
!        boxlen(id) size of the box in dimension id (id=1,2,3)        *
!        halfboxlen(id) half of boxlen                                *
!                                                                     *
!    If you choose, you can vary the boxlim values as functions       *
!    of time in the simulation, and in a way appropriate to your      *
!    problem, by writing your own subroutine. There are probably      *
!    reasons to do so, though it is hard to think of many at this     *
!    writing.                                                         *
!                                                                     *
!    These arrays could also be used with simulation regions          *
!    like pie slices or sections of spheres, rather than rectangular  *
!    regions implemented here, but would obviously require some       *
!    recoding elsewhere.                                              *
!                                                                     *
!**********************************************************************
#if PERIODIC
      use ewald, only:init_rand,init_ewald
#endif
      integer id,idummy

      select case(ibctype)
      case(id_freebnd)

         do id=1,ndim
            boxlims (1,id)=-fourth*huge(1d0)
            boxlims (2,id)= fourth*huge(1d0)
            boxlen    (id)=boxlims(2,id)-boxlims(1,id)
            halfboxlen(id)=boxlen(id)/2d0
         enddo
      case(id_perdbnd)

         boxlims(1,1)=xmin
         boxlims(2,1)=xmax
         boxlims(1,2)=ymin
         boxlims(2,2)=ymax
#if NDIM==3
         boxlims(1,3)=zmin
         boxlims(2,3)=zmax
#endif
         do id=1,ndim
            boxlen    (id)=boxlims(2,id)-boxlims(1,id)
            halfboxlen(id)=boxlen(id)/2d0
         enddo
!
! set random seed for boxshift
!
         idummy =-126
         call init_rand(idummy)

         if(igrav.eq.1)call init_ewald

      case(id_reflbnd)
          stop'reflecting boundaries not installed'
      case default 
          stop'Unknown boundary in setup_boundaries'
      end select

      return
      end subroutine setup_boundaries
!-----------------------------------------------------------------
      subroutine save(idump)
!************************************************************
!                                                           *
!  This routine drives the saving of the results on disk    *
!  and computes relevant quantities for print out           *
!                                                           *
!************************************************************
      use analyze
      use derivcalc, only:renew_deriv
      use vineutils
      use partvars, only: npart,npoim,posmas,posmaspm
#if GRAPE
      use forcelaw, only: unlock_grape
#endif
#if PERIODIC
      use ewald, only:shift_box_to_original,shift_box_to_current
#endif
      integer, intent(in):: idump

      logical do_it,required

      double precision tme1,tme2
      character(len=20), parameter:: where =('save')
      if(idbg.ge.2)print *,where

#if defined(LF)
      if(indts.eq.1 .and. igrav.eq.1)then

          do_it=.false.     !set true if you want synchronized gravpot.
          required=.false.

         call renew_deriv(do_it,required)
      endif
#endif

!===================================================
!Start various postprocessing analyses
!===================================================
      if(itme.ge.1)call cpu_time(tme1)
!
! compute center of mass, velocity of cm
!
      call centmass
!
! compute energies
!
      call toten
!
! compute total angular momentum
!
      call angmom
!
! shift simulation box back to original position
!
      if(ibctype.eq.id_perdbnd .and. ishift.eq.1)then
         call shift_box_to_original(npart,npoim,posmas,posmaspm)
      endif

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_analysis)=timing(itme_analysis)+tme2-tme1
         ncalls(itme_analysis)=ncalls(itme_analysis)+1
      endif
!
! write a restart dump to disk
!
      call writefiles(idump,.false.,1)
!
! shift simulation box again to current position
!
      if(ibctype.eq.id_perdbnd .and. ishift.eq.1)then
         call shift_box_to_current(npart,npoim,posmas,posmaspm)
      endif
!
! check for run termination
! use spacing(tstop) because spacing(t-tstop) when they are equal, i.e.
! t-tstop ca. 0, is typically much smaller than spacing tstop
!
      if(abs(t-tstop).lt.TEN*spacing(tstop) .or.
     $     n_dtmax_steps.gt.nstepmax)then
         call unlock_grape
         stop'Reached tstop or maximum number of time steps.'
      endif
!
! Set the next dump time
!
      call setnextout_time(nout_max,idump0_out,n_dtmax_steps,idump,
     &                     t0_out,t,dt_out,
     &                     tlast_out,tnext_out,next_outstep)

      return
      end subroutine save
!-----------------------------------------------------------------
      subroutine afterstep(idump,ifh,nactivepm,donefullstep)
!************************************************************
!                                                           *
!  After full timestep bookkeeping and on the fly analysis  *
!                                                           *
!  nactivepm is secretly either nactivepm or nactivefpm     *
!  in the caller, depending on whether we are being called  *
!  from the onestep or twostep integrator.                  *
!************************************************************
      use analyze
      use treemake , only: mtree
      use indivts  , only: report
      use derivcalc, only: renew_deriv
      use vineutils
      use partvars
      use clmpvars
#if PM
      use pm, only: remove
#endif
#if PERIODIC
      use ewald
#endif
      logical, intent(in):: donefullstep
      integer, intent(in):: idump,ifh,nactivepm

      logical writepm
      integer i12step,nremove,it,i

      character(len=20), parameter:: where =('afterstep')
      if(idbg.ge.2)print *,where

      i12step=ifh+1

      nremove=0
      if(npoim.gt.0)then
! 
! Accrete particles onto PMs and store PM info for later output
! (must have a PM at the end of its step),
!
         if(nactivepm.gt.0)then
            if(irem.ne.0)call remove(i12step,nremove)
            call dumcentmass
            call pmstore(i12step)
         endif

         writepm=.false.
         do i=1,npoim
            if(mod(nsteppm(i),nana_max).eq.0)writepm=.true.
            if(nstore(i).eq.maxstore)writepm=.true.
         enddo
         if(writepm) call pmoutput

      endif
!
! Rebuild the tree, if needed. Note `now' is out of date
! by one timestep since we just did the deriv at `now' and are
! going on to the future. Shouldn't matter much though.
!
      if(indts.eq.0)nbuild=nbuild+1

      if(rebuild .or. nremove.gt.0 .or.
     &         (nbuild.ge.maxbuild .and. indts.eq.0))then

         if(ifullextrap.eq.1 .and. indts.eq.1)then
            call synch2n(now)
         endif
         if(ibctype.eq.id_perdbnd .and. ishift.eq.1)then
            call boxshift(npart,npoim,posmas,posmaspm)
         endif

         call mtree(isoft,igrape,npart,npart_sph,sepmax,bsepmax,eps,
     &              boxlims,posmas,h,numclumps,lstclumps)
         if(indts.eq.0)nbuild=0

      endif
!
! Now do various kinds of diagnostic and analysis output,
! and reorder particles if needed.
!
      if(indts.eq.0)then
!--------------------
! Global timestep
!--------------------
         nstep=nstep+1
         if(iusebin.eq.0)then
            n_dtmax_steps=n_dtmax_steps+1
         elseif(iusebin.eq.1 .and. donefullstep)then
            n_dtmax_steps=n_dtmax_steps+1
         endif

         call insync(i12step)

         if(mod(nstep,10*maxbuild).eq.0)then
            call orderparticles
            nbuild=0
         endif

      else
!--------------------
! Individual timestep
!--------------------
!
! Only do something if we are at the end of a full, 
! large timestep and are starting a new one. 
!
!  Notes: 
!    1) Must have either mtree or orderparticles here. Commenting 
!       out both results in small, systematic jumps in the total 
!       energy calculation. Reasons unknown. Shouldn't happen.
!
!    2) Reorder/mtree frequency of 10 seems like a good number 
!       for some problems. Probably could be larger, but ordering 
!       is so cheap...why not?
!
!    3) centmass call is useful for RK integrator error calculation
!       in the pathological case when the particles are nearly at 
!       rest initially, and then start to move (see subroutine
!       rk_err in rk12.F and also the header documentation). 
!
         if(donefullstep)then

            nstep=nstep+1
            n_dtmax_steps=n_dtmax_steps+1

            call insync(i12step)

            if(ibctype.eq.id_perdbnd .and. ishift.eq.1)then
               call boxshift(npart,npoim,posmas,posmaspm)
            endif

            if(nbuild.le.10)then
               nbuild=nbuild+1
               call mtree(isoft,igrape,npart,npart_sph,sepmax,bsepmax,
     &                    eps,boxlims,posmas,h,numclumps,lstclumps)
            else
               call orderparticles
               nbuild=0
            endif

            if(i12step.eq.2)call centmass
         endif
      endif  !indts

!***************************************************************
! Here is where you should call routines that do intermittent,
! on-the-fly analysis. Must synch to `t_deriv' rather than 
! lastts_end or something similar because the latter can
! jump forward and back depending on the integrator (eg. the
! dkd leapfrog we use).
!
! Note `now' is the time at which we did the last derivative calc.
!     i.e.  tnow=t_deriv
!
!
      if( (t_deriv.ge.tnext_ana                               ) .or.
     &    (iusebin.eq.0 .and. n_dtmax_steps.ge.next_anastep)  )then

         if(ifullextrap.eq.1 .and. indts.eq.1)then
            call synch2n(now)
         endif

         n_ana_steps=n_ana_steps + 1
!
! Put your analysis routines and outputs here.
!

         !call your_routine(   )

         call setnextana_time(nana_max,idump0_ana,
     &                        n_dtmax_steps,n_ana_steps,
     &                        t0_ana,t_deriv,dt_ana,dt_ana,
     &                        tlast_ana,tnext_ana,next_anastep)

      endif
!
! Make a checkpoint dump
!
      if(makechkpnt.eq.1 .and. donefullstep)then
         call writefiles(idump,.true.,1)
      endif

      return
      end subroutine afterstep
!-----------------------------------------------------------------
      subroutine copyrho(npart_sph,nactivef_s,nactive_s,iactive_s,
     &                   rho,rhosave,divv,divvsave)

      integer, intent(in):: nactive_s,nactivef_s,npart_sph
      integer, intent(in):: iactive_s(npart_sph)
      double precision, intent(in   ):: rho     (npart_sph)
      double precision, intent(inout):: rhosave (npart_sph)
      double precision, intent(in   ):: divv    (npart_sph)
      double precision, intent(inout):: divvsave(npart_sph)
      integer i,j

!$omp parallel do if(nactive_s-(nactivef_s+1).gt.1024) 
!$omp& default(none) private(i,j)
!$omp&    shared(nactivef_s,nactive_s,iactive_s)
!$omp&    shared(rho,rhosave,divv,divvsave)
      do j=nactivef_s+1,nactive_s
         i=iactive_s(j)
         rhosave (i)=rho (i)
         divvsave(i)=divv(i)
      enddo

      return
      end subroutine copyrho
!-----------------------------------------------------------------
      end program vine
