      module energy
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!****************************************************************
!  This file contains the VINE module to do calculate the       *
!  change in internal energy of sph particles, due to a set of  *
!  sources and sink terms.                                      *
!                                                               *
!  This routine should be used as a 'master' routine from       *
!  which you call your own, special purpose energy varying      *
!  routines. In its current form, it includes only pdv energy   *
!  and shock heating. These quantities are passed to the udot   *
!  subrootine through the sphvars module, rather than directly  *
!  in the argument list.                                        *
!                                                               *
!  Written by Andy Nelson                                       *
!****************************************************************
#ifndef SPH
      energy module: SPH proprocessor flag not set
#endif
      use dimens
      use miscvars
      use sphvars
      use constants
      use perform

      implicit none

      private
      public udot

      contains
!--------------------------------------------------------------------
      subroutine udot(nactive_s,iactive_s,npart_sph,npoim,
     &                posmas,rho,u,posmaspm,dudt)
!****************************************************************
!                                                               *
!  compute the change in specific internal energy               *
!  For various EOSs (isothermal, polytropic etc), dudt is       *
!  zero and we set u explicitly in subr. eos.                   *
!                                                               *
!  Include posmas/posmaspm because we can have rad heating      *
!  or cooling dependent on position. For example from stellar   *
!  irradiation.                                                 *
!                                                               *
!****************************************************************

      integer, intent(in):: npart_sph,npoim
      integer, intent(in):: nactive_s,iactive_s(npart_sph)
      double precision, intent(in   ):: posmas  (ndim1,npart_sph)
      double precision, intent(in   ):: posmaspm(ndim1,npoim)
      double precision, intent(in   ):: rho (npart_sph)
      double precision, intent(in   ):: u   (npart_sph)
      double precision, intent(inout):: dudt(npart_sph)

      integer i,j
      double precision pdvon,shokon,cosmo_rho
      double precision tme1,tme2
      character(len=20), parameter:: where =('udot')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)


      if( (ieos.eq.id_lagrgiso .or. ieos.eq.id_euleriso) .and. 
     &         (ipdv.eq.1 .or. ishok.eq.1))then
         stop'udot: Isothermal EOS must set ipdv=ishok=0' 
      endif

      if(ieos.eq.id_lagrgiso .or. ieos.eq.id_euleriso)return

      pdvon=zero
      shokon=zero
      cosmo_rho=one
      if(ipdv.eq.1)pdvon=one
      if(ishok.eq.1)shokon=one
      if(iexpand.eq.1)cosmo_rho=one/ascale**ndim

!$omp parallel do default(none) private(i,j)
!$omp&  firstprivate(pdvon,shokon,cosmo_rho)
!$omp&  shared(dudt,pdv,dq,thermo,rho,u)
!$omp&  shared(nactive_s,iactive_s)
      do j=1,nactive_s
         i=iactive_s(j)
         dudt(i)=dudt(i) + pdvon*pdv(i)*thermo(i_pres,i)/
     &                     (rho(i)*cosmo_rho)**2
     &                   + shokon*half*dq(i)
      enddo

!************************************
! Put external cooling sources here.*
!************************************
      if(icool.ne.0)then   !cooling of some kind selected
!
! radiative cooling (template)
!
!         if(iradcool1.eq.1)call radcool1(u,dudt,ducool)
!
! some other radiative cooling (template) 
!
!        if(iradcool2.eq.1)call radcool2(u,dudt,ducool)
!
! Add your favorite cooling mechanism
! etc etc for other cases.
!

!         call yourcooling

      endif

!************************************
! Put external heating sources here.*
!************************************
      if(iheat.ne.0)then  !heating of some kind selected
!
! radiative heating (template)
!
!        if(iradheat1.eq.1)call radheat1(u,dudt,ducool)
!
! Some other radiative heating (template)
!
!        if(iradheat2.eq.1)call radheat2(u,dudt,ducool)
!
! Add your favorite heating mechanism
! etc etc for other cases.
!
      endif

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_udot)=timing(itme_udot) + tme2-tme1
         ncalls(itme_udot)=ncalls(itme_udot) + 1
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine udot
!--------------------------------------------------------------------
      end module energy
