      module forcelaw
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!****************************************************************
!  This file contains the VINE module to calculate gravity      *
!  on particles, using either a GRAPE board or a hierarchical   *
!  tree or both working together.                               *
!                                                               *
!  Some of the routines in this module could be replaced with   *
!  other routines that calculate other force laws. For example, *
!  if someone wanted to do molecular dynamics with vine, they   *
!  might wish to put in a force law for interparticle forces    *
!  and wave functions relevant for that case. This would        *
!  probably also mean that you would need to replace some of    *
!  the node opening criteria in the tree traversals, though     *
!  the traversals themselves would not require modification.    *
!                                                               *
!  The externally visible routines are:                         *
!                                                               *
!    calcgravgroup: Calculate the forces on a small set of      *
!                 particles due to a list of nodes and atoms    *
!                 culled from a tree traversal. Identical       *
!                 lists are used for all particles needing      *
!                 forces                                        *
!     Arguments:                                                *
!       igrape   : Use grape or not                             *
!       igrapelim: if grape, then limit calls to gt this many   *
!                  particles, otherwise host.                   *
!       isoft    : Type of softening                            *
!       eps2     : Softening coeff for Plummer softening        *
!       npart    : Total number of particles in simulation      *
!       iclump   : Index of clump in treept array. Needed for   *
!                  grape3 only                                  *
!       nact     : Number of active particles in this clump     *
!       ntot     : Number of total  particles in this clump     *
!       listpt   : list of active particles in this clump       *
!       na       : Number of atoms on interaction list          *
!       nn       : Number of nodes on interaction list          *
!       latoms   : List   of atoms on interaction list          *
!       lnodes   : List   of nodes on interaction list          *
!       vel      : Velocity of particles. Needed for grape6 only*
!       acc      : Accelerations of particles. Returned by      *
!                  this routine                                 *
!       poten    : Interparticle potential of particle. Returned*
!                  by this routine (grav potential currently)   *
!                                                               *
!    calcgrav: Calculate the force on a single particle due to  *
!              a list of nodes and atoms culled from a tree     *
!              traversal.                                       *
!     Arguments:                                                *
!       itreept  : index of particle in treept list on which    *
!                  forces are to be calculated                  *
!       isoft    : Type of softening                            *
!       eps2     : Softening coeff for Plummer softening        *
!       npart    : Total number of particles in simulation      *
!       na0      : Number of atoms on interaction list,         *
!                  known not to be neighbors                    *
!       na1      : First atom on list not known whether its a   *
!                  neighbor or not                              *
!       na       : Total number of atoms on the list            *
!       nn       : Number of nodes on interaction list          *
!       latoms   : List   of atoms on interaction list          *
!       lnodes   : List   of nodes on interaction list          *
!       acc      : Accelerations of particles. Returned by      *
!                  this routine                                 *
!       poten    : Interparticle potential of particle. Returned*
!                  by this routine (grav potential currently)   *
!                                                               *
!    acceln2: Calculate the forces on a small set of particles  *
!             due to each other, via direct summation.          *
!     Arguments:                                                *
!       npart : Total number of particles in simulation         *
!       ntot  : Number of particles in the list of particles    *
!               for which forces will be calculated.            *
!       listpt: The list of particles.                          *
!       isoft : Selector for type of softening                  *
!       eps2  : Squared softening length, for Plummer softening *
!       acc   : array with particle accelerations in it.        *
!       poten : array with particle potentials in it.           *
!                                                               *
!    dirsum: Calculate the forces on a large set of particles   *
!             due to each other, via direct summation.          *
!     Arguments:                                                *
!       igrape   : Use grape or not                             *
!       isoft    : Type of softening                            *
!       eps2     : Softening coeff for Plummer softening        *
!       npart    : Total number of particles in simulation      *
!       posmas   : Position and mass of particles, in driver    *
!                  ordering. Needed for some grape options      *
!       vel      : Velocity of particles, in driver ordering    *
!                  Needed for some grape options                *
!       acc   : array with particle accelerations in it.        *
!       poten : array with particle potentials in it.           *
!                                                               *
!    revpole: Update the multipole moments and node data in     *
!             the tree.                                         *
!     Arguments:                                                *
!       irevise: Full or partial node data revision. Currently  *
!                not functional                                 *
!       imac   : Which MAC you have selected to use for         *
!                testing nodes                                  *
!       treeacc: Setting for the MAC you selected.              *
!                                                               *
!    initpole: Initialize the multipole moments to zero         *
!     Arguments:                                                *
!       npart    : Total number of particles in simulation      *
!                                                               *
!    apply_ewald: See ewald module                              *
!                                                               *
!    lock_grape: see grape module                               *
!                                                               *
!    unlock_grape: see grape module                             *
!                                                               *
!                                                               *
!  It would be nice to merge acceln2 and dirsum someday.        *
!                                                               *
!  All other routines and variables should be private to this   *
!  module.                                                      *
!                                                               *
!  Written by Andy Nelson                                       *
!                                                               *
!****************************************************************
      use dimens
      use treevars
      use stubs
      use constants
      use perform

#ifdef HAVE_DIRSUM
      use forcedirect
#endif
#ifdef GRAPE
      use grape, only:lock_grape,unlock_grape
      use grape, only:grape_dirsum,grape_tree
#endif         
#ifdef PERIODIC
      use ewald, only:apply_ewald
#endif

      implicit none

      private
      public dirsum,acceln2,calcgravgroup,calcgrav
      public initpole,revpole
      public lock_grape,unlock_grape
      public apply_ewald

      include 'INCLUDE/hwinfo'

!************************************************************************
!  MULTIPOLE MOMENT ARRAY INDICES AND ARRAYS                            *
!************************************************************************
!
! multipole moment array locations. Someday it would be good
! to convert completely to reduced multipole moments to save
! array space. 5 vs 6 for quadrupole, and 7 vs 10 for octupole
!
      integer, parameter:: momquad=ndim*3-3   !3 or 6  in 2d or 3d
!     integer, parameter:: momoct =ndim*6-8   !4 or 10 in 2d or 3d
      integer, parameter:: moments=momquad  !+momoct
      integer, parameter:: iqxx=1,iqyy=2,iqxy=3
      integer, parameter:: iqzz=4,iqxz=5,iqyz=6    !quadrupole
      integer, parameter:: iqyx=iqxy,iqzx=iqxz,iqzy=iqyz
!      parameter (isxx=momquad+1,isyy=momquad+2,isxy=momquad+3) !octupole
!      parameter (isyx=momquad+4,iszz=momquad+4,isxz=momquad+5)
!      parameter (iszx=momquad+7,isyz=momquad+8,iszy=momquad+9)
!      parameter (isxyz=momquad+10)

!  pole: multipole moments of current node

      double precision, save:: pole(moments,mmax)


      contains
!----------------------------------------------------------------------
      subroutine dirsum(igrape,npart,isoft,eps2,posmas,vel,acc,poten)
!*********************************************************************
!  Calculate gravity on a list of particles, due to all the other    *
!   particles. Use either the host processor or, if present and      *
!   selected, a grape coprocessor.                                   *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in):: igrape,isoft,npart
      double precision, intent(in)::eps2
      double precision, intent(in):: posmas(ndim1,npart),vel(ndim,npart)
      double precision, intent(inout):: acc(ndim,npart), poten(npart)

      if(igrape.eq.1)then
         call grape_dirsum(npart,isoft,eps2,posmas,vel,acc,poten)
      else
         call host_dirsum (npart,isoft,eps2,posmas,vel,acc,poten)
      endif

      end subroutine dirsum
!----------------------------------------------------------------------
      subroutine calcgravgroup(igrape,igrapelim,isoft,eps2,npart,iclump,
     &                         nact,ntot,listpt,na,nn,latoms,lnodes,
     &                         vel,apcorr,acc,poten)
!*********************************************************************
!  Calculate gravity on a list of particles, due to a list of        *
!   distant atoms and nodes, known not to contain any neighbors.     *
!                                                                    *
!  Including the gravatomgrp call here is only really effective      *
!  if there are very many atoms. Usually this will not be the case.  *
!  For example, with the angle MAC, you would need theta<0.1 to      *
!  see much of any difference at all. Still, it doesn't hurt.        *
!                                                                    *
!  Do the calculation on grape coprocessors if they are available    *
!  and are selected.                                                 *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************

      integer, intent(in):: igrape,igrapelim,isoft
      integer, intent(in):: npart,iclump
      integer, intent(in):: nact,ntot,listpt(ntot)
      integer, intent(inout):: na,latoms(idim), nn,lnodes(nn)
      double precision, intent(in):: eps2
      double precision, intent(in):: vel(ndim,npart)
      double precision, intent(inout):: acc(ndim,npart), poten(npart)
      double precision, intent(inout):: apcorr(0:ndim,nact)

      integer nactuse,listact(max(maxbunchsize,maxclumpsize))

      integer j,n,itreept,nonay,na0,na_orig
      double precision e2
      double precision tme1,tme2

!
! Get a list of active particles 
!
      nactuse=0
      do n=1,ntot
         if(active(listpt(n)).ne.1)cycle

         nactuse=nactuse+1
         listact(nactuse)=listpt(n)
      enddo

      if(igrape.eq.0)then
!*********************************************************************
! Use list of nodes and atoms to calculate gravity on host processors*
!*********************************************************************

         e2=zero
         if(isoft.eq.id_Plummer)e2=eps2
         if(na.gt.0)then
            call gravatomgrp(isoft,npart,e2,nactuse,listact,na,latoms,
     &                       apcorr,acc,poten)
         endif
         if(nn.gt.0)then
            call gravnodegrp(isoft,npart,e2,nactuse,listact,nn,lnodes,
     &                       apcorr,acc,poten)
         endif

      else
!*********************************************************************
! Use GRAPE hardware to calculate gravity but only                   *
!    if there are enough active particles to make it worthwhile      *
!                                                                    *
! When their are not enough active particles do calculation on the   *
! host, using gravatomgrp for external atoms and acceln2 for         *
! internal-to-bunch atoms so as to avoid self potential              *
! calculation                                                        *
!                                                                    *
!*********************************************************************

         na_orig=na
         do n=1,nn
            na=na+1
            latoms(na)=lnodes(n)
         enddo
         na0=na
         do j=1,ntot
            itreept=listpt(j)
            na=na+1
            latoms(na)=itreept
         enddo

         if(nact.ge.igrapelim)then

            call grape_tree(iclump,isoft,eps2,npart,nactuse,listact,
     &                      na0,na,latoms,vel,apcorr,acc,poten)

         else

            if(itme.ge.1)call cpu_time(tme1)

            nonay=id_Plummer
            call gravatomgrp(nonay,npart,eps2,nact,listact,na0,latoms,
     &                       apcorr,acc,poten)
            call acceln2(npart,nact,ntot,listpt,nonay,eps2,
     &                   apcorr,acc,poten)

            if(itme.ge.1)then
               call cpu_time(tme2)
               timing(itme_hostcalc)=timing(itme_hostcalc) + tme2-tme1
               ncalls(itme_hostcalc)=ncalls(itme_hostcalc) + 1
            endif
         endif
      endif

      return
      end subroutine calcgravgroup
!----------------------------------------------------------------------
      subroutine calcgrav(itreept,iact,nact,isoft,npart,eps2,
     &                    na0,na1,na,nn,latoms,lnodes,apcorr,acc,poten)
!*********************************************************************
!  Calculate gravity on itreept, due to list of atoms and nodes.     *
!    Separate out atoms that are possibly neighbors (na0+1,na1)      *
!       from those that are definitely not (1,na0)                   *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      
      integer, intent(in):: itreept,iact,isoft,nact,na0,na1,npart
      integer, intent(in):: na,latoms(na), nn,lnodes(nn)
      double precision, intent(in):: eps2
      double precision, intent(inout):: apcorr(0:ndim,nact)
      double precision, intent(inout):: acc(ndim,npart), poten(npart)

      integer nonay
      double precision e2

      e2=zero
      if(isoft.eq.id_Plummer)e2=eps2
      if(na0.gt.0)then
         nonay=id_Plummer         
         call gravatom(itreept,iact,nonay,nact,npart,e2,1  ,na0,latoms,
     &                 apcorr,acc,poten)
      endif
      if(na.gt.na0)then
         call gravatom(itreept,iact,isoft,nact,npart,e2,na1,na, latoms,
     &                 apcorr,acc,poten)
      endif
      if(nn.gt.0)then
         call gravnode(itreept,iact,isoft,nact,npart,e2,    nn, lnodes,
     &                 apcorr,acc,poten)
      endif

      return
      end subroutine calcgrav
!----------------------------------------------------------------------
      subroutine acceln2(npart,nact,ntot,listpt,isoft,eps2,
     &                   apcorr,acc,poten)
!*********************************************************************
!                                                                    *
!  Do an n**2 acceleration calculation of particles in the listpt    *
!    list on each other                                              *
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************

      integer, intent(in):: isoft,npart
      integer, intent(in):: nact,ntot,listpt(ntot)
      double precision, intent(in)::eps2
      double precision, intent(inout):: apcorr(0:ndim,ntot)
      double precision, intent(inout):: acc(ndim,npart),poten(npart)

      integer j,itreept,iact,n,nla,nla1,ilast
      integer lista(max(maxclumpsize,maxbunchsize))

      nla=0
      do n=1,ntot
         nla=nla+1
         lista(nla)=listpt(n)
      enddo

      nla1=nla-1
      ilast=lista(nla)

      iact=0
      do j=1,nla
         itreept=listpt(j)
         if(active(itreept).ne.1)cycle
         iact=iact+1

         lista(j)=ilast
         call gravatom(itreept,iact,isoft,nact,npart,eps2,1,nla1,lista,
     &                 apcorr,acc,poten)
         lista(nla)=lista(j)
         lista(j  )=itreept
      enddo

      return
      end subroutine acceln2
!----------------------------------------------------------------------
      subroutine gravatom(itreept,iact,isoft,nact,npart,eps2,
     &                    na0,na1,latoms,apcorr,acc,poten)
!*********************************************************************
!                                                                    *
!  Evaluate acceleration of particle itreept,                        *
!  due to a list of na other particles contained in latoms           *
!                                                                    *
!  If kernel softening is used:                                      *
!     currently hard coded to do softening via the standard SPH      *
!     W4 kernel in 3d, with an additional modification for 2d.       *
!     For details of the modification and the reasons for it         *
!     see Nelson 2006 MNRAS "Resolution Requirements..."             *
!                                                                    *
!  Could be a good thing to use Kahan's algorithm in this summation  *
!  Try this sometime, using apcorr array                             *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in):: itreept,iact,isoft,nact,npart
      integer, intent(in):: na0,na1,latoms(na1)
      double precision, intent(in)::eps2
      double precision, intent(inout):: apcorr(0:ndim,nact)
      double precision, intent(inout):: acc(ndim,npart),poten(npart)

      double precision, parameter::twntysevths=20d0/7d0
      double precision, parameter::sixth=1d0/6d0
      double precision, parameter::fiftnth=1d0/15d0
      double precision, parameter::thirtyth=1d0/30d0
      integer j,n,m_sph
      double precision eps2sml,fx,fy,fz,pot
      double precision dx,dy,dz,em,emfrac,d2,di,di3,hsum,hmean
      double precision vinv,v,v2,v3,v4,v5,v6,sum,emdi3

      if(na1.eq.0)return

      eps2sml = small
      if(isoft.eq.id_Plummer)eps2sml = eps2sml + eps2
      fx =zero  ; fy =zero  ; fz =zero  ; pot=zero
      m_sph=linktreetosph(itreept)

#if NDIM==2

      select case(isoft)
      case(id_varkern,id_fixkern)

         do j=na0,na1
            n=latoms(j)
            dx  =treept(1,n)-treept(1,itreept)
            dy  =treept(2,n)-treept(2,itreept)
            d2  =dx*dx+dy*dy + eps2sml
            di  =one/sqrt(d2)
            di3 =di*di*di
            hsum=treept(ihnode,n)+treept(ihnode,itreept)
            if(d2.gt.hsum*hsum)then
               em=treept(ims,n)
               pot=pot + em*di
            else
               v=one/(di*hsum*half)
               v2=v*v
               v4=v2*v2
               v5=v4*v
               if(v2.gt.one)then         !v.le.2 guaranteed by d2.lt.hsum2
                  v3=v2*v
                  emfrac=-0.05d0*v5+0.375d0*v4-v3 + v2-0.05d0
               elseif(v2.gt.fourninth)then
                  emfrac= 0.15d0*v5-0.375d0*v4 + half*v2
               else
                  emfrac= 0.15d0*v5-0.375d0*v4 + half*v2
                  emfrac=emfrac*(three*v-ninefourth*v2)   !why? see Nelson 2006
               endif
               em=twntysevths*emfrac*treept(ims,n)
               pot= pot + em*di

            endif

            emdi3= em*di3
            fx   = fx  + emdi3*dx 
            fy   = fy  + emdi3*dy
         enddo

      case(id_Plummer)

         do j=na0,na1
            n=latoms(j)
            dx   =treept(1,n)-treept(1,itreept)
            dy   =treept(2,n)-treept(2,itreept)
            di   =one/sqrt(dx*dx+dy*dy + eps2sml)
            di3  =di*di*di
            emdi3=treept(ims,n)*di3
            fx   =fx + emdi3*dx
            fy   =fy + emdi3*dy
            pot  =pot + treept(ims,n)*di
         enddo

      case default
         stop 'gravatom:unknown softening'
      end select
      acc  (1,m_sph)=acc  (1,m_sph) + fx
      acc  (2,m_sph)=acc  (2,m_sph) + fy
      poten(  m_sph)=poten(  m_sph) + pot

#else

      select case(isoft)

      case(id_varkern,id_fixkern)

         do j=na0,na1
            n   =latoms(j)
            dx  =treept(1,n)-treept(1,itreept)
            dy  =treept(2,n)-treept(2,itreept)
            dz  =treept(3,n)-treept(3,itreept)
            d2  =dx*dx+dy*dy+dz*dz + eps2sml
            di  =one/sqrt(d2)
            di3 =di*di*di
            hsum=treept(ihnode,n)+treept(ihnode,itreept)

            if(d2.gt.hsum*hsum)then
               em =treept(ims,n)
               pot=pot + em*di
            else
               hmean=half*hsum
               vinv=hmean*di
               v=one/vinv
               v2=v*v
               v3=v2*v
               v4=v2*v2
               v5=v3*v2
               v6=v3*v3
               if(v2.gt.one)then          !v.le.2 guaranteed by d2.lt.hsum2
                  emfrac= -fiftnth+atethd*v3-three*v4+1.2d0*v5-sixth*v6
                  sum= fiftnth*vinv -1.6d0 + forthd*v2 - v3 + 
     &                   0.3d0*v4 -thirtyth*v5
                  pot=pot-treept(ims,n)*sum/hmean
               else
                  emfrac=forthd*v3 - 1.2d0*v5 + half*v6
                  sum=-1.4d0 + twothd*v2 - 0.3d0*v4 + 0.1d0*v5
                  pot=pot-treept(ims,n)*sum/hmean
               endif
               em=emfrac*treept(ims,n)

            endif
            emdi3=em*di3
            fx   =fx + emdi3*dx
            fy   =fy + emdi3*dy
            fz   =fz + emdi3*dz
         enddo

      case(id_Plummer)
         do j=na0,na1
            n    =latoms(j)
            dx   =treept(1,n)-treept(1,itreept)
            dy   =treept(2,n)-treept(2,itreept)
            dz   =treept(3,n)-treept(3,itreept)
            di   =one/sqrt(dx*dx+dy*dy+dz*dz + eps2sml)
            di3  =di*di*di
            emdi3=treept(ims,n)*di3
            fx   =fx + emdi3*dx
            fy   =fy + emdi3*dy
            fz   =fz + emdi3*dz
            pot  =pot + treept(ims,n)*di
         enddo

      case default

         stop 'gravatom:unknown softening'

      end select
      acc  (1,m_sph)=acc  (1,m_sph) + fx
      acc  (2,m_sph)=acc  (2,m_sph) + fy
      acc  (3,m_sph)=acc  (3,m_sph) + fz
      poten(  m_sph)=poten(  m_sph) + pot

#endif

      return
      end subroutine gravatom
!----------------------------------------------------------------------
      subroutine gravatomgrp(isoft,npart,eps2,nact,listact,na,latoms,
     &                       apcorr,acc,poten)
!*********************************************************************
!                                                                    *
!  Evaluate acceleration of particle itreept,                        *
!  due to a list of na other particles contained in latoms           *
!                                                                    *
!  This version expect that *no* particles are neighbors that        *
!  need kernel softening.                                            *
!                                                                    *
!  Could be a good thing to use Kahan's algorithm in this summation  *
!  Try this sometime, using apcorr array                             *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in):: isoft,npart
      integer, intent(in):: nact,listact(nact)
      integer, intent(in):: na,latoms(na)
      double precision, intent(in)::eps2
      double precision, intent(inout):: apcorr(0:ndim,nact)
      double precision, intent(inout):: acc(ndim,npart),poten(npart)

      integer, parameter:: nvals=ndim1
      integer, parameter:: npartvals=L1data/nvals/idblesize
      integer, parameter:: jpartx=1,jparty=2,jpartz=3,jpartm=ims
      double precision partval(nvals,npartvals)

      integer iter,jstart,j,jj,jpart,n,n1,n2,num
      integer itreept,m_sph
      double precision eps2sml
      double precision dx,dy,dz,emdi3,di,di3
      double precision fx,fy,fz,pot

      if(na.eq.0)return

      eps2sml = small
      if(isoft.eq.id_Plummer)eps2sml = eps2sml + eps2

      iter=1
      n2=0
      do while(n2.lt.na)
         n1=1 + npartvals*(iter-1)
         n2=    npartvals*iter
         n2=min(n2,na)
         iter=iter+1

         jstart=1-n1
         do jj=n1,n2
            n=latoms(jj)
            j=jstart+jj
            partval(jpartm,j)=treept(ims,n)
            partval(jpartx,j)=treept(1  ,n)
            partval(jparty,j)=treept(2  ,n)
#if NDIM==3
            partval(jpartz,j)=treept(3  ,n)
#endif
         enddo
         num=n2-n1+1

         do jpart=1,nact
            itreept=listact(jpart)
            m_sph=linktreetosph(itreept)
            fx =zero  ; fy =zero  ; fz= zero  ; pot=zero
#if NDIM==2
            do n=1,num
               dx   =partval(jpartx,n)-treept(1,itreept)
               dy   =partval(jparty,n)-treept(2,itreept)
               di   =one/sqrt(dx*dx+dy*dy + eps2sml)
               di3  =di*di*di
               emdi3=partval(jpartm,n)*di3

               fx  = fx  + emdi3*dx
               fy  = fy  + emdi3*dy
               pot = pot + partval(jpartm,n)*di
            enddo
#else
            do n=1,num
               dx   =partval(jpartx,n)-treept(1,itreept)
               dy   =partval(jparty,n)-treept(2,itreept)
               dz   =partval(jpartz,n)-treept(3,itreept)
               di   =one/sqrt(dx*dx+dy*dy+dz*dz + eps2sml)
               di3  =di*di*di
               emdi3=partval(jpartm,n)*di3

               fx = fx + emdi3*dx
               fy = fy + emdi3*dy
               fz = fz + emdi3*dz
               pot = pot + partval(jpartm,n)*di

            enddo
            acc  (3,m_sph)=acc  (3,m_sph) + fz
#endif
            acc  (1,m_sph)=acc  (1,m_sph) + fx
            acc  (2,m_sph)=acc  (2,m_sph) + fy
            poten(  m_sph)=poten(  m_sph) + pot
         enddo !jpart particle loop
      enddo    !more atoms to calculate

      return
      end subroutine gravatomgrp
!----------------------------------------------------------------------
      subroutine gravnode(itreept,iact,isoft,nact,npart,eps2,nn,lnodes,
     &                    apcorr,acc,poten)
!*********************************************************************
!                                                                    *
!  Evaluate acceleration of itreept,                                 *
!  due to a list of nn other nodes contained in lnodes               *
!                                                                    *
!  Could be a good thing to use Kahan's algorithm in this summation  *
!  Try this sometime, using apcorr array                             *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in):: itreept,iact,isoft,nact,npart
      integer, intent(in):: nn,lnodes(nn)
      double precision, intent(in)::eps2
      double precision, intent(inout):: apcorr(0:ndim,nact)
      double precision, intent(inout):: acc(ndim,npart),poten(npart)

      double precision, parameter:: sixthneg=-1d0/6d0

      integer j,n,m_sph
      double precision dx,dy,dz,di,di2,di3,fpr,fpprr
      double precision tx,ty,tz,rqr,fac
      double precision eps2sml,fx,fy,fz,pot


      eps2sml = small
      if(isoft.eq.id_Plummer)eps2sml = eps2sml + eps2
      fx =zero  ; fy =zero  ; fz =zero  ; pot=zero
      m_sph=linktreetosph(itreept)

#if NDIM==2

      do j=1,nn
         n=lnodes(j)
         dx=treept(1,n)-treept(1,itreept)
         dy=treept(2,n)-treept(2,itreept)
         di=one/sqrt(dx*dx+dy*dy + eps2sml)        ! 1.0/r
         di2=di*di                                 ! 1.0/r**2
         di3=di2*di                                ! 1.0/r**3
         fpr=-three*di3*di2                        !-3.0/r**5
         fpprr=-fivhalf*fpr*di2                    !-7.5/r**7
         tx=pole(iqxx,n)*dx+pole(iqxy,n)*dy
         ty=pole(iqxy,n)*dx+pole(iqyy,n)*dy
         rqr=dx*tx+dy*ty
         fac=treept(ims,n)*di3 + rqr*fpprr
     &           + half*(pole(iqxx,n)+pole(iqyy,n))*fpr
         fx=fx + fac*dx + fpr*tx
         fy=fy + fac*dy + fpr*ty
         pot=pot + treept(ims,n)*di + sixthneg*rqr*fpr
      enddo
#else
      do j=1,nn
         n=lnodes(j)
         dx=treept(1,n)-treept(1,itreept)
         dy=treept(2,n)-treept(2,itreept)
         dz=treept(3,n)-treept(3,itreept)
         di =one/sqrt(dx*dx+dy*dy+dz*dz + eps2sml)   ! 1.0/r
         di2=di*di                                   ! 1.0/r**2
         di3=di2*di                                  ! 1.0/r**3
         fpr  =-three*di3*di2                        !-3.0/r**5
         fpprr=-fivhalf*fpr*di2                      !-7.5/r**7
         tx=pole(iqxx,n)*dx+pole(iqxy,n)*dy+pole(iqxz,n)*dz
         ty=pole(iqxy,n)*dx+pole(iqyy,n)*dy+pole(iqyz,n)*dz
         tz=pole(iqxz,n)*dx+pole(iqyz,n)*dy+pole(iqzz,n)*dz
         rqr=dx*tx+dy*ty+dz*tz
         fac=treept(ims,n)*di3 + rqr*fpprr
     &        + half*(pole(iqxx,n)+pole(iqyy,n)+pole(iqzz,n))*fpr
         fx=fx + fac*dx + fpr*tx
         fy=fy + fac*dy + fpr*ty
         fz=fz + fac*dz + fpr*tz
         pot=pot + treept(ims,n)*di + sixthneg*rqr*fpr
      enddo
      acc  (3,m_sph)=acc  (3,m_sph) + fz
#endif
      acc  (1,m_sph)=acc  (1,m_sph) + fx
      acc  (2,m_sph)=acc  (2,m_sph) + fy
      poten(  m_sph)=poten(  m_sph) + pot

      return
      end subroutine gravnode
!----------------------------------------------------------------------
      subroutine gravnodegrp(isoft,npart,eps2,nact,listact,
     &                       nn,lnodes,apcorr,acc,poten)
!*********************************************************************
!                                                                    *
!  Evaluate acceleration of active particles in listpt               *
!  due to a list of nn other nodes contained in lnodes               *
!                                                                    *
!  Could be a good thing to use Kahan's algorithm in this summation  *
!  Try this sometime, using apcorr array                             *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************

      integer, intent(in):: isoft,npart
      integer, intent(in):: nact,listact(nact)
      integer, intent(in):: nn,lnodes(nn)
      double precision, intent(in):: eps2
      double precision, intent(inout):: apcorr(0:ndim,nact)
      double precision, intent(inout):: acc(ndim,npart),poten(npart)

      integer, parameter:: nvals=ndim1+moments
      integer, parameter:: npoles=L1data/nvals/idblesize
      integer, parameter:: ipolx=1,ipoly=2,ipolz=3,ipolm=ims

! Need this to avoid annoying `out of array limits' compiler warnings 
!#ifdef GRAPE
!      integer, parameter:: ipolqxx=ndim1+1,ipolqyy=ndim1+1
!      integer, parameter:: ipolqxy=ndim1+1
!      integer, parameter:: ipolqzz=ndim1+1,ipolqxz=ndim1+1
!      integer, parameter:: ipolqyz=ndim1+1
!#else
      integer, parameter:: ipolqxx=ndim1+1,ipolqyy=ndim1+2
      integer, parameter:: ipolqxy=ndim1+3
      integer, parameter:: ipolqzz=ndim1+4,ipolqxz=ndim1+5
      integer, parameter:: ipolqyz=ndim1+6
!#endif
      double precision, parameter:: sixthneg=-1d0/6d0
      integer iter,n1,n2,jstart,j,jj,n,num,jpart,itreept,m_sph
      double precision poleval(nvals,npoles)
      double precision dx,dy,dz,di,di2,di3,fpr,fpprr
      double precision tx,ty,tz,trace,rqr,fac
      double precision eps2sml,fx,fy,fz,pot

      eps2sml = small
      if(isoft.eq.id_Plummer)eps2sml = eps2sml + eps2

      iter=1
      n2=0
      do while(n2.lt.nn)
         n1=1 + npoles*(iter-1)
         n2=    npoles*iter
         n2=min(n2,nn)
         iter=iter+1

         jstart=1-n1
         do jj=n1,n2
            n=lnodes(jj)
            j=jstart+jj
            poleval(ipolm  ,j)=treept(ims,n)
            poleval(ipolx  ,j)=treept(1  ,n)
            poleval(ipoly  ,j)=treept(2  ,n)
            poleval(ipolqxx,j)=pole(iqxx ,n)
            poleval(ipolqyy,j)=pole(iqyy ,n)
            poleval(ipolqxy,j)=pole(iqxy ,n)
#if NDIM==3
            poleval(ipolz  ,j)=treept(3  ,n)
            poleval(ipolqzz,j)=pole(iqzz ,n)
            poleval(ipolqxz,j)=pole(iqxz ,n)
            poleval(ipolqyz,j)=pole(iqyz ,n)
#endif
         enddo
         num=n2-n1+1

         do jpart=1,nact
            itreept=listact(jpart)

            fx =zero  ; fy =zero  ; fz =zero  ; pot=zero
            m_sph=linktreetosph(itreept)
#if NDIM==2
            do n=1,num
               dx=poleval(ipolx,n)-treept(1,itreept)
               dy=poleval(ipoly,n)-treept(2,itreept)
               di=one/sqrt(dx*dx+dy*dy + eps2sml)             ! 1.0/r
               di2=di*di                                      ! 1.0/r**2
               di3=di2*di                                     ! 1.0/r**3
               fpr=-three*di3*di2                             !-3.0/r**5
               fpprr=-fivhalf*fpr*di2                         !-7.5/r**7
               tx=poleval(ipolqxx,n)*dx+poleval(ipolqxy,n)*dy
               ty=poleval(ipolqxy,n)*dx+poleval(ipolqyy,n)*dy
               rqr=dx*tx+dy*ty
               trace=poleval(ipolqxx,n)+poleval(ipolqyy,n)
               fac=poleval(ipolm,n)*di3 + rqr*fpprr + half*trace*fpr
               fx=fx + fac*dx + fpr*tx
               fy=fy + fac*dy + fpr*ty
               pot=pot + poleval(ipolm,n)*di + sixthneg*rqr*fpr
            enddo
#else
            do n=1,num
               dx=poleval(ipolx,n)-treept(1,itreept)
               dy=poleval(ipoly,n)-treept(2,itreept)
               dz=poleval(ipolz,n)-treept(3,itreept)
               di =one/sqrt(dx*dx+dy*dy+dz*dz + eps2sml)        ! 1.0/r
               di2=di*di                                        ! 1.0/r**2
               di3=di2*di                                       ! 1.0/r**3
               fpr  =-three*di3*di2                             !-3.0/r**5
               fpprr=-fivhalf*fpr*di2                           !-7.5/r**7
               tx=poleval(ipolqxx,n)*dx + poleval(ipolqxy,n)*dy
               ty=poleval(ipolqxy,n)*dx + poleval(ipolqyy,n)*dy
               tz=poleval(ipolqxz,n)*dx + poleval(ipolqyz,n)*dy
               trace=poleval(ipolqxx,n) + poleval(ipolqyy,n)
               trace=trace + poleval(ipolqzz,n)
               tx   =tx    + poleval(ipolqxz,n)*dz
               ty   =ty    + poleval(ipolqyz,n)*dz
               tz   =tz    + poleval(ipolqzz,n)*dz
               rqr  =dx*tx+dy*ty+dz*tz
               fac=poleval(ipolm,n)*di3 + rqr*fpprr + half*trace*fpr
               fx=fx + fac*dx + fpr*tx
               fy=fy + fac*dy + fpr*ty
               fz=fz + fac*dz + fpr*tz
               pot=pot + poleval(ipolm,n)*di + sixthneg*rqr*fpr
            enddo
            acc  (3,m_sph)=acc  (3,m_sph) + fz
#endif
            acc  (1,m_sph)=acc  (1,m_sph) + fx
            acc  (2,m_sph)=acc  (2,m_sph) + fy
            poten(  m_sph)=poten(  m_sph) + pot

         enddo !particle loop
      enddo    !done with all nodes in list

      return
      end subroutine gravnodegrp
!----------------------------------------------------------------------
      subroutine initpole(npart)

      integer, intent(in):: npart

      integer i


!$omp parallel do default(none)
!$omp& shared(npart,pole)
!$omp& private(i)
      do i=1,2*npart
         pole(iqxx,i)=zero
         pole(iqxy,i)=zero
         pole(iqyy,i)=zero
#if NDIM==3
         pole(iqzz,i)=zero
         pole(iqxz,i)=zero
         pole(iqyz,i)=zero
#endif
      enddo

      return
      end subroutine initpole
!----------------------------------------------------------------------
      subroutine revpole(igrape,irevise,imac,treeacc)
!*********************************************************************
!                                                                    *
! Calculate center of mass and multipole moments of                  *
! progressively higher nodes in the tree                             *
!                                                                    *
! treept(iconvrg,i)= multipole convergence radius/treeacc            *
! treept(ihnode ,i)= physical size of the node. At the end of rtree  *
!             (call gethnode) it is redefined for clumps and         *
!             their children, if the clumps are small enough         *
!             to make this redefinition efficient [[need to specify]]*
!                                                                    *
! pole( ,i) = multipole moment                                       *
!                                                                    *
! Use a backwards (end to root) traversal to update the tree.        *
! In other words, start at ntotnodes and proceed to nroot,           *
! hitting all children before their parents.                         *
!                                                                    *
! n1 and n2 are the last and first nodes to be examined in an        *
! independent branch. That means that n1 is the last atom in         *
! the branch and n2 is the branch root.  In the fixup stage,         *
! the roles are reversed. We are looking at orphan nodes in          *
! between independent branches. That means that n1 is the node       *
! immediately preceeding the branch root, and n2 is either the       *
! last atom in the previous branch, or the root of the entire        *
! tree.                                                              *
!                                                                    *
!  Subroutine by A. Nelson, with some MACs by M. Wetzstein           *
!                                                                    *
!*********************************************************************

      integer, intent(in):: igrape,irevise,imac
      double precision, intent(in):: treeacc

      logical fixup,succeed
      integer imac_inloop
      integer icpu,nbotbranch,ntopbranch
      integer n1,n2,i
      integer imacabs,j,jbranch,me,my_dau,m1,m2
      double precision treeaccinv,b2,bmax,sb2,ssb2
      double precision emratm1,emratm2,emred,difx,dify,difz,rmag
      character(len=20), parameter:: where = ('revpole')

      if(idbg.ge.3)print *,where

!#ifdef GRAPE
!!
!! See the grape ifdef in the header of this file 
!! for why these conditions present problems. 
!!
!      if(igrape.eq.0)then
!         print *,'You have compiled in the grape hardware'
!         print *,'option, but are not using it. This is'
!         print *,'currently disabled to save memory usage'
!         print *,'given to the multipole moments'
!         print *,'You must compile with GRAPE=0 or run with'
!         print *,'igrape.ne.0'
!         stop'You have compiled in grape, but are not using it'
!      endif
!      if(igrape.eq.1 .and. imac.eq.id_SWabs)then
!         print *,'You are using the grape hardware and'
!         print *,'have selected the Salmon & Warren'
!         print *,'Absolute error criterion MAC'
!         print *,'Currently this is disabled to save memory'
!         stop'Using grape and SWabs MAC'
!      endif
!#endif

      imacabs=abs(imac)
      treeaccinv=one/treeacc

      if(imacabs.ne.id_angle .and. imacabs.ne.id_SWabs .and.
     &   imacabs.ne.id_gadget)stop'revpole:unknown MAC'

      fixup=.false.

      do i=1,2

      succeed=.true.

!$omp parallel if(.not.fixup) default(none)
!$omp& firstprivate(imacabs,treeaccinv) shared(ntotnodes)
!$omp& shared(fixup,succeed,imac_inloop,nroot,numbranches,lstbranches)
!$omp& shared(link,treept,pole,nbrnchstart,numcpus_sav)
!$omp& private(icpu,ntopbranch,nbotbranch)
!$omp& private(jbranch,n1,n2,me,my_dau,m1,m2)
!$omp& private(emratm1,emratm2,emred,rmag,difz,dify,difx)
!$omp& private(b2,bmax,sb2,ssb2)
!$omp do
      do icpu=numcpus_sav,1,-1
         ntopbranch=nbrnchstart(2,icpu)
         nbotbranch=nbrnchstart(1,icpu)
         do jbranch=ntopbranch,nbotbranch,-1
            if(.not.fixup)then
               if(jbranch.eq.numbranches)then
                  n1=ntotnodes
               else
                  me=lstbranches(jbranch+1)
                  my_dau=abs(link(idau,me))
                  do while(my_dau.ne.0)
                     me=me-1
                     my_dau=abs(link(idau,me))
                  enddo
                  n1=me
               endif
               n2=lstbranches(jbranch)
            else
               n1=lstbranches(jbranch)-1
               if(jbranch.eq.1)then
                  n2=nroot
               else
                  me=n1
                  my_dau=abs(link(idau,me))
                  do while(my_dau.ne.0)
                     me=me-1
                     my_dau=abs(link(idau,me))
                  enddo
                  n2=me
               endif
            endif

            do me=n1,n2,-1
               if(link(idau,me).eq.0)then
                  pole(iqxx,me)=zero
                  pole(iqyy,me)=zero
                  pole(iqxy,me)=zero
#if NDIM==3
                  pole(iqzz,me)=zero
                  pole(iqxz,me)=zero
                  pole(iqyz,me)=zero
#endif
                  cycle
               endif

               m1=abs(link(idau,me))
               m2=abs(link(isib,m1))
               treept(ims,me)=treept(ims,m1) + treept(ims,m2)
               emratm1      =treept(ims,m1)/treept(ims,me)
               emratm2      =one - emratm1
               emred        =emratm1*emratm2*treept(ims,me)
               difx         =treept(1,m2) - treept(1,m1)
               dify         =treept(2,m2) - treept(2,m1)
               treept(1,me) =treept(1,m1) + emratm2*difx
               treept(2,me) =treept(2,m1) + emratm2*dify
#if NDIM==2
               rmag         =sqrt(difx*difx + dify*dify)
#else
               difz         =treept(3,m2) - treept(3,m1)
               treept(3,me) =treept(3,m1) + emratm2*difz
               rmag         =sqrt(difx*difx + dify*dify + difz*difz)
               pole(iqxz,me)=emred*difx*difz+pole(iqxz,m1)+pole(iqxz,m2)
               pole(iqyz,me)=emred*dify*difz+pole(iqyz,m1)+pole(iqyz,m2)
               pole(iqzz,me)=emred*difz*difz+pole(iqzz,m1)+pole(iqzz,m2)
#endif
               pole(iqxx,me)=emred*difx*difx+pole(iqxx,m1)+pole(iqxx,m2)
               pole(iqyy,me)=emred*dify*dify+pole(iqyy,m1)+pole(iqyy,m2)
               pole(iqxy,me)=emred*difx*dify+pole(iqxy,m1)+pole(iqxy,m2)
               treept(ihnode,me)=max(emratm2*rmag+treept(ihnode,m1),
     &                               emratm1*rmag+treept(ihnode,m2))
!
! The MACs
!
               select case(imacabs)
               case(id_angle)
                  treept(iconvrg,me)= treeaccinv*treept(ihnode,me)
               case(id_gadget)
                  treept(iconvrg,me)=
     &                treeaccinv*treept(ims,me)*treept(ihnode,me)**4
               case(id_SWabs)
                  b2=pole(iqxx,me)+pole(iqyy,me)
#if NDIM==3
     &                            +pole(iqzz,me)
#endif
                  bmax=treept(ihnode,me)
                  sb2=sqrt(three*b2*treeaccinv)    !treeaccinv = 1/err_max in paper
                  ssb2=sqrt(fourth*bmax**2+sb2)

                  treept(iconvrg,me)= half*bmax+ssb2
               case default
                  succeed=.false.
                  imac_inloop=imacabs
               end select
            enddo   !nodes in branch

         enddo      !branches
      enddo         !icpu
!$omp end parallel

      if(.not.succeed)then
         print *,'MAC setting in revpole was not valid'
         print *,'imac inside  the loop was:',imac_inloop
         print *,'imac outside the loop was:',imacabs
         stop'Unknown MAC in revpole'
      endif

      fixup=.true.

      enddo

      return
      end subroutine revpole
!----------------------------------------------------------------------
#ifndef GRAPE
!*********************************************************************
!  Below are routines that serve as stubs for the GRAPE routines     *
!  that don't exist, since the GRAPE preprocessor flag is not set    *
!*********************************************************************
      subroutine grape_tree(iclump,isoft,eps2,npart,nact,listact,
     &                      nla0,nla,latoms,vel,apcorr,acc,poten)

      integer, intent(in):: iclump,isoft,npart
      integer, intent(in):: nact,listact(nact)
      integer, intent(inout):: nla0, nla ,latoms(nla)
      double precision, intent(in)   :: eps2
      double precision, intent(in)   :: vel(ndim,npart)
      double precision, intent(inout):: apcorr(0:ndim,nact)
      double precision, intent(inout):: acc(ndim,npart), poten(npart)

      integer listpt(max(maxbunchsize,maxclumpsize)),ntot
      integer nonay
      integer n,nn

      ntot = nla - nla0+1
      do n=nla0+1,nla
         nn=n-nla0
         listpt(nn)=latoms(n)
      enddo

      nonay=id_Plummer
      call gravatomgrp(nonay,npart,eps2,ntot,listpt,nla0,latoms,
     &                 apcorr,acc,poten)
      call acceln2(npart,nact,ntot,listpt,nonay,eps2,apcorr,acc,poten)

      return
      end subroutine grape_tree
!----------------------------------------------------------------------
      subroutine grape_dirsum(npart,isoft,eps2,posmas,vel,acc,poten)

      integer, intent(in):: npart,isoft
      double precision, intent(in)::eps2
      double precision, intent(in):: posmas(ndim1,npart),vel(ndim,npart)
      double precision, intent(inout):: acc(ndim,npart), poten(npart)

      call host_dirsum(npart,isoft,eps2,posmas,vel,acc,poten)
      return
      end subroutine grape_dirsum
!----------------------------------------------------------------------
#endif
      end module forcelaw
