      module grape
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!*********************************************************************
! This file contains the VINE module to do gravity using the         *
! GRAPE 3 special purpose hardware boards. The routines here         *
! are called from the tree.F module.                                 *
!                                                                    *
!  Written by Markus Wetzstein, Andy Nelson and Thorsten Naab        *
!*********************************************************************
#if GRAPE==3
#ifndef SELFFORCE
      grape3 module: Grape flag set, but self force flag is not'
#endif
#else
      grape3 module: No or wrong grape preprocessor flag exists!
#endif

      use dimens
      use treevars
      use miscvars, only: ibctype
      use indivtsvars
      use constants
      use perform

      implicit none
      save

      private
      public grape_tree,grape_dirsum,lock_grape,unlock_grape

!  GRAPE-3 parameters
!
! set ig3boards to number of g3 boards in your system
!
      integer, parameter :: ig3boards=1
      integer, parameter :: ig3chips=ig3boards*8
!
! GRAPE-3 functions
!
      integer g3init,g3free
      integer g3nchips,g3maxptcl
!
! gravitational smoothing length for GRAPE
!
      real g3eps2
!
! Data for j particles to send to grape
!
      integer g3_j
      real g3_xj(3), g3_mj
!
! Data for i particles to send to grape:
!
      real g3_xi(3,ig3chips)
!
! Data to retrieve from grape
!
      real g3acc(3,ig3chips),g3pot(ig3chips)

      contains
!-----------------------------------------------------
      subroutine lock_grape()
      integer ierr

      ierr=g3init()
      if(ierr.ne.0)stop'error opening Grape3'
      
      return
      end subroutine lock_grape
!-----------------------------------------------------
      subroutine unlock_grape()
      integer ierr

      ierr=g3free()
      if(ierr.ne.0)stop'error closing Grape3'

      return
      end subroutine unlock_grape
!-----------------------------------------------------
      subroutine grape_tree(ibunch,isoft,eps2,npart,nact,listact,
     &                      nla0,nla,latoms,vel,apcorr,acc,poten)

! vel,npart aren't used here, they are just passed because the 
! G6 routine needs them and the call sequence has to be the same 
! for all grape_* routines in the different files

      integer, intent(in):: ibunch,isoft,npart
      integer, intent(in):: nact, listact(nact)
      integer, intent(in):: nla0, nla, latoms(nla)
      double precision, intent(in   ):: eps2
      double precision, intent(in   ):: vel(ndim,npart)
      double precision, intent(inout):: apcorr(0:ndim,nact)
      double precision, intent(inout):: acc(ndim,npart), poten(npart)

      integer nchips
      integer NJMAX,i,ii,j,k
      integer njlow,njhigh,nj
      integer nilow,nihigh,ni
      double precision gconst
      double precision g3massmin,g3massmax,dxmax

      real scalemass

      double precision tme1,tme2,tmeio1,tmeio2
      double precision tmecalc1,tmecalc2
      character(len=20), parameter:: where=('grape_tree')
      if(idbg.ge.2)print *,'where'
      if(itme.ge.1)call cpu_time(tme1)

      if(isoft.ne.id_Plummer)then
         print *,'You have not selected Plummer softening'
         print *,'and you have selected grape. This is not'
         print *,'possible.' 
         stop 'Not using Plummer softening with Grape'
      endif

      gconst=one
      g3eps2 = real(eps2)

      nchips = g3nchips()
      NJMAX  = g3maxptcl()

      if (nchips .gt. ig3chips) then
         print *,'nchips                      :',nchips
         print *,'max compiled number of chips:',ig3chips
         print *,'Using grape inefficiently in',where
         print *,'Check module grape3.F'
      endif

      dxmax=-1d100
      g3massmin=1d100 ; g3massmax=0d0
      do i=1,nla
         ii=latoms(i)
         dxmax=max(dxmax,abs(treept(1,ii)-treept(1,ibunch)),
     &                   abs(treept(2,ii)-treept(2,ibunch)),
     &                   abs(treept(3,ii)-treept(3,ibunch)))
         g3massmin=min(treept(ims,ii),g3massmin)
         g3massmax=max(treept(ims,ii),g3massmax)
      enddo

      call setresolution(g3massmin,g3massmax,dxmax)
!
! mass correction factor for tree code
!
      scalemass=one
      if(npart.gt.100000)scalemass=1e5/real(npart)
      call g3setmasscor(scalemass)

      njlow=1  ; njhigh=min(NJMAX,nla)
      do while(njlow.le.nla)
         if(itme.ge.1)call cpu_time(tmeio1)
!
! Cycle through groups of nj particles/nodes that contribute forces 
!
         nj=njhigh - njlow + 1

         call g3setn(nj)

         do i=njlow,njhigh
            ii=latoms(i)
            g3_j =i-njlow                     !j numbered from 0
            g3_xj(1) = real(treept(1  ,ii))
            g3_xj(2) = real(treept(2  ,ii))
            g3_xj(3) = real(treept(3  ,ii))
            g3_mj    = real(treept(ims,ii))
            call g3setxj(g3_j,g3_xj)
            call g3setmj(g3_j,g3_mj)
         enddo
         do i=1,nchips
            call g3seth2(0e0,i-1)
         enddo
         call g3seteps2(g3eps2)

         if(itme.ge.1)then
            call cpu_time(tmeio2)
            timing(itme_grpIO)=timing(itme_grpIO)+(tmeio2-tmeio1)
            call cpu_time(tmecalc1)
         endif

         nilow = 1 ; nihigh = min(nact,nchips,ig3chips)
         do while(nilow.le.nact)
!
! Cycle through groups of ni particles needing forces
!
            ni = nihigh - nilow + 1

            do i=nilow,nihigh
               ii=i - nilow + 1
               k=listact(i)
               g3_xi(1,ii)=real(treept(1,k))
               g3_xi(2,ii)=real(treept(2,k))
               g3_xi(3,ii)=real(treept(3,k))
            enddo

            call g3frc(g3_xi,g3acc,g3pot,ni)

            do i=nilow,nihigh
               ii = i - nilow + 1
               k=linktreetosph(listact(i))
               acc  (1,k) = acc  (1,k) + gconst*dble(g3acc(1,ii))
               acc  (2,k) = acc  (2,k) + gconst*dble(g3acc(2,ii))
               acc  (3,k) = acc  (3,k) + gconst*dble(g3acc(3,ii))
               poten(  k) = poten(  k) - gconst*dble(g3pot(  ii))
            enddo

            nilow  = nihigh + 1
            nihigh = min(nihigh + min(ig3chips,nchips),nact) 
         enddo
         if(itme.ge.1)then
            call cpu_time(tmecalc2)
            timing(itme_grpcalc)=timing(itme_grpcalc)+
     &                                (tmecalc2-tmecalc1)
         endif

         njlow  = njhigh + 1
         njhigh = min(njhigh + NJMAX,nla)
      enddo

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_grape)=timing(itme_grape)+(tme2-tme1)
         ncalls(itme_grape)=ncalls(itme_grape)+1
         ncalls(itme_grpIO)=ncalls(itme_grpIO)+1
         ncalls(itme_grpcalc)=ncalls(itme_grpcalc)+1
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine grape_tree
!-----------------------------------------------------
      subroutine grape_dirsum(npart,isoft,eps2,posmas,vel,acc,poten)

      integer, intent(in):: npart,isoft
      double precision, intent(in   ):: eps2
      double precision, intent(in   ):: posmas(ndim1,npart)
      double precision, intent(in   ):: vel   (ndim ,npart)
      double precision, intent(inout):: acc(ndim,npart),poten(npart)

      double precision gconst
      double precision g3massmin,g3massmax,dxmax
      integer nchips
      integer i,ii,j,k,NJMAX
      integer njlow,njhigh,nj
      integer nilow,nihigh,ni
      real scalemass

      double precision tme1,tme2,tmeio1,tmeio2
      double precision tmecalc1,tmecalc2
      character(len=20), parameter:: where=('grape_dirsum')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

      if(isoft.ne.id_Plummer)then
         print *,'You have not selected Plummer softening'
         print *,'and you have selected grape. This is not'
         print *,'possible.' 
         stop 'Not using Plummer softening with Grape'
      endif

      gconst=one
      g3eps2  = real(eps2)

      nchips  = g3nchips ()
      NJMAX   = g3maxptcl()
      if (nchips .gt. ig3chips) then
         print *,'nchips                      :',nchips
         print *,'max compiled number of chips:',ig3chips
         print *,'Using grape inefficiently in',where
         print *,'Check module grape3.F'
      endif

      g3massmin=1d100 ; g3massmax=0d0
      do i=1,npart
         g3massmin=min(g3massmin,posmas(ims,i))
         g3massmax=max(g3massmax,posmas(ims,i))
      enddo

      if(ibctype.eq.0)then
         dxmax=128d0
      else
         dxmax=sqrt(12d0)
      endif

      call setresolution(g3massmin,g3massmax,dxmax)
!
! mass correction factor: only needed for tree code
!
      scalemass=1.
      call g3setmasscor(scalemass)

      njlow=1 ; njhigh=min(npart,NJMAX)
      do while(njlow.le.npart)
         if(itme.ge.1)call cpu_time(tmeio1)
!
! Cycle through groups of nj particles that contribute forces
!
         nj=njhigh-njlow + 1

         call g3setn(nj)
         do i=njlow,njhigh
            g3_j =i-njlow                   !j numbered from 0
            g3_xj(1) = real(posmas(1  ,i))
            g3_xj(2) = real(posmas(2  ,i))
            g3_xj(3) = real(posmas(3  ,i))
            g3_mj    = real(posmas(ims,i))
            call g3setxj(g3_j,g3_xj)
            call g3setmj(g3_j,g3_mj)
         enddo
         do i=1,nchips
            call g3seth2(0.0,i-1)
         enddo

         call g3seteps2(g3eps2)

         if(itme.ge.1)then
            call cpu_time(tmeio2)
            timing(itme_grpIO)=timing(itme_grpIO)+(tmeio2-tmeio1)
            call cpu_time(tmecalc1)
         endif

         nilow = 1 ; nihigh = min(ig3chips,nchips,nactive)
         do while(nilow.le.nactive)
!
! Cycle through groups of ni particles needing forces
!
            ni = nihigh - nilow + 1

            do i=nilow,nihigh 
               ii=i - nilow + 1
               k=iactive(i)
               g3_xi(1,ii)=real(posmas(1,k))
               g3_xi(2,ii)=real(posmas(2,k))
               g3_xi(3,ii)=real(posmas(3,k))
            enddo

            call g3frc(g3_xi,g3acc,g3pot,ni)

            do i=nilow,nihigh
               ii=i - nilow + 1
               k=iactive(i)
               acc  (1,k)=acc  (1,k) + gconst*dble(g3acc(1,ii))
               acc  (2,k)=acc  (2,k) + gconst*dble(g3acc(2,ii))
               acc  (3,k)=acc  (3,k) + gconst*dble(g3acc(3,ii))
               poten(  k)=poten(  k) - gconst*dble(g3pot(  ii))
            enddo

            nilow = nihigh + 1
            nihigh = min(nihigh + min(ig3chips,nchips),nactive) 
         enddo
         if(itme.ge.1)then
            call cpu_time(tmecalc2)
            timing(itme_grpcalc)=timing(itme_grpcalc)+
     &                                (tmecalc2-tmecalc1)
         endif

         njlow  = njhigh + 1
         njhigh = min(njhigh + NJMAX,npart)
      enddo

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_grape)=timing(itme_grape)+(tme2-tme1)
         ncalls(itme_grape)=ncalls(itme_grape)+1
         ncalls(itme_grpIO)=ncalls(itme_grpIO)+1
         ncalls(itme_grpcalc)=ncalls(itme_grpcalc)+1
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine grape_dirsum
!-----------------------------------------------------
      subroutine setresolution(g3massmin,g3massmax,dxmax)
!
! Resolution limits as defined in the grape3 users manual
!
!   position resolution is limited to 2**18 different values
!   Therefore we want to scale the positions such that
!   the size of the simulation is smaller that this, and
!   so that the increment between neighboring particles is
!   at least one position increment.
!
!   Similarly mass resolution must be greater than the 
!   minimum mass of any particle 
!
!   The combination must satisfy the condition:
!
!   maximum mass < g3_mres * 64*g3eps/g3_xres
!
      double precision, intent(in):: dxmax,g3massmin,g3massmax

      real g3res,g3masslim
      real g3_xres,g3_mres

      g3_xres=real(dxmax/262144d0)
      g3_mres=real(g3massmin)

      g3res = sqrt(g3eps2)/g3_xres
      g3masslim = g3_mres * (64.0*min(g3res,g3res**2))

      if(g3masslim.lt.real(g3massmax))then
         print *,'Upper mass limit too low:',g3masslim
         print *,'Should be               :',g3massmax
         print *,'Overflows are likely'
         print *,'Try decreasing gmasslim parameter in insph.'
         stop'Overflows likely in grape3 module'
      endif

      call g3setscales(g3_xres,g3_mres) 

      return
      end subroutine setresolution
!-----------------------------------------------------
      end module grape
