      module indivts
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!****************************************************************
!  This file contains the VINE module to allow individual       *
!  time steps in the integration of particles forward in time.  *
!                                                               *
!  These subroutines include the possibility of not using       *
!  individual time steps at all, via the flag `indts'           *
!                                                               *
!         With global timesteps, the timestep can fail,         *
!         and return a smaller timestep, so that the            *
!         code can go back and try again. This is not           *
!         possible with the individual timestep options.        *
!                                                               *
!    IF indts = 1 then it will use individual timesteps         *
!                 with the timesteps binned automatically       *
!                 in factors of two increments.                 *
!    IF indts = 0 then it will use a single timestep            *
!                 for all particles.                            *
!         if iusebin=0 then it will use the calculated best     *
!                   timestep, without regard to binning         *
!         if iusebin=1 then it will use the calculated best     *
!                   timestep bin, as defined by the             *
!                   individual timestep machinery.              *
!                                                               *
!  Various routines require as inputs:                          *
!      npart    :Number of particles                            *
!      npart_sph:Number of SPH particles                        *
!      dtrat    :Ratio of timesteps (next/last)                 *
!                                                               *
!  The routines return variously,                               *
!                                                               *
!      nactive  :The number of currently active particles of    *
!                any type (sph or nbody)                        *
!      nactivef :The number of currently active particles of    *
!                any type (sph or nbody), that are on full      *
!                timestep updates right now.                    *
!      iactive  :Index of the array location of particles.      *
!                                                               *
!        Storage order in iactive array:                        *
!           both full and half active (1..nactive)              *
!           fullstep active           (1..nactivef)             *
!           halfstep active           (nactivef+1..nactive)     *
!           inactive                  (nactive+1..npart)        *
!                                                               *
!      nactive_s :Same as for nactive and nactivef              *
!      nactivef_s:  but limited only to                         *
!      iactive_s :                  SPH particles               *
!                                                               *
!      nactive_n :Same as for nactive and nactivef              *
!      nactivef_n:  but limited only to                         *
!      iactive_n :                  Nbody particles             *
!                                                               *
!      nactivepm :Same as for nactive and nactivef              *
!      nactivefpm:  but limited only to                         *
!      iactivepm :                  Point mass particles        *
!                                                               *
!      it0: Integer time for last completed full timestep       *
!      it1: Integer time for next 1/2 timestep.                 *
!      it0pm: Integer time for last completed full timestep (PM)*
!      it1pm: Integer time for next 1/2 timestep. (PM)          *
!                                                               *
!      itstepbin  :Integer size of timestep.                    *
!      itstepbinpm:Integer size of timestep for PMs.            *
!                  Calculate the real timestep from itstepbin   *
!                  as dt=dtincr*itstepbin                       *
!                                                               *
!      imaxstep :the maximum integer  timestep value            *
!      dtmax    :the maximum physical timestep value            *
!      dtmin    :the minimum physical timestep value            *
!      dtincr   :The smallest timestep increment allowed        *
!                dtincr=dtmax/dble(imaxstep),                   *
!                This is 1/2 of the minimum time step, to allow *
!                for integrators on half or whole update steps  *
!                with itstepbin=1 or 2                          *
!                for global, non-binned timesteps               *
!                dtincr=half*deltnew                            *
!                                                               *
!      deltat   :Time from (or sometimes until-for offset       *
!                integrators like leapfrog) particles last      *
!                completed full timestep to the current time.   *
!                deltat=dtincr*dble(now-it0)                    *
!      deltatpm :Same as deltat, for point mass particles       *
!                                                               *
!===============================================================*
!      NOTE : deltat is only related to the particle's          *
!            current timestep if the particle is active!        *
!===============================================================*
!      NOTE2: The deltat calculated by initdt is different      *
!            from these definitions for a onestep offset        *
!            integrator, in order to set up the initial offset  *
!            calculation from the saved conditions.             *
!===============================================================*
!      NOTE3: This code should be replaced with some kind of    *
!             sort based algorithm. It would go faster and      *
!             be more parallel. -AFN                            *
!===============================================================*
!  Written by Andy Nelson, Markus Wetzstein and Thorsten Naab   *
!                                                               *
!****************************************************************

      use dimens
      use indivtsvars
      use miscvars, only: iusebin,indts,igrape,iexpand,nstep
      use miscvars, only: t,t_endts,deltnew,deltold,dtinit
      use miscvars, only: hnow,ascale,dtmax,dtmin,t_unit
      use perform
      use constants

      implicit none

      private
      public initstepping, initdt, init_active
      public getdt, setnewdt
      public getactive, getnext
      public update_stepping,update_time
      public restore_activpm,save_activpm 
      public report

      contains
!----------------------------------------------------------------
      subroutine init_active(npart,npart_sph,npoim)
!****************************************************************
!                                                               *
! Turn on all particles as active particles that require        *
! derivative calculations (i.e. either a full or half timestep  *
! update).                                                      *
!                                                               *
! Returns: nactive*, iactive*                                   *
!                                                               *
!****************************************************************
      integer, intent(in):: npart,npart_sph,npoim

      integer ifh

      double precision tme1,tme2
      character(len=20), parameter:: where = ('allactive')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

!
! For init_active, we either don't care whether we are half or
! full active, so choose half active (ifh=0) because we expect 
! to be extrapolating from some initial condition to a 1/2 step.
!
      ifh=0

      call activ_glob(ifh,npart,npart_sph,npoim)

      if(itme.ge.1)then
         call cpu_time(tme2)
         nactiveftot(1)=nactiveftot(1) + nactivef_s
         nactiveftot(2)=nactiveftot(2) + nactivef_n
         nactivetot (1)=nactivetot (1) + nactive_s
         nactivetot (2)=nactivetot (2) + nactive_n
         timing(itme_getactiv)=timing(itme_getactiv)+tme2-tme1
         ncalls(itme_getactiv)=ncalls(itme_getactiv)+1
         timing(itme_indts)=timing(itme_indts)+tme2-tme1
!         ncalls(itme_indts) increment only in setnewdt
      endif

      return
      end subroutine init_active
!----------------------------------------------------------------
      subroutine getactive(ifh,npart,npart_sph,npoim)
!****************************************************************
!                                                               *
! get lists of clumps and particles which require derivative    *
! calculations (i.e. either a full or half timestep update)     *
!                                                               *
! Returns: nactive*, iactive*                                   *
!                                                               *
!****************************************************************
      integer, intent(in):: ifh,npart,npart_sph,npoim

      double precision tme1,tme2
      character(len=20), parameter:: where = ('getactive')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

      if(indts.eq.1)then
         call activ_indt(npart,npart_sph,npoim)
      else
         call activ_glob(ifh,npart,npart_sph,npoim)
      endif

      if(itme.ge.1)then
         call cpu_time(tme2)
         nactiveftot(1)=nactiveftot(1) + nactivef_s
         nactiveftot(2)=nactiveftot(2) + nactivef_n
         nactiveftot(3)=nactiveftot(3) + nactivefpm
         nactivetot (1)=nactivetot (1) + nactive_s
         nactivetot (2)=nactivetot (2) + nactive_n
         nactivetot (3)=nactivetot (3) + nactivepm
         timing(itme_getactiv)=timing(itme_getactiv)+tme2-tme1
         ncalls(itme_getactiv)=ncalls(itme_getactiv)+1
         timing(itme_indts)=timing(itme_indts)+tme2-tme1
!         ncalls(itme_indts) increment only in setnewdt
      endif

      return
      end subroutine getactive
!----------------------------------------------------------------
      subroutine initdt(npart,npoim)
!****************************************************************
!                                                               *
! Initialize the integer time it0 and it1 for each particle     *
!  and the timestep deltnew if global unbinned timestep         *
!                                                               *
! Return:                                                       *
!        it0, it1, it0pm, it1pm, itstepbin, itstepbinpm,        *
!        dtincr, deltnew, deltold, deltat                       *
!                                                               *
!****************************************************************
      integer, intent(in):: npart,npoim
      integer max_idt,min_idt,max_idtpm,min_idtpm,i

      character(len=20), parameter:: where = ('initdt')
      if(idbg.ge.2)print *,where

      if(indts.eq.1)then
         call initbin_indt(npart,npoim)
      else
         call initbin_glob(npart,npoim)
      endif

      min_idt=imaxstep
      max_idt=0

!$omp parallel do default(none) private(i)
!$omp& firstprivate(dtincr)
!$omp& shared(npart,itstepbin,it0,it1,deltat)
!$omp& reduction(min:min_idt) reduction(max:max_idt)
      do i=1,npart
         it0(i)=0
         it1(i)=itstepbin(i)/2
         max_idt=max(max_idt,itstepbin(i))
         min_idt=min(min_idt,itstepbin(i))
         deltat(i)=dtincr*dble(it1(i))
      enddo

      min_idtpm=imaxstep
      max_idtpm=0
      do i=1,npoim
         it0pm(i)=0
         it1pm(i)=itstepbinpm(i)/2
         max_idtpm=max(max_idtpm,itstepbinpm(i))
         min_idtpm=min(min_idtpm,itstepbinpm(i))
         deltatpm(i)=dtincr*dble(it1pm(i))
      enddo


      levlocc_bigdt=1
      do while(2**(ntimelevs-levlocc_bigdt).gt.max_idt)
         levlocc_bigdt=levlocc_bigdt+1
      enddo

      levlocc_smldt=1
      do while(2**(ntimelevs-levlocc_smldt).gt.min_idt)
         levlocc_smldt=levlocc_smldt+1
      enddo

      levldone_bigdt=levlocc_bigdt  !for init only
      levldone_smldt=levlocc_smldt  !for init only

      lastts_end=0

      return
      end subroutine initdt
!----------------------------------------------------------------
      subroutine setnewdt(npart,npoim,success)
!****************************************************************
!                                                               *
! Determine the timestep for each particle for the              *
! next timestep, using the minimum of new/old dt ratio          *
! for each particle. Put into appropriate bins, if              *
! necessary.                                                    *
!                                                               *
! Return:                                                       *
!     if indts                                                  *
!        itstepbin  (can't know deltnew,deltold so set to huge) *
!     if not indts                                              *
!        if iusebin                                             *
!           itstepbin,deltnew,deltold                           *
!        if not iusebin                                         *
!           itstepbin(=2),deltnew,deltold and dtincr            *
!                                                               *
!****************************************************************
      integer, intent(in):: npart,npoim
      logical, intent(out):: success

      double precision tme1,tme2
      character(len=20), parameter:: where = ('setnewdt')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)
!
! get level of timestep scheme we just worked on. Needed
! for updating time, and must be done before updating 
! the time*step*.
!
      if(indts.eq.1 .or. iusebin.eq.1)call getlevl(npart,npoim)

      if(indts.eq.1)then
         call newdt_indt
         success=.true.
      else
         call newdt_glob(npart,npoim,success)
      endif

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_setnewdt)=timing(itme_setnewdt) + tme2-tme1
         ncalls(itme_setnewdt)=ncalls(itme_setnewdt) + 1
         timing(itme_indts)=timing(itme_indts) + tme2-tme1
         ncalls(itme_indts)=ncalls(itme_indts) + 1
      endif

      return
      end subroutine setnewdt
!----------------------------------------------------------------
      subroutine init_min_dtbin

      integer ibin,istep
      double precision dtmin_use

      if(iusebin.eq.0)then

         min_dtbin=1
      else

         ibin=int(log10(dtmax/max(TWO*dtincr,dtmin))*dlog2inv+ 1.001d0) 

         if(ibin.lt.1)stop 'Error with initial minimum timestep'

         min_dtbin= ibin
         istep=imaxstep/2**(min_dtbin-1)
         dtmin_use= t_unit*dtincr*dble(istep)

         print *,'========================================='
         print *,' '
         print *,' '
         print *,'Minimum time step bin is   : ',min_dtbin
         print *,'Corresponding to time step : ',dtmin_use
         print *,' '
         print *,' '
         print *,'========================================='

      endif

      return
      end subroutine init_min_dtbin
!----------------------------------------------------------------
      subroutine initstepping(istepmin)
!****************************************************************
! Get initial integer timestep value, istepmin                  *
!****************************************************************
      integer, intent(out)::istepmin
      integer ibin
      double precision dtmin_use

      if(dtmax.lt.dtinit)then
         print *,'=================='
         print *,'dtinit is too big '
         print *,'resetting to dtmax'
         print *,'=================='
         dtinit=dtmax
      endif
      if(dtmin.gt.dtinit)then
         print *,'==================='
         print *,'dtinit is too small'
         print *,'resetting to dtmin '
         print *,'==================='
         dtinit=dtmin
      endif

      ibin = int(log10(dtmax/dtinit)*dlog2inv + 1.001d0)
      ibin = min(ibin,min_dtbin)

      if(ibin.lt.0)stop 'Error with initial timesteps'

      istepmin = imaxstep/2**(ibin-1)
      dtmin_use= dtincr*dble(istepmin)

      print *,'========================================='
      print *,' '
      print *,' '
      print *,'        dtmax is:',dtmax*t_unit
      print *,'        dtmin is:',dtmin*t_unit
      print *,'    Using dtinit:',dtinit*t_unit
      print *,'     Yields ibin:',ibin
      print *,'So integer ts is:',istepmin 
      print *,'  and real ts is:',dtmin_use*t_unit
      print *,' '
      print *,' '
      print *,'========================================='

      return
      end subroutine initstepping
!----------------------------------------------------------------
      subroutine update_stepping(i12step)
!****************************************************************
! Update the integer times after completing                     *
!  a full or 1/2 timestep                                       *
!****************************************************************
      integer, intent(in):: i12step
      integer i,j

      double precision tme1,tme2
      character(len=20), parameter:: where = ('update_stepping')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)


      if(i12step.eq.1)then
!=====================================================
! Appropriate for single step integrator like leapfrog
!=====================================================

!$omp parallel do if(nactive.gt.1024) default(none) private(i,j)
!$omp& firstprivate(lastts_end)
!$omp& shared(nactive,iactive,it0,it1,itstepbin)
         do j=1,nactive
            i=iactive(j)
            it0(i)=lastts_end
            it1(i)=it0(i) + itstepbin(i)/2
         enddo

         do j=1,nactivepm
            i=iactivepm(j)
            it0pm(i)=lastts_end
            it1pm(i)=lastts_end + itstepbinpm(i)/2
         enddo

      elseif(i12step.eq.2)then
!=====================================================
! Appropriate for two step integrator like Runge Kutta
!=====================================================

! particles at the end of a timestep need new it0, it1
! but particles at the middle of a timestep only need it1
! out of the way

!$omp parallel do if(nactivef.gt.1024) default(none) private(i,j)
!$omp& firstprivate(lastts_end)
!$omp& shared(nactivef,iactive,it0,it1,itstepbin)
         do j=1,nactivef
            i=iactive(j)
            it0(i)=lastts_end
            it1(i)=it0(i) + itstepbin(i)/2
         enddo

         do j=1,nactivefpm
            i=iactivepm(j)
            it0pm(i)=lastts_end
            it1pm(i)=it0pm(i) + itstepbinpm(i)/2
         enddo

!$omp parallel do if(nactive-nactivef.gt.1024) default(none)
!$omp&  private(j,i) shared(nactivef,nactive,iactive,it1)
         do j=nactivef+1,nactive
            i=iactive(j)
            it1(i)=imax
         enddo

         do j=nactivefpm+1,nactivepm
            i=iactivepm(j)
            it1pm(i)=imax
         enddo

      else
         stop'update_stepping:don''t know integrator stepping'
      endif

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_updtstep)=timing(itme_updtstep) + tme2-tme1
         ncalls(itme_updtstep)=ncalls(itme_updtstep) + 1
         timing(itme_indts)=timing(itme_indts) + tme2-tme1
!         ncalls(itme_indts)=ncalls(itme_indts) + 1
      endif

      return
      end subroutine update_stepping
!----------------------------------------------------------------
      subroutine update_time(npart,npoim,donefullstep)
!********************************************************************
! Update the time if neccessary after completing a full timestep    *
! Time update is not done for indts.eq.1 until the end of a         *
! full dtmax step.                                                  *
!                                                                   *
! Also, reset the steps it0, it1 to values less than imaxstep.      *
!                                                                   *
!********************************************************************
      integer, intent(in ):: npart,npoim
      logical, intent(out):: donefullstep
      integer i

      double precision tme1,tme2
      if(itme.ge.1)call cpu_time(tme1)

      if(indts.eq.0 .and. iusebin.eq.0)then  !global unbinned ts
         t      = t + deltold
         t_endts= t
         donefullstep=.true.
         call printtime(npart,npoim)
      else                                   !global binned and indts

         t_endts= t + dble(lastts_end)*dtincr

         if(lastts_end.ge.imaxstep .and. 
     &               levldone_smldt.eq.levlocc_smldt)then
         
            call report(npart,npoim)
            if(itme.gt.0)call perfsummary(npart)

            t=t+dtmax
            donefullstep=.true.
!$omp parallel do default(none) private(i) shared(npart,it0,it1)
            do i=1,npart
               it0(i)=it0(i)-imaxstep
               it1(i)=it1(i)-imaxstep
            enddo
            do i=1,npoim
               it0pm(i)=it0pm(i)-imaxstep
               it1pm(i)=it1pm(i)-imaxstep
            enddo
            lastts_end=lastts_end - imaxstep

         else
            if(idbg.ge.2)then
               call report(npart,npoim)
            else
               call printtime(npart,npoim)
            endif
            donefullstep=.false.
         endif
      endif



      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_updttime)=timing(itme_updttime) + tme2-tme1
         ncalls(itme_updttime)=ncalls(itme_updttime) + 1
         timing(itme_indts)=timing(itme_indts) + tme2-tme1
!         ncalls(itme_indts)=ncalls(itme_indts) + 1
      endif

      return
      end subroutine update_time
!----------------------------------------------------------------
      subroutine getnext(npart,npoim,nexthalf,nextfull)
!****************************************************************
!                                                               *
! get the next integer time for which particles                 *
!    need to be updated                                         *
!                                                               *
!****************************************************************

      integer, intent(in ):: npart,npoim
      integer, intent(out):: nexthalf,nextfull

      double precision dtval
      integer ibin,j

      integer i
      double precision tme1,tme2
      character(len=20), parameter:: where = ('getnext')
      if(idbg.ge.4)print *,where
      if(itme.ge.1)call cpu_time(tme1)

!---------------------------
! Global binned or individual timesteps
!---------------------------
      nexthalf = imax
      nextfull = imax
      
!$omp parallel do default(none) private(i)
!$omp& shared(it1,it0,itstepbin,npart)
!$omp& reduction(min:nexthalf,nextfull)
      do i=1,npart
         nexthalf=min(nexthalf,it1(i))
         nextfull=min(nextfull,it0(i)+itstepbin(i))
      enddo

      do i=1,npoim
         nexthalf=min(nexthalf,it1pm(i))
         nextfull=min(nextfull,it0pm(i)+itstepbinpm(i))
      enddo

      now=min(nexthalf,nextfull)

      call getdt(npart,npoim,now)

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_getnext)=timing(itme_getnext) + tme2-tme1
         ncalls(itme_getnext)=ncalls(itme_getnext) + 1
         timing(itme_indts  )=timing(itme_indts  ) + tme2-tme1
!         ncalls(itme_indts) only increment in setnewdt
      endif

      return
      end subroutine getnext
!----------------------------------------------------------------
      subroutine getdt(npart,npoim,n)
!****************************************************************
!                                                               *
!  calculate deltat for particles and point masses,             *
!  relative to the integer time n                               *
!                                                               *
!****************************************************************

      integer, intent(in):: npart,npoim,n
      integer i

!$omp parallel do default(none) private(i)
!$omp& firstprivate(dtincr)
!$omp& shared(n,deltat,it0,npart)
      do i=1,npart
         deltat(i)=dtincr*dble(n-it0(i))
      enddo

      do i=1,npoim
         deltatpm(i)=dtincr*dble(n-it0pm(i))
      enddo

      return
      end subroutine getdt
!****************************************************************
! Below are the versions of several routines for using          *
! individual timesteps.                                         *
!****************************************************************
!----------------------------------------------------------------
      subroutine initbin_indt(npart,npoim)
!****************************************************************
!  Individual timestep init routine                             *
!                                                               *
!  Sets min_dtbin,deltnew,deltold,itstepbin,itstepbinpm         *
!****************************************************************
      integer, intent(in):: npart,npoim
      integer i,istepmin


      dtincr=dtmax/dble(imaxstep)
      TWOdtincr=TWO*dtincr

      call init_min_dtbin

      istepmin = imax
!$omp parallel do default(none) 
!$omp&   shared(npart,itstepbin) private(i)
!$omp&   reduction(min:istepmin)
      do i=1,npart
         istepmin=min(istepmin,itstepbin(i))
      enddo
      do i=1,npoim
         istepmin=min(istepmin,itstepbinpm(i))
      enddo

      if(t.eq.zero .or. istepmin.eq.0)then    !define itstepbin 

         call initstepping(istepmin)
         deltnew=dtincr*dble(istepmin)
         deltold=deltnew

!$omp parallel do default(none) private(i)
!$omp&  shared(itstepbin,istepmin,npart)
         do i=1,npart
            itstepbin(i)=istepmin
         enddo
         do i=1,npoim
            itstepbinpm(i)=istepmin
         enddo

      else                           !itstepbin already stored

         deltnew=dtincr*dble(istepmin)
         deltold=deltnew

      endif

      return
      end subroutine initbin_indt
!----------------------------------------------------------------
      subroutine getlevl(npart,npoim)
!****************************************************************
!                                                               *
! find 1) the finest timestep level                             *
!      2) the finest timestep level last worked on              *
!                                                               *
!   levlocc_* : biggest and smallest timestep bins              *
!               that are currently occupied                     *
!   levldone_*: biggest and smallest timestep bins              *
!               that were done at the last update               *
!                                                               *
! Needed for afterstep subroutine...                            *
!   are we at the end of a dtmax step?                          *
!     Yes if 1)lastts_end is sync'ed to imaxstep                *
!            2)we just worked on finest level                   *
!                                                               *
! recall for defn of levldone_smldt and levlocc_smldt that      *
! the definition of imaxstep is imaxstep= 2**(ntimelevs-1)      *
!****************************************************************

      integer, intent(in):: npart,npoim
      integer i,j,min_idt,max_idt,min_idtpm,max_idtpm
      integer loopindx,loopindxpm

      if(nactivef.eq.0 .and. nactivefpm.eq.0)then    !see newdt_indt below 
         loopindx=nactive                            !for reason why
         loopindxpm=nactivepm
      else
         loopindx=nactivef
         loopindxpm=nactivefpm
      endif
!
! Get the levels that were just completed
!
      min_idt=imaxstep
      max_idt=0
!$omp parallel do if(loopindx.gt.1024) default(none) private(i,j)
!$omp& shared(iactive,loopindx,itstepbin)
!$omp& reduction(min:min_idt) reduction(max:max_idt)
      do j=1,loopindx
         i=iactive(j)
         min_idt=min(min_idt,itstepbin(i))
         max_idt=max(max_idt,itstepbin(i))
      enddo

      do j=1,loopindxpm
         i=iactivepm(j)
         min_idt=min(min_idt,itstepbinpm(i))
         max_idt=max(max_idt,itstepbinpm(i))
      enddo

      levldone_bigdt=1
      do while(2**(ntimelevs-levldone_bigdt-1).gt.max_idt)
         levldone_bigdt=levldone_bigdt+1
      enddo
      levldone_bigdt=levldone_bigdt+1   !kludge fix for off by one error

      levldone_smldt=1
      do while(2**(ntimelevs-levldone_smldt-1).gt.min_idt)
         levldone_smldt=levldone_smldt+1
      enddo
      levldone_smldt=levldone_smldt+1   !kludge fix for off by one error
!
! Get the levels that are occupied. Separate out point masses
!
      min_idt=imaxstep
      max_idt=0
!$omp parallel do default(none) private(i) shared(npart,itstepbin)
!$omp& reduction(min:min_idt) reduction(max:max_idt)
      do i=1,npart
         min_idt=min(min_idt,itstepbin(i))
         max_idt=max(max_idt,itstepbin(i))
      enddo

      do i=1,npoim
         min_idt=min(min_idt,itstepbinpm(i))
         max_idt=max(max_idt,itstepbinpm(i))
      enddo

      levlocc_bigdt=1
      do while(2**(ntimelevs-levlocc_bigdt-1).gt.max_idt)
         levlocc_bigdt=levlocc_bigdt+1
      enddo
      levlocc_bigdt= levlocc_bigdt+1    !same here as above

      levlocc_smldt=1
      do while(2**(ntimelevs-levlocc_smldt-1).gt.min_idt)
         levlocc_smldt=levlocc_smldt+1
      enddo
      levlocc_smldt= levlocc_smldt+1    !same here as above

      return
      end subroutine getlevl
!----------------------------------------------------------------
      subroutine newdt_indt
!****************************************************************
!  Individual timestep variant of the routine to get new        *
!  timesteps.                                                   *
!                                                               *
!  We have to do two different options in this routine to       *
!  take care of offset and non-offset integrators. In other     *
!  words, leapfrog or Runge-Kutta (or some similar ones)        *
!  We only want to update timesteps for particles that actually *
!  finished their timestep just now. For non-offset integrators *
!  the list of particles that finished their timesteps is       *
!  bounded by 1..nactivef. For offset integrators the lists are *
!  bounded by 1..nactive instead.                               *
!                                                               *
!  Therefore have to make a distinction in which loop index     *
!  we iterate over.                                             *
!                                                               *
!  Also, the offset and non-offset integrators have different   *
!  requirements for what defines 'now'. In both leapfrog and    *
!  Runge-Kutta, 'now' is the time of the last derivative calc,  *
!  but that time is different. For lf, it is 1/2 step back      *
!  from the end of the ts. We can use any (here the first)      *
!  active particle to define idthalf, since in our version      *
!  of lf, only one level can be done at a time due to           *
!  synchronization issues. Then we add idthalf to get to the    *
!  end of the ts. For rk, we can just use now as it is.         *
!                                                               *
!  The reason we need lastts_end in the first place is for      *
!  setting the it1 for the next step (see update_stepping)      *
!                                                               *
!****************************************************************

      integer loopindx,loopindxpm,idthalf
      integer i,j,ibin,istep2,nbelow,nbelow_tot,nbelowpm
      double precision dt_new,dt_old

      if(nactivef.eq.0 .and. nactivefpm.eq.0)then
!
! Offset integrator version (eg leapfrog)
!
         loopindx=nactive
         loopindxpm=nactivepm
         if(nactive.gt.0)then
            idthalf=itstepbin  (iactive  (1))/2
         elseif(nactivepm.gt.0)then
            idthalf=itstepbinpm(iactivepm(1))/2
         else
            stop'newdt_indt:Unknown type of active particles'
         endif
         lastts_end = now + idthalf

      else
!
! Non-offset integrator like Runge-Kutta, don't want to
! reset guys on 1/2 steps (i.e. iactive index >nactivef)
!
         loopindx=nactivef
         loopindxpm=nactivefpm
         lastts_end = now
      endif
!
! Now set new time steps for particles.
!
      nbelow_tot=0
!$omp parallel default(none) 
!$omp& firstprivate(dtincr,TWOdtincr,lastts_end,min_dtbin,dtmin)
!$omp& private(i,j,dt_old,dt_new,ibin,istep2,nbelow)
!$omp& shared(nbelow_tot)
!$omp& shared(loopindx,iactive,itstepbin,dtrat,dtmax)

      nbelow = 0
!$omp do
      do j=1,loopindx
         i=iactive(j)
         dt_old = dble(itstepbin(i))*dtincr
         if(dtrat(i).lt. one)then          !Decrease timestep

            if(dtmin.gt.dt_old*dtrat(i))nbelow=nbelow+1
            dt_new = max(dtmin,TWOdtincr,dt_old*dtrat(i))
            ibin = int(log10(dtmax/dt_new)*dlog2inv + 1.001d0)
            itstepbin(i) = imaxstep/(2**(ibin-1))
         elseif(dtrat(i).ge.two)then       !Increase timestep
                                           !if synchronized 
            istep2 = 2*itstepbin(i)
            if(mod(lastts_end,istep2).eq.0)then
               itstepbin(i)=min(istep2,imaxstep)
            endif
         endif
      enddo
!$omp critical
      nbelow_tot=nbelow_tot + nbelow
!$omp end critical
!$omp end parallel
!
! Point masses
!
      nbelowpm=0
      do j=1,loopindxpm
         i=iactivepm(j)
         dt_old = dble(itstepbinpm(i))*dtincr
         if(dtratpm(i).lt. ONE)then          !Decrease timestep

            if(dtmin.gt.dt_old*dtratpm(i))nbelowpm=nbelowpm+1
            dt_new = max(dtmin,TWOdtincr,dt_old*dtratpm(i))
            ibin = int(log10(dtmax/dt_new)*dlog2inv) + 1
            itstepbinpm(i) = imaxstep/(2**(ibin-1))

         elseif(dtratpm(i).ge.TWO)then       !Increase timestep
                                             !if synchronized 
            istep2 = 2*itstepbinpm(i)
            if(mod(lastts_end,istep2).eq.0)then
               itstepbinpm(i)=min(istep2,imaxstep)
            endif

         endif
      enddo
!
! These values are `out of date' in the sense that they
! refer to the minimum timestep in the last timstep, not
! this one. However they are not used, except in one very
! rare circumstance where the out of date-ness doesn't matter.
! That circumstance is when a dump is made with ind ts and
! then restarted with global timsteps, the code needs a 
! reasonable value of deltold to start up with.
!
      deltnew=dtmax/(2**(levlocc_smldt-1))
      deltold=dtmax/(2**(levlocc_smldt-1))

!
! Finally, check for error/danger conditions:
!
      if(nbelow_tot.gt.0)then
         print *,'WARNING WARNING WARNING WARNING WARNING'
         print *,'Particles in your simulation need'
         print *,'time steps smaller than the minimum'
         print *,'you set. Your simulation is in danger'
         print *,'of producing nonsensical results'
         print *,'Hope you know what you are doing'
         print *,' '
         print *,'Number wanting smaller steps:',nbelow_tot
         print *,' '
      endif
      if(nbelowpm.gt.0)then
         print *,'WARNING WARNING WARNING WARNING WARNING'
         print *,'Point masses in your simulation need'
         print *,'time steps smaller than the minimum'
         print *,'you set. Your simulation is in danger'
         print *,'of producing nonsensical results'
         print *,'Hope you know what you are doing'
         print *,' '
         print *,'Number wanting smaller steps:',nbelowpm
         print *,' '
      endif

      return
      end subroutine newdt_indt
!----------------------------------------------------------------
      subroutine activ_indt(npart,npart_sph,npoim)
!******************************************************************
!                                                                 *
!  Determine whether each particle needs an update, doing         *
!  a search through the entire list of particles, to determine    *
!  activity.                                                      *
!  Count the numbers of actives, half actives, and inactives      *
!  and place on appropriate lists for update.                     *
!  Fully active particles go first on each list,                  *
!  Followed by half actives, then inactives.                      *
!  On each sublist, sph  particles are first,                     *
!              then nbody particles                               *
!                                                                 *
!  In order to do this routine in parallel, we have to            *
!  calculate some offsets from the beginning of each sublist      *
!  for each processor, in order to know where to start putting    *
!  its portion of the active, half-active and inactive particle   *
!  indices. Define nf_*, nh_* and ni_* for this purpose.          *
!  We need these offsets                                          *
!                                                                 *
! iactive needs 6 locations. Immediately before:                  *
!    1) first fully active sph  : 0                       + nf_s  *
!    2) first fully active nbody: nactivef_s              + nf_n  *
!    3) first half  active sph  : nactivef                + nh_s  *
!    4) first half  active nbody: nactivef   + nacthalf_s + nh_n  *
!    5) first idle         sph  : nactive                 + ni_s  *
!    6) first idle         nbody: nactive    + nidle_s    + ni_n  *
! 3 each for iactive_[s,n]                                        *
!    1) First fully active particle: 0              + nf_[s,n]    *
!    2) First half  active particle: nactivef_[s,h] + nh_[s,n]    *
!    3) First idle         particle: nactive_[s,n]  + ni_[s,n]    *
!                                                                 *
!******************************************************************

      integer, intent(in):: npart,npart_sph,npoim

      integer my_cpu,numcpus,icpu

      integer nf_n,nh_n

      integer i
      integer nidle_s,nacthalf_s
      integer nf,nh,ni
      integer nperproc_sph ,limlo_sph ,limhi_sph ,nah_s,naf_s,ni_s,ns
      integer nperproc_nbod,limlo_nbod,limhi_nbod,nah_n,naf_n,ni_n,nn

      integer, parameter:: id_nfs =1,id_nhs =2, id_nis=3
      integer, parameter:: id_nfn =4,id_nhn =5, id_nin=6
      integer ncounts(6,maxcpus)


      nactivef_s=0 ; nactivef_n=0 ; nactivef =0 ; nactivefpm=0
      nactive_s =0 ; nactive_n =0 ; nactive  =0 ; nactivepm =0

!$omp parallel default(none) 
!$omp&  firstprivate(now)
!$omp&  shared(npart,npart_sph,it0,it1,itstepbin)
!$omp&  shared(nactive  ,nactivef  ,iactive  )
!$omp&  shared(nactive_s,nactivef_s,iactive_s)
!$omp&  shared(nactive_n,nactivef_n,iactive_n)
!$omp&  shared(ncounts)
!$omp&  private(my_cpu,numcpus,icpu)
!$omp&  private(i,nperproc_sph,limlo_sph,limhi_sph ,nah_s,naf_s,ni_s,ns)
!$omp&  private(nperproc_nbod,limlo_nbod,limhi_nbod,nah_n,naf_n,ni_n,nn)
!$omp&  private(nf,nh,ni,nidle_s,nacthalf_s)

      call get_mycpu(my_cpu)
      call numthreads(numcpus)
!
! Divide up work among processors
!
      nperproc_sph =npart_sph/numcpus
      nperproc_nbod=(npart-npart_sph)/numcpus

      limlo_sph= 1 + nperproc_sph*(my_cpu-1)
      limhi_sph=     nperproc_sph*my_cpu

      limlo_nbod= npart_sph + 1 + nperproc_nbod*(my_cpu-1)
      limhi_nbod= npart_sph +     nperproc_nbod*my_cpu
      if(my_cpu.eq.numcpus)then
         limhi_sph =npart_sph
         limhi_nbod=npart
      endif
!
! Examine each particle for activity status
!
!    SPH particles
!
      nah_s=0 ; naf_s=0

      do i=limlo_sph,limhi_sph
         if(it1(i).eq.now)then                     ! half timestep
            nah_s  = nah_s  + 1
         elseif(it0(i)+itstepbin(i).eq.now)then    ! full timestep
            naf_s  = naf_s  + 1
         endif
      enddo

      ns= limhi_sph  - limlo_sph  + 1
      ncounts(id_nfs,my_cpu)=naf_s
      ncounts(id_nhs,my_cpu)=nah_s
      ncounts(id_nis,my_cpu)=ns - naf_s - nah_s
!
!    Nbody particles
!
      nah_n=0 ; naf_n=0

      do i=limlo_nbod,limhi_nbod
         if(it1(i).eq.now)then                     ! half timestep
            nah_n  = nah_n  + 1
         elseif(it0(i)+itstepbin(i).eq.now)then    ! full timestep
            naf_n  = naf_n  + 1
         endif
      enddo

      nn= limhi_nbod - limlo_nbod + 1
      ncounts(id_nfn,my_cpu)=naf_n
      ncounts(id_nhn,my_cpu)=nah_n
      ncounts(id_nin,my_cpu)=nn - naf_n - nah_n

!$omp critical
      nactivef_s = nactivef_s + naf_s
      nactivef_n = nactivef_n + naf_n
      nactivef   = nactivef   + naf_s + naf_n
      nactive_s  = nactive_s  + naf_s + nah_s
      nactive_n  = nactive_n  + naf_n + nah_n
      nactive    = nactive    + nah_s + naf_s + naf_n + nah_n
!$omp end critical

!$omp barrier
!
! Compute list offsets for each processor. 

      naf_n=0 ; nah_n= 0 ; ni_n= 0
      naf_s=0 ; nah_s= 0 ; ni_s= 0

      do icpu=1,my_cpu-1
         naf_n= naf_n + ncounts(id_nfn,icpu)
         naf_s= naf_s + ncounts(id_nfs,icpu)

         nah_n= nah_n + ncounts(id_nhn,icpu)
         nah_s= nah_s + ncounts(id_nhs,icpu)

         ni_n= ni_n + ncounts(id_nin,icpu)
         ni_s= ni_s + ncounts(id_nis,icpu)
      enddo

!iactive and iactive_s arrays (set up for SPH particles)

      nf = naf_s             ; naf_s = naf_s
      nh = nah_s + nactivef  ; nah_s = nah_s + nactivef_s
      ni = ni_s  + nactive   ; ni_s  = ni_s  + nactive_s

!
! Now sort onto appropriate lists.
!
      do i=limlo_sph,limhi_sph

         if(it1(i).eq.now)then
            nah_s=nah_s+1      ; nh=nh+1
            iactive_s(nah_s)=i ; iactive(nh)=i
         elseif(it0(i)+itstepbin(i).eq.now)then
            naf_s=naf_s+1      ; nf=nf+1
            iactive_s(naf_s)=i ; iactive(nf)=i
         else
            ni_s=ni_s+1        ; ni=ni+1
            iactive_s(ni_s )=i ; iactive(ni)=i
         endif
      enddo

!iactive and iactive_n arrays (set up for Nbody particles)

      nidle_s   = npart_sph-nactive_s
      nacthalf_s= nactive_s - nactivef_s 

      nf = naf_n + nactivef_s              ; naf_n = naf_n
      nh = nah_n + nactivef   + nacthalf_s ; nah_n = nactivef_n + nah_n
      ni = ni_n  + nactive    + nidle_s    ; ni_n  = nactive_n  + ni_n

      do i=limlo_nbod,limhi_nbod

         if(it1(i).eq.now)then
            nah_n=nah_n+1      ; nh=nh+1 
            iactive_n(nah_n)=i ; iactive(nh)=i
         elseif(it0(i)+itstepbin(i).eq.now)then
            naf_n=naf_n+1      ; nf=nf+1 
            iactive_n(naf_n)=i ; iactive(nf)=i
         else
            ni_n=ni_n+1        ; ni=ni+1
            iactive_n(ni_n )=i ; iactive(ni)=i
         endif
      enddo

!$omp end parallel

!
! Do the same for point masses, if any
!
      if(npoim.eq.0)return

      do i = 1, npoim
         if(it0pm(i)+itstepbinpm(i).eq.now)then
            nactivefpm = nactivefpm + 1
            nactivepm  = nactivepm  + 1
         elseif(it1pm(i).eq.now)then
            nactivepm  = nactivepm  + 1
         endif
      enddo

      nf  =0 ; nh  =0 ; ni  =0
      do i=1,npoim
         if(it0pm(i)+itstepbinpm(i).eq.now)then
            nf=nf+1
            iactivepm(nf)=i
         elseif(it1pm(i).eq.now)then
            nh=nh+1
            iactivepm(nactivefpm+nh)=i
         else
            ni=ni+1
            iactivepm(nactivepm+ni)=i
         endif
      enddo


      return
      end subroutine activ_indt
!----------------------------------------------------------------
      subroutine restore_activpm(npoim)
!****************************************************************
!  There are situations wehre we need to save the list of       *
!  active point masses, then restore it. This routine does the  *
!  restore.                                                     *
!****************************************************************

      integer, intent(in):: npoim
      integer i

      if(npoim.eq.0)return
      do i=1,npoim
         iactivepm(i)=iactivepmsav(i)
      enddo
      nactivepm =nactivepmsav
      nactivefpm=nactivefpmsav

      return
      end subroutine restore_activpm
!----------------------------------------------------------------
      subroutine save_activpm(npoim)
!****************************************************************
!  There are situations wehre we need to save the list of       *
!  active point masses, then restore it. This routine does the  *
!  save.                                                        *
!****************************************************************

      integer, intent(in):: npoim
      integer i

      if(npoim.eq.0)return
      do i=1,npoim
         iactivepmsav(i)=iactivepm(i)
      enddo
      nactivepmsav =nactivepm
      nactivefpmsav=nactivefpm

      return
      end subroutine save_activpm
!----------------------------------------------------------------

!****************************************************************
! Below are the versions of several routines for global         *
! timesteps.                                                    *
!****************************************************************

!----------------------------------------------------------------
      subroutine initbin_glob(npart,npoim)
!****************************************************************
!                                                               *
!  Initialize timestep information for global timesteps, both   *
!  binned and unbinned variants.                                *
!                                                               *
!  Sets min_dtbin,deltnew,deltold,itstepbin,itstepbinpm         *
!                                                               *
!****************************************************************
      integer, intent(in):: npart,npoim

      integer i,istepmin


      if(iusebin.eq.0)then            !global unbinned

         if(t.eq.zero)then            !initial timestep: assume dtinit given 
                                      !deltnew is not available 
            deltnew=max(dtmin,min(dtinit,dtmax))
         endif

         istepmin =2
         dtincr    = deltnew/dble(istepmin)
         TWOdtincr = TWO*dtincr
         call init_min_dtbin

      else

         dtincr=dtmax/dble(imaxstep)
         TWOdtincr = TWO*dtincr
         call init_min_dtbin

         if(t.eq.zero)then            !initial timestep, assume dtmax given 

            call initstepping(istepmin)

         else                         !use stored timestep information

            istepmin=imax
!$omp parallel do default(none) private(i)
!$omp&  shared(npart,itstepbin) reduction(min:istepmin)
            do i=1,npart
               istepmin=min(istepmin,itstepbin(i))
            enddo
            do i=1,npoim
               istepmin=min(istepmin,itstepbinpm(i))
            enddo

            if(istepmin.eq.0)call initstepping(istepmin)

         endif

         deltnew=dtincr*dble(istepmin)
      endif
      deltold=deltnew

!$omp parallel do default(none) private(i)
!$omp& firstprivate(istepmin)
!$omp&  shared(npart,itstepbin)
      do i=1,npart
         itstepbin(i)=istepmin
      enddo
      do i=1,npoim
         itstepbinpm(i)=istepmin
      enddo

      return
      end subroutine initbin_glob

!----------------------------------------------------------------
      subroutine newdt_glob(npart,npoim,success)
!****************************************************************
!                                                               *
! GLOBAL TIMESTEP OPTION                                        *
! set new timestep and determine previous ts success or failure * 
!                                                               *
!  Two variants here.                                           *
!     'Continuous' timesteps:                                   *
!        set largest possible deltnew                           *
!                                                               *
!     Binned timesteps:                                         *
!        get new ts using the same method as for indiv          *
!        timesteps                                              *
!        Note: .5<dtratmin<1 will keep old bin, since you       *
!          need a full factor two decrease in ts (dtratmin) to  *
!          get smaller bin (This behavior is similar to         *
!          binned individual timesteps in that the same bin     *
!          is chosen for the critical particle in either case)  *
!                                                               *
!****************************************************************

      integer, intent(in) :: npart,npoim
      logical, intent(out):: success
      integer i,idthalf,istepmin,irat,newbin,newstep
      integer istep2
      double precision dtratmin,try_dtnew

      dtratmin=huge(ONE)
!$omp parallel do default(none) private(i)
!$omp&  shared(dtrat,npart) reduction(min:dtratmin)
      do i=1,npart
         dtratmin=min(dtratmin,dtrat(i))
      enddo
      do i=1,npoim
         dtratmin=min(dtratmin,dtratpm(i))
      enddo
!
! update synchronization time, if the ts succeeded
!
      success=.true.
      if(dtratmin.lt.half)then
         success=.false.
         print *,'FAILED TIME STEP, dtratmin=',dtratmin
         if(dtratmin.lt.1e-4)then
            print *,'Check to see if your unit (iunits) selection'
            print *,'is set correctly. dtratmin < 1e-4 is ' 
            print *,'pretty extreme!'
         endif
      else
         if(nactivef.eq.0 .and.nactivefpm.eq.0)then
            idthalf=itstepbin(iactive(1))/2
            lastts_end = now + idthalf
         else
            lastts_end = now
         endif
         if(iusebin.eq.0)lastts_end=0
      endif

      if(dtratmin*deltnew.lt.dtmin)then
         print *,'You are attempting to run your simulation'
         print *,'with a minimum time step that is larger'
         print *,'than some particles need for accurate'
         print *,'integration of their trajectories'
         print *,'You must set dtmin smaller or relax your'
         print *,'error criteria'
         print *,'dtmin=',dtmin*t_unit
         print *,'Next desired time step:',dtratmin*deltnew*t_unit
         stop 'Inconsistent dtmin and error criteria'
      endif
!
! Now get the new dt
!
      if(iusebin.eq.0)then    !Using unbinned timesteps

         deltold=deltnew
         deltnew=max(dtmin,min(dtmax,min(1.5d0,dtratmin)*deltnew))
         istepmin=2
         dtincr=deltnew/dble(istepmin)
         TWOdtincr = TWO*dtincr

      else                    !Using binned timesteps

         istepmin=itstepbin(1)
         irat=imaxstep/istepmin
         deltold= dtmax/dble(irat)

         try_dtnew= max(dtmin,deltold*dtratmin)
         newbin = max(1,int(log10(dtmax/try_dtnew)*dlog2inv+1.001d0))
         newstep= imaxstep/(2**(newbin-1))

         if(dtratmin.gt.TWO)then
            istep2= 2*istepmin
            if(mod(lastts_end,istep2).eq.0)istepmin=istep2

         elseif(dtratmin.lt.ONE)then
            istepmin=newstep
         endif
         istepmin=min(istepmin,imaxstep)

         irat=imaxstep/istepmin
         deltnew= dtmax/dble(irat)
!
! Store istepmin in itstepbin/pm
!
!$omp parallel do default(none)
!$omp& shared(npart,itstepbin)
!$omp& firstprivate(istepmin)
!$omp& private(i) 
         do i=1,npart
            itstepbin(i) = istepmin
         enddo

         do i=1,npoim
            itstepbinpm(i) = istepmin
         enddo

      endif

      return
      end subroutine newdt_glob
!----------------------------------------------------------------
      subroutine activ_glob(ifh,npart,npart_sph,npoim)
!****************************************************************
! get lists of particles which require derivative               *
! calculations (i.e. either a full or half timestep update).    *
!                                                               *
! No individual timesteps: all particles are active             *
!                                                               *
! Returns everything via module indivtsvars                     *
!     (everything= nactive* and iactive*)                       *
!****************************************************************
      integer, intent(in)::ifh,npart,npart_sph,npoim
      integer j

!$omp parallel default(none) private(j)
!$omp& shared(npart,npart_sph,iactive,iactive_s,iactive_n)
!$omp do 
      do j=1,npart_sph
         iactive  (j)=j
         iactive_s(j)=j
      enddo
!$omp do
      do j=1,npart-npart_sph 
         iactive  (j+npart_sph)=j+npart_sph
         iactive_n(j          )=j+npart_sph
      enddo
!$omp end parallel

      do j=1,npoim
         iactivepm(j)=j
      enddo

      if(ifh.eq.1)then            !full time update

         nactivef_s = npart_sph
         nactivef_n = npart-npart_sph
         nactivef   = npart
         nactivefpm = npoim

      else                         !half time update

         nactivef_s = 0
         nactivef_n = 0
         nactivef   = 0
         nactivefpm = 0

      endif

      nactive_s  = npart_sph
      nactive_n  = npart-npart_sph
      nactive    = npart
      nactivepm  = npoim

      return
      end subroutine activ_glob
!----------------------------------------------------------------
      subroutine report(npart,npoim)
!****************************************************************
! Report number of particles with each timestep                 *
!                                                               *
! Note that due to the order of integrating each level in       *
! leapfrog, the `Current Time' doesn't always increment         *
! smoothly and sometimes appears to decrement, since the        *
! end of a timestep on a high level might be after the end      *
! of a timestep on a finer level                                *
!                                                               *
!****************************************************************
      integer, intent(in)::npart,npoim

      integer ndistrtemp(ntimelevs)
      integer ntstepdistr(ntimelevs)
      integer ntstepdistrpm(ntimelevs)
      integer j,ibin,iexit
      double precision dtval,percent

      character(len=20), parameter:: where = ('report')
      if(idbg.ge.2)print *,where

      if(idbg.ge.1)then

         call printtime(npart,npoim)

         if(indts.eq.0 .and. iusebin.eq.0)return

         do j=1,ntimelevs
            ndistrtemp   (j) = 0
            ntstepdistr  (j) = 0
            ntstepdistrpm(j) = 0
         enddo
         do j=1,npoim
            dtval = itstepbinpm(j)*dtincr
            ibin = int((log10(dtmax/dtval)*dlog2inv) + 1.001d0)
            if((ibin.ge.1).and.(ibin.lt.ntimelevs))then
               ntstepdistrpm(ibin) = ntstepdistrpm(ibin) + 1
            else
               ntstepdistrpm(ntimelevs) = ntstepdistrpm(ntimelevs) + 1
               print *,ibin,itstepbinpm(j),dtmax,dtval
            endif
         enddo

         iexit=0
!$omp parallel default(none)
!$omp&    firstprivate(ndistrtemp,dtmax,dtincr)
!$omp&    private(j,dtval,ibin)
!$omp&    shared(npart,iexit,itstepbin,ntstepdistr)
!$omp do
         do j=1,npart
            dtval = itstepbin(j)*dtincr
            ibin = int((log10(dtmax/dtval)*dlog2inv) + 1.001d0)
            if(ibin.ge.1 .and. ibin.lt.ntimelevs)then
               ndistrtemp(ibin) = ndistrtemp(ibin) + 1
            else
               iexit=1
               ndistrtemp(ntimelevs) = ndistrtemp(ntimelevs) + 1
            endif
         enddo
!$omp critical
         do j=1,ntimelevs
             ntstepdistr(j)=ntstepdistr(j) + ndistrtemp(j)
         enddo
!$omp end critical
!$omp end parallel

         if(iexit.eq.1)then
            print *,'************'
            print *,'************'
            print *,'************'
            print *,'WARNING: Particles in smallest timestep bin!'
            print *,'************'
            print *,'************'
            print *,'************'
         endif

         print *,'Next fully sync''ed time step binning will have:'
         j=1
         do while(j.le.ntimelevs)
            percent=dble(ntstepdistr(j))/dble(npart)*100d0
            print 55,j,t_unit*dtmax/2**(j-1),ntstepdistr(j),percent,
     &                ntstepdistrpm(j)
            j=j+1
         enddo
         print *,' '
      endif

55    format(i4,1pd20.12,0p,i16,'    ',f18.13,' %','  ',i5)
      return
      end subroutine report
!----------------------------------------------------------------
      subroutine printtime(npart,npoim)
!****************************************************************
!  Do a log file print out of the current time and information  *
!  about particles that are active and finishing their ts.      *
!****************************************************************

      integer, intent(in):: npart,npoim

      double precision frac,currtime

      frac=ONE
      if(iusebin.eq.1)frac=dble(lastts_end)*dtincr/dtmax

      currtime=t_unit*t_endts

      print 100,currtime,frac
      if(iexpand.eq.1)then
         print *,'Redshift',(ONE/ascale)-ONE,'Hubble',hnow
         print *,'Scale parameter',ascale
      endif
      print 101,npart,nactive,levldone_bigdt,levldone_smldt,
     &                        levlocc_bigdt,levlocc_smldt
      if(npoim.gt.0)then
         print 102,npoim,nactivepm,levldone_bigdt,levldone_smldt,
     &                             levlocc_bigdt,levlocc_smldt
      endif

      if(indts.eq.0.and.iusebin.eq.0)then
         print *,'dt=',TWOdtincr*t_unit,'  nstep=',nstep
      elseif(indts.eq.0 .and. iusebin.eq.1)then
         print *,'dt=',dtincr*dble(itstepbin(1))*t_unit,'  nstep=',nstep
      endif
      print *,' '
      call bufflush(5)

100   format('Current end timestep time=',f25.8,' TS fraction',f14.10)
101   format('npart=',i8,' nactive  =',i8,' just did levels ',i2,
     &  '-',i2,' of ',i2,'-',i2,' now occupied')
102   format('npoim=',i8,' nactivepm=',i8,' just did levels ',i2,
     &  '-',i2,' of ',i2,'-',i2,' now occupied')
      return
      end subroutine printtime
!----------------------------------------------------------------
      end module indivts
