      module integ
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!*********************************************************************
! This file contains the VINE module to do forward time              *
! integrations on particles using a 2nd order Leapfrog               *
! integrator. The specific implementation follows some of the        *
! discussion in:                                                     *
!    Springel et al. 2001 New Astronomy 6 79                         *
!                                                                    *
! Leapfrog is by nature an `offset' integrator, meaning that         *
! some variables are offset by 1/2 timestep from others.             *
! In simplest form, position and acceleration are on 1/2 steps,      *
! while velocity is on integer steps. That is what allows it         *
! to be 2nd order.                                                   *
!                                                                    *
! AN IMPORTANT IMPLEMENTATION PECULIARITY OF THIS FACT IN VINE       *
! is that the array deltat is defined strangely. Specifically,       *
! the timestep from one update to the next (i.e. n to n+1) is        *
! *2* times deltat because deltat is the time from the last          *
! full update to `now', and now is the 1/2 time step when the        *
! all of the derivative calculations are done.                       *
!                                                                    *
! For inactive particles, deltat is still the time of the last       *
! full update to `now' so that quantities can be extrapolated        * 
! to the present time correctly. This means that deltat has          *
! essentially nothing to do with the actual timestep for             *
! inactive particles, and can even be negative. Caveat Emptor.       *
!                                                                    *
! Another note: for individual timesteps density must be specially   *
! extrapolated in halfupdate because its last value relative to      *
! velocity or position is not at n or n+1/2 but rather at n-1/2.     *
! This means that it is necessary to give it a different dt.         *
!                                                                    *
!  `active' means that a particle is on its 1/2 timestep             *
!  and therefore requires a derivative calculation.                  *
!                                                                    *
! ============================================                       *
! Note about Nbody variable smoothing lengths.                       *
! ============================================                       *
! This code currently only implements smoothing lengths              *
! for nbody particles that are constant in time. Reasons             *
! for this are essentially that small scale random motions           *
! (when they occur) such as when bulk motions become                 *
! `thermalized' tend to couple to the smoothing length,              *
! causing a nasty resonance that oscillates faster and               *
! faster until the whole system turns into a quivering bowl          *
! of jelly. This is essentially an effect of the breakdown of        *
! the `collisionless' assumption in most nbody simulations.          *
! SPH particles don't suffer from this problem because the           *
! energy cascade from large scale to small ends up in heat,          *
! as small scale random motions are dissipated.                      *
!                                                                    *
! If you still want to try and include such effects, and this        *
! warning didn't scare you off, here is how:                         *
!                                                                    *
! 1) Redimension the arrays rho ,  rhosave,  rhohalf                 *
!                           divv, divvsave, divvhalf                 *
!                           dhdt1, nneightot                         *
!    to idim rather than isphdim. Further all invocations            *
!    of these arrays in various subroutines need the same            *
!    treatment. (i.e. in the i/o routines, you need to               *
!    output all npart particles, not just npart_sph)                 *
!                                                                    *
! 2) In subroutine deriv, move the if(npart_sph.eq.0) line           *
!    down to below the hdot call.                                    *
!                                                                    *
! 3) In subroutine density (in sph.F) change the value of iboth      *
!    to iboth=1, and in the storage loop at the end, save rho        *
!    and divv for all particles, not just sph particles.             *
!                                                                    *
! 4) Change the hdot subroutine arguments from the current           *
!    iactive_s and nactive_s to iactive and nactive.                 *
!                                                                    *
! 5) Move the h, rho and divv updates from the sph only loops in     *
!    of halfupdate and fullupdate, into the loops done for all       *
!    particles. Note that this means deleting lines like:            *
!    dumh(i)=h(i) because now you actually will change h, not        *
!    just copy it to the dummy/advanced time array.                  *
!                                                                    *
! 6) Move the h, rho and divv copyback values into the main          *
!    loop rather than their current location in the sph only loop    *
!    of copyback                                                     *
!                                                                    *
! 7) In lf_correct, include lines to extrapolate rho and             *
!    and divv backwards or forwards when particles change            *
!    timestep bins, for all particles, not just sph, as is done      *
!    now.                                                            *
!                                                                    *
! 8) Do a similar modification to extrap_half.                       *
!                                                                    *
! 9) In the hchange subroutine, include all particles, not just      *
!    sph particles.                                                  *
!                                                                    *
!10) Tell us if it works.                                            *
!                                                                    *
!                                                                    *
!  Written by Andy Nelson, Thorsten Naab and Markus Wetzstein        *
!*********************************************************************
      use dimens 
      use stubs
      use perform
      use partvars
      use integvars
      use miscvars, only:iexpand,ifullextrap
      use miscvars, only:umin,hmin,hmax,dhmax,alphastar
      use miscvars, only:t_unit
      use constants
#if COSMO
      use cosmo, only: cosmo_vel
#endif
      implicit none

      private
      public copyback,halfupdate,fullupdate,extrap_half
      public lf_correct
!
! Make sure that the preprocessor flag for leapfrog integrator
! is set, and all other integrators (defined by you--who wrote
! that other integrator) are not set, before going any further.
!
#ifndef LF
      leapfrog module: preprocessor flag LF is not set
#endif

#ifdef RK
      leapfrog module: preprocessor flag RK is set
#endif

!replace this with some other integrator that 
!    code has been written for 
#ifdef YOUR_INTEGRATOR
      leapfrog module: preprocessor flag YOUR_INTEGRATOR is set
#endif


      contains
!----------------------------------------------------------------
      subroutine copyback(nactive_s,nactive,nactivepm,
     &                    iactive_s,iactive,iactivepm)
!************************************************************
!                                                           *
!  copy temporary values and derivatives for particles      *
!  which have finished their time steps into the            *
!  beginning of time step arrays.                           *
!                                                           *
!************************************************************
      integer, intent(in):: nactive_s,iactive_s(nactive_s)
      integer, intent(in):: nactive  ,iactive  (nactive  )
      integer, intent(in):: nactivepm,iactivepm(nactivepm)

      integer i,j
      double precision tme1,tme2
      character(len=20), parameter:: where=('copyback')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

      do j=1,nactivepm
         i=iactivepm(j)
         nsteppm     (i)=nsteppm(i)+1
         oldpotpm    (i)=potpm          (i)
         posmaspm(ims,i)=dumposmaspm(ims,i)
         posmaspm(1  ,i)=dumposmaspm(1  ,i)
         posmaspm(2  ,i)=dumposmaspm(2  ,i)
         velpm   (1  ,i)=dumvelpm   (1  ,i)
         velpm   (2  ,i)=dumvelpm   (2  ,i)
#if NDIM==3
         posmaspm(3  ,i)=dumposmaspm(3  ,i)
         velpm   (3  ,i)=dumvelpm   (3  ,i)
#endif
      enddo

!$omp parallel do if(nactive_s .gt. 1024) default(none) 
!$omp& shared(nactive_s,iactive_s,rho,rhosave,divv,divvsave)
!$omp& shared(h,u,alpha,dumh,dumu,dumalf)
!$omp& private(i,j)
      do j=1,nactive_s
         i=iactive_s(j)
         rhosave (i)=rho    (i)
         divvsave(i)=divv   (i)
         h       (i)=dumh   (i)
         u       (i)=dumu   (i)
         alpha   (i)=dumalf (i)
      enddo

!$omp parallel do if(nactive .gt. 1024) default(none) 
!$omp& shared(nactive,iactive)
!$omp& shared(poten,oldpot,posmas,dumposmas,vel,dumvel)
!$omp& private(i,j)
      do j=1,nactive
         i=iactive(j)
         oldpot    (i)=poten        (i)
         posmas(ims,i)=dumposmas(ims,i)
         posmas(1  ,i)=dumposmas(1  ,i)
         posmas(2  ,i)=dumposmas(2  ,i)
         vel   (1  ,i)=dumvel   (1  ,i)
         vel   (2  ,i)=dumvel   (2  ,i)
#if NDIM==3
         posmas(3  ,i)=dumposmas(3  ,i)
         vel   (3  ,i)=dumvel   (3  ,i)
#endif
      enddo

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_copyback)=timing(itme_copyback)+ (tme2-tme1)
         ncalls(itme_copyback)=ncalls(itme_copyback)+ 1
         timing(itme_intgratr)=timing(itme_intgratr)+ (tme2-tme1)
!        ncalls(itme_intgratr) increment only for fullupdate calls
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine copyback
!----------------------------------------------------------------
      subroutine fullupdate(nactive_s,nactive_n,nactive,nactivepm,
     &                      iactive_s,iactive_n,iactive,iactivepm,
     &                      deltat,deltatpm)
!********************************************************************
!                                                                   *
!  Update particles on the update list                              *
!                                                                   *
! Notes:                                                            *
!  1) Order is important. Velocity updates precede position updates.*
!  2) the time step is 2*deltat because deltat is the time from     *
!     the last full update to `now', and now is the 1/2 time step   *
!                                                                   *
!********************************************************************
      integer, intent(in):: nactive_s, iactive_s(nactive_s)
      integer, intent(in):: nactive_n, iactive_n(nactive_n)
      integer, intent(in):: nactive  , iactive  (nactive  )
      integer, intent(in):: nactivepm, iactivepm(nactivepm)
      double precision, intent(in):: deltat  (idim)
      double precision, intent(in):: deltatpm(ipmdim)

      integer i,j
      double precision dthalf,dtfull,atmp,utmp,dh

      double precision tme1,tme2
      character(len=20), parameter:: where =('fullupdate')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

      if(iexpand.eq.1)then
         call cosmo_vel(nactive,nactivepm,iactive,iactivepm,
     &                  npart,npoim,deltat,deltatpm,
     &                  acc1,vel,dumvel,acc1pm,velpm,dumvelpm)

!$omp parallel do if(nactive .gt. 1024) default(none)
!$omp& shared(nactive,iactive,deltat,posmas,dumposmas,vel,dumvel)
!$omp& private(i,j,dthalf,dtfull) 
         do j=1,nactive
            i=iactive(j)
            dthalf=deltat(i)
            dtfull=two*deltat(i)

            dumposmas(ims,i)=posmas(ims,i)
            dumposmas(1  ,i)=posmas(1  ,i) + 
     &                        dthalf*(dumvel(1,i)+vel(1,i))
            dumposmas(2  ,i)=posmas(2  ,i) + 
     &                        dthalf*(dumvel(2,i)+vel(2,i))
#if NDIM==3
            dumposmas(3  ,i)=posmas(3  ,i) +
     &                        dthalf*(dumvel(3,i)+vel(3,i))
#endif
         enddo

         do j=1,nactivepm
            i=iactivepm(j)
            dthalf=    deltatpm(i)
            dtfull=two*deltatpm(i)

            dumhpm     (    i)=hpm(i)
            dumposmaspm(ims,i)=posmaspm(ims,i)
            dumposmaspm(1  ,i)=posmaspm(1  ,i) +
     &                        dthalf*(dumvelpm(1,i)+velpm(1,i))
            dumposmaspm(2  ,i)=posmaspm(2  ,i) +
     &                        dthalf*(dumvelpm(2,i)+velpm(2,i))
#if NDIM==3
            dumposmaspm(3  ,i)=posmaspm(3  ,i) +
     &                        dthalf*(dumvelpm(3,i)+velpm(3,i))
#endif
         enddo

      else 

!$omp parallel do if(nactive .gt. 1024) default(none)
!$omp&  shared(nactive,iactive,deltat,posmas,dumposmas,dumvel,vel,acc1)
!$omp&  private(i,j,dthalf,dtfull) 
         do j=1,nactive
            i=iactive(j)
            dthalf=    deltat(i)
            dtfull=two*deltat(i)

            dumvel   (1  ,i)=vel   (1  ,i) + dtfull*acc1(1,i)
            dumvel   (2  ,i)=vel   (2  ,i) + dtfull*acc1(2,i)
            dumposmas(ims,i)=posmas(ims,i)
            dumposmas(1  ,i)=posmas(1  ,i) + 
     &                         dthalf*(dumvel(1,i)+vel(1,i))
            dumposmas(2  ,i)=posmas(2  ,i) + 
     &                         dthalf*(dumvel(2,i)+vel(2,i))
#if NDIM==3
            dumvel   (3  ,i)=vel   (3  ,i) + dtfull*acc1(3,i)
            dumposmas(3  ,i)=posmas(3  ,i) + 
     &                         dthalf*(dumvel(3,i)+vel(3,i))
#endif
         enddo

         do j=1,nactivepm
            i=iactivepm(j)
            dthalf=    deltatpm(i)
            dtfull=two*deltatpm(i)

            dumvelpm   (1  ,i)=velpm   (1  ,i) + dtfull*acc1pm(1,i)
            dumvelpm   (2  ,i)=velpm   (2  ,i) + dtfull*acc1pm(2,i)
            dumhpm     (    i)=hpm(i)
            dumposmaspm(ims,i)=posmaspm(ims,i)
            dumposmaspm(1  ,i)=posmaspm(1  ,i) +
     &                         dthalf*(dumvelpm(1,i)+velpm(1,i))
            dumposmaspm(2  ,i)=posmaspm(2  ,i) +
     &                         dthalf*(dumvelpm(2,i)+velpm(2,i))
#if NDIM==3
            dumvelpm   (3  ,i)=velpm   (3  ,i) + dtfull*acc1pm(3,i)
            dumposmaspm(3  ,i)=posmaspm(3  ,i) +
     &                         dthalf*(dumvelpm(3,i)+velpm(3,i))
#endif
         enddo

      endif
!
! Hydro
!
!$omp parallel do if(nactive_s .gt. 1024) default(none)
!$omp& firstprivate(dhmax,hmax,hmin,umin,alphastar)
!$omp& shared(nactive_s,iactive_s,deltat)
!$omp& shared(h,dhdt1,dumh,u,dudt1,dumu,alpha,dalfdt1,dumalf)
!$omp& private(i,j,atmp,utmp,dh,dtfull) 
      do j=1,nactive_s
         i=iactive_s(j)
         dtfull=two*deltat(i)

         dh=dtfull*dhdt1(i)
         dh= sign(min(abs(dh),dhmax*h(i)),dh)
         dumh(i)=min(hmax,max(h(i) + dh,hmin))

         utmp  =u    (i) + dtfull*dudt1  (i)
         atmp  =alpha(i) + dtfull*dalfdt1(i)
         dumu  (i)=max(utmp,umin)
         dumalf(i)=max(atmp,alphastar)
      enddo

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_fullupdt)=timing(itme_fullupdt)+ (tme2-tme1)
         ncalls(itme_fullupdt)=ncalls(itme_fullupdt)+ 1
         timing(itme_intgratr)=timing(itme_intgratr)+ (tme2-tme1)
         ncalls(itme_intgratr)=ncalls(itme_intgratr)+ 1
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine fullupdate
!----------------------------------------------------------------
      subroutine halfupdate(nactivef_s,nactivef_n,nactivef,nactivefpm,
     &                      nactive_s ,nactive_n ,nactive ,nactivepm ,
     &                      iactive_s ,iactive_n ,iactive ,iactivepm ,
     &                      itstepbin,itstepbinpm,
     &                      dtincr,deltat,deltatpm)
!************************************************************
!  First:                                                   *
!  Extrapolate velocities to n+1/2 for SPH particles        *
!  so that integrator remains formally 2nd order.           *
!  This is required because SPH forces depend on the        *
!  velocity through the artificial viscosity.               *
!  (Recall that forces are calculated at the 1/2 timestep)  *
!                                                           *
!  deltat is the time since (or until, if negative)         *
!     the last full update step. In other words,            *
!     since the last velocity update at timestep n          *
!     (i.e. NOT n+1/2)                                      *
!                                                           *
!  Second:                                                  *
!  Synchronize quantities for particles that are inactive   *
!  to the values they should have at the current time,      *
!  including the densities, which are extrapolated          *
!  using Monaghan 1990 equation 3.9                         *
!                                                           *
!  NOTE:                                                    *
!  nactivef* are not used here. They are only included      *
!      to make the driver interface identical to the one    *
!      for other integrators.                               *
!                                                           *
!************************************************************
      integer, intent(in):: nactive_s,nactive_n,nactive,nactivepm
      integer, intent(in):: nactivef_s, iactive_s(isphdim)
      integer, intent(in):: nactivef_n, iactive_n(nboddim)
      integer, intent(in):: nactivef  , iactive  (   idim)
      integer, intent(in):: nactivefpm, iactivepm( ipmdim)
      integer, intent(in):: itstepbin(idim),itstepbinpm(ipmdim)
      double precision, intent(in):: dtincr, deltat(idim)
      double precision, intent(in):: deltatpm(ipmdim)

      integer i,j
      double precision dh,utmp,atmp,drho,dt_dens

      double precision tme1,tme2
      character(len=20), parameter:: where= ('halfupdate')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)


      select case(ifullextrap)
      case(1)
!
!==============ACTIVE and INACTIVE PARTICLES=========================
!
! Extrapolate particle quantities defined at n to n+1/2. 
!
!$omp parallel do default(none)
!$omp&  shared(npart,deltat)
!$omp&  shared(posmas,dumposmas,vel,dumvel,acc1,h,dumh)
!$omp&  private(i,j,dh)
!$omp&  shared(t_unit)
         do i=1,npart
            dumh     (    i)=h     (    i) !redo for sph below
            dumposmas(ims,i)=posmas(ims,i) 
            dumposmas(1  ,i)=posmas(1  ,i) + deltat(i)*vel (1,i)
            dumposmas(2  ,i)=posmas(2  ,i) + deltat(i)*vel (2,i)
            dumvel   (1  ,i)=vel   (1  ,i) + deltat(i)*acc1(1,i)
            dumvel   (2  ,i)=vel   (2  ,i) + deltat(i)*acc1(2,i)
#if NDIM==3
            dumposmas(3  ,i)=posmas(3  ,i) + deltat(i)*vel (3,i)
            dumvel   (3  ,i)=vel   (3  ,i) + deltat(i)*acc1(3,i)
#endif
         enddo

         do i=1,npoim
            dumhpm     (    i)=hpm(i)
            dumposmaspm(ims,i)=posmaspm(ims,i) 
            dumposmaspm(1  ,i)=posmaspm(1  ,i) + deltatpm(i)*velpm (1,i)
            dumposmaspm(2  ,i)=posmaspm(2  ,i) + deltatpm(i)*velpm (2,i)
            dumvelpm   (1  ,i)=velpm   (1  ,i) + deltatpm(i)*acc1pm(1,i)
            dumvelpm   (2  ,i)=velpm   (2  ,i) + deltatpm(i)*acc1pm(2,i)
#if NDIM==3
            dumposmaspm(3  ,i)=posmaspm(3  ,i) + deltatpm(i)*velpm (3,i)
            dumvelpm   (3  ,i)=velpm   (3  ,i) + deltatpm(i)*acc1pm(3,i)
#endif
         enddo
!
! Extrapolate hydrodynamic quantities to the current time.
!
! Density is special in this loop because its last value
! relative to velocity or position is not at n or n+1/2 but 
! rather at n-1/2. This means that it is necessary to give it
! its own dt in the loop below. Also, as implemented, dhmax will
! either limit the change in h to a maximum of dhmax*h at:
!              1) the end of the timestep and  (if i_hts.eq.1)
!                 any time during the timestep 
!       or     2) any time during the timestep (if i_hts.eq.0)
!
! Density for active particles gets overwritten when the sph
! quantities get updated, but we extrapolate it here anyway,
! for loop simplicity.
!
!$omp parallel do default(none)
!$omp& firstprivate(umin,alphastar,hmin,hmax,dhmax)
!$omp&   shared(dtincr,itstepbin,deltat,npart_sph)
!$omp&   shared(rhosave,divvsave,rho)
!$omp&   shared(u    ,dumu  ,  dudt1)
!$omp&   shared(alpha,dumalf,dalfdt1)
!$omp&   shared(h    ,dumh  ,  dhdt1)
!$omp&   private(i,dh,utmp,atmp,dt_dens,drho)
         do i=1,npart_sph

            dt_dens = deltat(i) + half*dtincr*dble(itstepbin(i))
            drho    = dt_dens*rhosave(i)*divvsave(i)
            rho  (i)= rhosave(i) - drho

            dh=deltat(i)*dhdt1(i)
            dh= sign(min(abs(dh),dhmax*h(i)),dh)
            dumh (i)=min(hmax,max(h(i) + dh,hmin))

            utmp=u    (i) + deltat(i)*dudt1  (i)
            atmp=alpha(i) + deltat(i)*dalfdt1(i)
            dumu  (i)=max(utmp,umin)
            dumalf(i)=max(atmp,alphastar)
         enddo

      case default
!
! Same as above, but for active particles only
!

!$omp parallel do default(none) if(nactive.gt.1024)
!$omp&  shared(nactive,iactive,deltat)
!$omp&  shared(posmas,dumposmas,vel,dumvel,acc1,h,dumh)
!$omp&  private(i,j,dh)
         do j=1,nactive
            i=iactive(j)

            dumh     (    i)=h     (    i) !redo for sph below
            dumposmas(ims,i)=posmas(ims,i) 
            dumposmas(1  ,i)=posmas(1  ,i) + deltat(i)*vel (1,i)
            dumposmas(2  ,i)=posmas(2  ,i) + deltat(i)*vel (2,i)
            dumvel   (1  ,i)=vel   (1  ,i) + deltat(i)*acc1(1,i)
            dumvel   (2  ,i)=vel   (2  ,i) + deltat(i)*acc1(2,i)
#if NDIM==3
            dumposmas(3  ,i)=posmas(3  ,i) + deltat(i)*vel (3,i)
            dumvel   (3  ,i)=vel   (3  ,i) + deltat(i)*acc1(3,i)
#endif
         enddo

         do j=1,nactivepm
            i=iactivepm(j)  
            dumhpm     (    i)=hpm(i)
            dumposmaspm(ims,i)=posmaspm(ims,i) 
            dumposmaspm(1  ,i)=posmaspm(1  ,i) + deltatpm(i)*velpm (1,i)
            dumposmaspm(2  ,i)=posmaspm(2  ,i) + deltatpm(i)*velpm (2,i)
            dumvelpm   (1  ,i)=velpm   (1  ,i) + deltatpm(i)*acc1pm(1,i)
            dumvelpm   (2  ,i)=velpm   (2  ,i) + deltatpm(i)*acc1pm(2,i)
#if NDIM==3
            dumposmaspm(3  ,i)=posmaspm(3  ,i) + deltatpm(i)*velpm (3,i)
            dumvelpm   (3  ,i)=velpm   (3  ,i) + deltatpm(i)*acc1pm(3,i)
#endif
         enddo

!$omp parallel do default(none) if(nactive_s.gt.1024)
!$omp& firstprivate(umin,alphastar,hmin,hmax,dhmax)
!$omp&   shared(dtincr,itstepbin,deltat,nactive_s,iactive_s)
!$omp&   shared(rhosave,divvsave,rho)
!$omp&   shared(u    ,dumu  ,  dudt1)
!$omp&   shared(alpha,dumalf,dalfdt1)
!$omp&   shared(h    ,dumh  ,  dhdt1)
!$omp&   private(i,j,dh,utmp,atmp,dt_dens,drho)
         do j=1,nactive_s
            i=iactive_s(j)

            dt_dens = deltat(i) + half*dtincr*dble(itstepbin(i))
            drho    = dt_dens*rhosave(i)*divvsave(i)
            rho  (i)= rhosave(i) - drho              !will be overwritten

            dh=deltat(i)*dhdt1(i)
            dh= sign(min(abs(dh),dhmax*h(i)),dh)
            dumh (i)=min(hmax,max(h(i) + dh,hmin))

            utmp=u    (i) + deltat(i)*dudt1  (i)
            atmp=alpha(i) + deltat(i)*dalfdt1(i)
            dumu  (i)=max(utmp,umin)
            dumalf(i)=max(atmp,alphastar)
         enddo

      end select

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_halfupdt)=timing(itme_halfupdt)+ (tme2-tme1)
         ncalls(itme_halfupdt)=ncalls(itme_halfupdt)+ 1
         timing(itme_intgratr)=timing(itme_intgratr)+ (tme2-tme1)
!        ncalls(itme_intgratr) increment only for fullstep calls
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine halfupdate
!----------------------------------------------------------------
      subroutine extrap_half(deltat,deltatpm)
!*******************************************************************
!  Apply correction factor to make the leap-frog integrator        *
!     algorithm self-starting.                                     *
!   For density and divergence, do only first (zeroth) order       *
!   correction, and hope that is good enough.                      *
!*******************************************************************
!
! initial deltat defn is 1/2 of the current step.
!
      double precision, intent(in):: deltat(idim),deltatpm(ipmdim)

      integer i
      double precision drho
      character(len=20), parameter:: where= ('extrap_half')
      if(idbg.ge.2)print *,where

! Note minus sign in drho is correct. Need to extrap 1/2 step backwards.

!$omp parallel do if(npart .gt. 1024) default(none)
!$omp& firstprivate(npart_sph)
!$omp& shared(npart,rhosave,divvsave,poten,oldpot,deltat)
!$omp& private(i,drho) 
      do i=1,npart
         oldpot(i)=poten(i)

         if(i.le.npart_sph)then
            drho=-deltat(i)*rhosave(i)*divvsave(i)
            rhosave (i)=rhosave (i)-drho
            divvsave(i)=divvsave(i)
         endif
      enddo


      do i=1,npoim
         oldpotpm(i)=potpm(i)
      enddo

      return
      end subroutine extrap_half
!----------------------------------------------------------------
      subroutine lf_correct(nactive_s,nactive,nactivepm,
     &                      iactive_s,iactive,iactivepm,
     &                      itstepbin,itstepbinpm,
     &                      dtincr,deltat,deltatpm)
!*******************************************************************
!                                                                  *
! Correct several variables for the fact that the particle's       *
! timestep has changed.                                            *
!                                                                  *
!  This subroutine depends on the fact that deltat has not         *
!  yet been reset, and therefore contains its `old' value.         *
!  The rho and divv corrections will be incorrect if this          *
!  condition is violated.                                          *
!                                                                  *
!  Included here is a crude method of updating the potential to    *
!  the full timestep (where the position is) so that total         *
!  energy calculations (if desired) can be made self               *
!  consistently. Can be done better with an explicit               *
!  calculation.                                                    * 
!                                                                  *
!  The rho and divv corrections are done in the following manner:  *
!   Consider a particle moving upwards in a timestep hierarchy     *
!   from a sequency of positions on n-1/2, n+1/2, n+3/2 etc        *
!   and velocities on n,n+1,n+2 etc. This means that densities     *
!   and divvs will be calculated on n-1/2,n+1/2,n+3/2 etc and      *
!   at any given time we have                                      *
!            rho(n-1/2), v(n), x(n)                                *
!   Now we want to move to a sequence where the stepping is        *
!            rho(n-1)  , v(n), x(n)                                *
!  Therefore we need to move density backwards in time from        *
!  where it was originally calculated. The exact opposite will     *
!  be true for moving downwards in the hierarchy: density must     *
!  move forward in time.                                           *
!  How much?                                                       *
!    We need to go 1/2 of 1 old timestep forward, and 1/2 of       *
!    1 new timestep backwards again. The net result is a           *
!    time offset of deltat(i) - half*dt_new, since deltat is       *
!    defined as 1/2 of one timestep and has not yet been updated   *
!    to its new value.                                             *
!  How do we change the quantities?                                *
!    We assume divv changes linearly between its last update       *
!    and the one before that (we still have both, since the        *
!    copyback hasn't yet happened), then we can get a slope        *
!    and extrapolate divv forwards. Then we use the new divv       *
!    to update the new rho.                                        *
!*******************************************************************
      integer, intent(in):: nactive_s, iactive_s(nactive_s)
      integer, intent(in):: nactive  , iactive  (nactive  )
      integer, intent(in):: nactivepm, iactivepm(nactivepm)
      integer, intent(in):: itstepbin(idim),itstepbinpm(ipmdim)
      double precision, intent(in):: dtincr, deltat(idim)
      double precision, intent(in):: deltatpm(ipmdim)

      integer i,j
      double precision tme1,tme2
      double precision dpotdt,dt_old,dt_new,dtoffset,divvdot,drho
      character(len=20), parameter:: where =('lf_correct')
      if(itme.ge.1)call cpu_time(tme1)

!$omp parallel if(nactive .gt. 1024) default(none) 
!$omp&  shared(nactive,nactive_s,iactive,iactive_s,itstepbin)
!$omp&  shared(dtincr,deltat,oldpot,poten,divv,rho,divvsave)
!$omp&  private(i,j,drho)
!$omp&  private(dt_old,dt_new,dtoffset,divvdot,dpotdt)
!$omp do
      do j=1,nactive
         i=iactive(j)

         dt_old= two*deltat(i)
         dpotdt=(poten(i) - oldpot(i))/dt_old
         poten(i)= poten(i) + half*dt_old*dpotdt
      enddo
!$omp end do nowait
!$omp do
      do j=1,nactive_s
         i=iactive_s(j)

         dt_old= two*deltat(i)
         dt_new= dtincr*dble(itstepbin(i))

         if(abs(dt_old-dt_new).gt.dtincr)then

            dtoffset=deltat(i) - half*dt_new
            divvdot= (divv(i) - divvsave(i))/dt_old
            divv(i)=divv(i) + dtoffset*divvdot

            drho=  dtoffset*rho(i)*divv(i)
            rho (i)=rho (i)- drho

         endif
      enddo
!$omp end parallel

      do j=1,nactivepm
         i=iactivepm(j)

         dt_old= two*deltatpm(i)
         dpotdt=(potpm(i) - oldpotpm(i))/dt_old
         potpm(i)= potpm(i) + half*dt_old*dpotdt

      enddo


      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_correct )=timing(itme_correct )+ (tme2-tme1)
         ncalls(itme_correct )=ncalls(itme_correct )+ 1
         timing(itme_intgratr)=timing(itme_intgratr)+ (tme2-tme1)
!        ncalls(itme_intgratr) increment only for fullstep calls
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine lf_correct
!----------------------------------------------------------------
      end module integ
