      module pmcalcs
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!*********************************************************************
!  This file contains the VINE module to do force calculations       *
!  involving point masses and a routine for doing mass accretion     *
!  onto point masses from sph/nbody particles, by bleeding mass      *
!  away from the sph/nbody particles. The latter doesn't             *
!  actually work. Back in the dark ages, I tried bleeding particle   *
!  accretion and didn't like the results. Frequently particles would *
!  lose most of their mass, then escape to the fluid again. That     *
!  meant that you'd have some of your sph particles as regular       *
!  particles, and some that had substantially less mass than a       *
!  regular particle. It is possible to end up with what amounts to   *
!  Brownian motion for these low mass particles.                     *
!                                                                    *
!  Written by Andy Nelson                                            *
!*********************************************************************
      use dimens
      use perform
      use constants

      implicit none

      private
      public pmgrav,mdot

      contains
!------------------------------------------------------------------
      subroutine pmgrav(nactive,iactive,npart,nactivepm,iactivepm,npoim,
     &           posmas,acc,poten,posmaspm,hpm,accpm,potpm)
!*********************************************************************
!                                                                    *
!  This subroutine computes the acceleration on the particles due to * 
!  the presence of point masses. It also computes the forces on the  *
!  point masses due to the particles. The only force considered here *
!  is gravity.                                                       *
!                                                                    *
!  Someday this should really be converted to use kernel softening   *
!  too/instead.                                                      *
!                                                                    *
!  Potential is calculated as a positive quantity                    *
!                                                                    *
!*********************************************************************
      integer, intent(in):: npart, nactive  , iactive  (npart)
      integer, intent(in):: npoim, nactivepm, iactivepm(npoim) 

      double precision, intent(in):: posmas  (ndim1,npart)
      double precision, intent(in):: posmaspm(ndim1,npoim), hpm(npoim)
      double precision, intent(inout):: acc  (ndim,npart), poten(npart)
      double precision, intent(inout):: accpm(ndim,npoim), potpm(npoim)

      integer i,j,ipoim,jpoim,kpoim
      double precision tme1,tme2
      double precision dxij,dyij,dzij,d2,di,di2,fpi,fpj
      double precision fpmxi,fpmyi,fpmzi,pmpoteni
      double precision pmeps2(ipmdim)

      character(len=20), parameter:: where =('pmgrav')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

      if(npoim.eq.0)return

      do i=1,npoim
         pmeps2(i)= hpm(i)**2
      enddo
!
!--compute accelerations on point masses
!  ------------------------------------- 
!
      do jpoim=1,nactivepm
         ipoim=iactivepm(jpoim)
         fpmxi=zero
         fpmyi=zero
         fpmzi=zero
         pmpoteni=zero
!$omp parallel do default(none) 
!$omp&  shared(npart,ipoim,posmas,posmaspm,pmeps2) 
!$omp&  private(i,dxij,dyij,dzij,d2,di,di2,fpi)
!$omp&  reduction(-:fpmxi,fpmyi,fpmzi) reduction(+:pmpoteni)
         do i=1,npart

            dxij=posmaspm(1,ipoim)-posmas(1,i)
            dyij=posmaspm(2,ipoim)-posmas(2,i)
            d2=dxij*dxij + dyij*dyij + pmeps2(ipoim)
#if NDIM==3
            dzij=posmaspm(3,ipoim)-posmas(3,i)
            d2=d2 + dzij*dzij 
#endif
            di=one/sqrt(d2)
            di2=di*di

            pmpoteni=pmpoteni + posmas(ims,i)*di
            fpi=posmas(ims,i)*di2*di
            fpmxi=fpmxi - fpi*dxij
            fpmyi=fpmyi - fpi*dyij
#if NDIM==3
            fpmzi=fpmzi - fpi*dzij
#endif
         enddo
         potpm(  ipoim)=pmpoteni
         accpm(1,ipoim)=fpmxi
         accpm(2,ipoim)=fpmyi
#if NDIM==3
         accpm(3,ipoim)=fpmzi
#endif
      enddo

!
!--compute accelerations on particles
!  ---------------------------
!
! mipspro7.4 compiler bug activates here if loop fusion is allowed.
!
      do ipoim=1,npoim
!$omp parallel do default(none)
!$omp& private(i,j,dxij,dyij,dzij,di,di2,fpj,d2)
!$omp&  shared(ipoim,nactive,iactive)
!$omp&  shared(posmas,posmaspm,acc,poten,pmeps2)
         do j=1,nactive
            i=iactive(j)

            dxij=posmas(1,i) - posmaspm(1,ipoim)
            dyij=posmas(2,i) - posmaspm(2,ipoim)
            d2=dxij*dxij + dyij*dyij + pmeps2(ipoim)
#if NDIM==3
            dzij=posmas(3,i) - posmaspm(3,ipoim)
            d2= dzij*dzij + d2
#endif
            di=one/sqrt(d2)
            di2=di*di

            poten(i)=poten(i) + posmaspm(ims,ipoim)*di
            fpj=posmaspm(ims,ipoim)*di2*di
            acc(1,i)=acc(1,i) - fpj*dxij
            acc(2,i)=acc(2,i) - fpj*dyij
#if NDIM==3
            acc(3,i)=acc(3,i) - fpj*dzij
#endif
         enddo
      enddo
!
!--compute interaction between point masses (standard N-body)
!  ----------------------------------------------------------
!
      do jpoim=1,nactivepm
         ipoim=iactivepm(jpoim)
         fpmxi   =zero
         fpmyi   =zero
         fpmzi   =zero
         pmpoteni=zero
         do kpoim=1,npoim
            if(ipoim.ne.kpoim)then
               dxij=posmaspm(1,ipoim)-posmaspm(1,kpoim)
               dyij=posmaspm(2,ipoim)-posmaspm(2,kpoim)
               d2=dxij*dxij + dyij*dyij + pmeps2(ipoim)+pmeps2(kpoim)
#if NDIM==3
               dzij=posmaspm(3,ipoim)-posmaspm(3,kpoim)
               d2  =d2 + dzij*dzij
#endif
               di=one/sqrt(d2)
               di2=di*di

               pmpoteni=pmpoteni + posmaspm(ims,kpoim)*di
               fpi  =posmaspm(ims,kpoim)*di2*di
               fpmxi=fpmxi - fpi*dxij
               fpmyi=fpmyi - fpi*dyij
#if NDIM==3
               fpmzi=fpmzi - fpi*dzij
#endif
            endif
         enddo
         potpm(  ipoim)=potpm(  ipoim) + pmpoteni
         accpm(1,ipoim)=accpm(1,ipoim) + fpmxi
         accpm(2,ipoim)=accpm(2,ipoim) + fpmyi
#if NDIM==3
         accpm(3,ipoim)=accpm(3,ipoim) + fpmzi
#endif
      enddo

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_PMgrav)=timing(itme_PMgrav)+tme2-tme1
         ncalls(itme_PMgrav)=ncalls(itme_PMgrav)+1
         timing(itme_gravity)=timing(itme_gravity)+tme2-tme1
!         ncalls(itme_gravity) increment only in gravity sub
      endif

      if(idbg.ge.3)print *,'done ',where

      return
      end subroutine pmgrav
!------------------------------------------------------------------
      subroutine mdot(npart,posmas,dm,npoim,posmaspm,dmpm)
!**********************************************************************
!                                                                     *
!  subroutine computing the rate of change of particle masses due to  *
!  the presence of sink particles (the point masses).                 *
!                                                                     *
!**********************************************************************
      use miscvars, only: irem

      integer, intent(in):: npart,npoim
      double precision, intent(in):: posmas  (ndim1,npart)
      double precision, intent(in):: posmaspm(ndim1,npoim)
      double precision, intent(out):: dm(npart), dmpm(npoim)
!
! xmu is a rate constant for mass transfer.
!
      integer i
      double precision, parameter ::xmu=.05d0
      character(len=20), parameter:: where =('mdot')
      if(idbg.ge.2)print *,where    
!
!--initialize
!
!$omp parallel do default(none) private(i) shared(npart,dm)
      do i=1,npart
         dm(i)=0d0
      enddo
      do i=1,npoim
         dmpm(i)=0d0  
      enddo

      if(irem.ne.2)return

!  fix mdot to remember how to bleed particles!

      end subroutine mdot
!------------------------------------------------------------------
      end module pmcalcs
