      module sph
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!****************************************************************
!  This file contains the VINE module to do hydrodynamic        *
!  calculations on particles, using the SPH method.             *
!                                                               *
!  This module requires the tree module, in order to            *
!  determine and return a list of neighbors for each active     *
!  SPH and (in some circumstances-for the density) Nbody        *
!  particle.                                                    *
!                                                               *
!  The SPH kernel can be easily swapped for another kernel      *
!  by swapping out the currently implemented kernel module      *
!  for another one, so long as the same arguments are returned  *
!                                                               *
!  Some important variables used in these routines:             *
!    iclump         : Clump that we're working on now.          *
!    ntot_a         : Number of active particles in iclump      *
!    ntot_i         : Number of inactive particles in iclump    *
!    listpt_a       : list of active particles in iclump        *
!    listpt_i       : list of inactive particles in iclump      *
!    distmas(idx,,) : xi-xj                                     *
!    distmas(idy,,) : yi-yj                                     *
!    distmas(idz,,) : zi-zj                                     *
!    distmas(ims,,) : mj                                        *
!    distmas(ir2,,) : squared distance between i and j          *
!    distmas(ih2,,) : squared smoothing length                  *
!    listnay(  ,  ) : list of neighbor indices                  *
!    nneigh (     ) : Number of neighbors                       *
!    WdW(1, ) : SPH kernel value                                *
!    WdW(2, ) : SPH kernel derivative value                     *
!                                                               *
! COSMOLOGY                                                     *
!                                                               *
! Positions and velocities are in comoving coordinates.         *
! The hydro equations are computed in physical coordinates      *
! as in Katz, Weinberg, Hernquist ApJS (1996) 105, 19.          *
! So we need to correct those quantities which are comoving     *
! to their value in the physical frame. We define correction    *
! factors for these quantities and if no cosmological           *
! expansion is used, the factors are 1.                         *
! The density array rho is stored in comoving, so any           *
! time rho is used, it needs to be transformed to physical      *
! frame.                                                        *
!                                                               *
!  Written by Andy Nelson, Markus Wetzstein and Thorsten Naab   *
!****************************************************************
#ifndef SPH
      sph module: SPH proprocessor flag not set
#endif
      use dimens
      use miscvars
      use sphvars
      use constants
      use perform
      use tree, only: getparticles, initgetneigh, getneigh
      use kernel, only: W0, v2max, SPHkerntablinit
      use kernel, only: SPHkern => SPHkern_ana    !switch *_ana or 
                                                  ! *_tab as desired
      implicit none

      private
      public SPHkerntablinit
      public density,sphaccel,viscdot,hdot,cour_ts


      integer ntot_a,ntot_i
      integer listpt_a(maxclumpsize), listpt_i(maxclumpsize)
      integer nneigh(maxclumpsize),listnay(maxbite,maxclumpsize)
      double precision distmas(ndim3,maxbite,maxclumpsize)
      double precision WdW(2,maxbite,maxclumpsize)

      contains
!--------------------------------------------------------------------
      subroutine density(nactive_c,iactive_c,
     &                   npart,npart_sph,posmas,vel,h,rho,divv)
!****************************************************************
!                                                               *
!  Compute the density and the velocity divergence              *
!  at each particle location.                                   *
!                                                               *
!****************************************************************
      integer, intent(in):: npart,npart_sph
      integer, intent(in):: nactive_c,iactive_c(nactive_c)
      double precision, intent(in):: posmas(ndim1,npart)
      double precision, intent(in):: vel(ndim,npart),h(npart)
      double precision, intent(out):: rho(npart_sph), divv(npart_sph) 

      integer nnaytot(maxclumpsize)
      double precision rhoi(maxclumpsize),divvi(maxclumpsize)
      double precision curlvx(maxclumpsize),curlvy(maxclumpsize)
      double precision curlvz(maxclumpsize)

      integer iboth,jclump,iclump,j,i,k,knay
      integer nstart,moreneig
      double precision dvx,dvy,dvz,projv,crosx,crosy,crosz
      double precision absdivv,abscurlv,dself

      double precision tme1,tme2
      character(len=20), parameter:: where =('density')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)
!
! compute density, velocity divergence and curl (times density) 
! for all particles, using the neighbors returned by getneigh
! and the kernel SPHkern. 
!  If iboth=0, getneigh returns zero neighbors for nbody particles
!               meaning that rho, divv, and curlv aren't calculated.
!  If iboth=1, then we do get nbody neighbors of nbody particles.
!              doing so is useful if you implement a dynamically 
!              changing smoothing length for nbody particles.
!              Note that if you do this, the storage loop below
!              needs to be modified to allow storage of nbody particles
!
      iboth=0

!$omp parallel default(none)
!$omp&  firstprivate(ibals,ibctype,iboth)
!$omp&  shared(npart_sph,nneightot,posmas,vel,h,rho,divv,thermo)
!$omp&  shared(nactive_c,iactive_c)
!$omp&  private(jclump,iclump,ntot_a,listpt_a,ntot_i,listpt_i,
!$omp&      i,j,k,knay,nnaytot,rhoi,divvi,absdivv,abscurlv,moreneig,
!$omp&      nstart,nneigh,listnay,distmas,WdW,dself,
!$omp&      dvx,dvy,dvz,projv,crosx,crosy,crosz,curlvx,curlvy,curlvz)

      call initgetneigh

!$omp do schedule(dynamic,32)
      do jclump=1,nactive_c
         iclump=iactive_c(jclump)
         call getparticles(iclump,maxclumpsize,
     &                     ntot_a,ntot_i,listpt_a,listpt_i)
         do j=1,ntot_a
            i=listpt_a(j)
            nnaytot(j)=0
            rhoi  (j)=zero
            divvi (j)=zero
            curlvx(j)=zero
            curlvy(j)=zero
            curlvz(j)=zero
         enddo

         moreneig=1
         nstart=0
         do while(moreneig.ne.0)
            call getneigh(ibctype,iboth,iclump,ntot_a,ntot_i,
     &                listpt_a,listpt_i,nstart,
     &                nneigh,listnay,distmas,moreneig)
            call SPHkern(1,nneigh,ntot_a,distmas,WdW)

            do j=1,ntot_a
               i=listpt_a(j)
               nnaytot(j)=nnaytot(j) + nneigh(j)
!
! compute each neighbor's density contribution, then calculate 
! pairwise velocity divergence and curl (both times density)
!
               do k=1,nneigh(j)
                  knay=listnay(k,j)
                  rhoi(j)=rhoi(j) + distmas(ims,k,j)*WdW(1,k,j)

                  dvx=vel(1,i)-vel(1,knay)
                  dvy=vel(2,i)-vel(2,knay)
                  projv=WdW(2,k,j)*(dvx*distmas(idx,k,j) +
     &                              dvy*distmas(idy,k,j))
#if NDIM==3
                  dvz=vel(3,i)-vel(3,knay)
                  projv= projv + WdW(2,k,j)*dvz*distmas(idz,k,j)
                  crosx=dvz*distmas(idy,k,j) - dvy*distmas(idz,k,j)
                  crosy=dvx*distmas(idz,k,j) - dvz*distmas(idx,k,j)
                  curlvx(j)=curlvx(j)-WdW(2,k,j)*distmas(ims,k,j)*crosx
                  curlvy(j)=curlvy(j)-WdW(2,k,j)*distmas(ims,k,j)*crosy
#endif
                  divvi(j)=divvi(j) - distmas(ims,k,j)*projv
                  crosz=dvy*distmas(idx,k,j) - dvx*distmas(idy,k,j)
                  curlvz(j)=curlvz(j)-WdW(2,k,j)*distmas(ims,k,j)*crosz

               enddo
            enddo
         enddo
!
! save quantities for later.
!
         do j=1,ntot_a
            i=listpt_a(j)
            if(i.le.npart_sph)then
               nneightot(i)=nnaytot(j)

               dself=posmas(ims,i)*W0/h(i)**ndim
               rho (i)=(rhoi (j) + dself) 
               divv(i)=divvi(j)/rho(i) 
               absdivv =abs(divvi (j))
               abscurlv=abs(curlvz(j))
#if NDIM==3
               abscurlv=sqrt(curlvx(j)**2 + curlvy(j)**2 + curlvz(j)**2)
#endif
               thermo(i_bal,i)=absdivv/(absdivv+abscurlv+small)
            endif
         enddo
         if(ibals.eq.0)then
            do j=1,ntot_a
               i=listpt_a(j)
               if(i.le.npart_sph)thermo(i_bal,i)=one
            enddo
         endif
      enddo        !jclump loop
!$omp end parallel

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_density)=timing(itme_density) + tme2-tme1
         ncalls(itme_density)=ncalls(itme_density) + 1
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine density
!--------------------------------------------------------------------
      subroutine sphaccel(nactive_c,iactive_c,npart,npart_sph,
     &                     vel,rho,alpha,acc)
!****************************************************************
!                                                               *
!  Compute the gas dynamical accelerations of the particles     *
!  that need new updates.                                       *
!                                                               *
!****************************************************************
      integer, intent(in):: npart,npart_sph
      integer, intent(in):: nactive_c,iactive_c(nactive_c)
      double precision, intent(in):: vel(ndim,npart)
      double precision, intent(in):: rho(npart_sph),alpha(npart_sph)
      double precision, intent(inout):: acc(ndim,npart)

      double precision, parameter:: epsil=1d-2

      double precision fxi(maxclumpsize),fyi(maxclumpsize)
      double precision fzi(maxclumpsize)
      double precision pdvi(maxclumpsize),dqi(maxclumpsize)
      double precision rmui(maxclumpsize)

      integer i,j,k,knay,iclump,jclump,moreneig
      integer nstart
      double precision press_on,visc_on,viscratio
      double precision rhoi,vsoundi,fbali,alphai,betai,artvij
      double precision projv,emgrW,gradpij,pro2j,pro2i,divvel,PIij
      double precision alfabar,betabar,fbalbar,vsbar,robar
      double precision cosmo_a1,cosmo_a2,cosmo_rho
      double precision cosmo_a1i,cosmo_a2i,cosmo_a3i,cosmo_a4i

      double precision tme1,tme2
      character(len=20), parameter:: where =('sphaccel')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

      press_on=dble(ipres)
      visc_on=dble(ivisc)
      viscratio=betastar/alphastar
!
! Cosmological expansion. As h, vel and pos are in comoving coordinates,
! we need to convert them and derived quantities to physical coordinates.
! For more details, see the density routine.
!
      if(iexpand.eq.1)then
         cosmo_a1=ascale       ; cosmo_a1i=one/cosmo_a1
         cosmo_a2=ascale**2    ; cosmo_a2i=one/cosmo_a2
         cosmo_a3i=one/ascale**3
         cosmo_a4i=one/ascale**4
         cosmo_rho=one/ascale**ndim
      else
         cosmo_a1=one          ; cosmo_a1i=one
         cosmo_a2=one          ; cosmo_a2i=one
         cosmo_a3i=one
         cosmo_a4i=one
         cosmo_rho=one
      endif

!$omp parallel default(none) 
!$omp&  firstprivate(press_on,visc_on,viscratio,ibctype)
!$omp&  firstprivate(cosmo_a1,cosmo_a2,cosmo_a3i,cosmo_a4i,cosmo_a2i)
!$omp&  firstprivate(cosmo_a1i,cosmo_rho)
!$omp&  shared(npart_sph,nactive_c,iactive_c)
!$omp&  shared(vel,acc,rho,thermo,alpha)
!$omp&  shared(pdv,dq,rmu)
!$omp&  private(j,i,iclump,jclump,
!$omp&    ntot_a,listpt_a,ntot_i,listpt_i,
!$omp&    rhoi,pro2i,vsoundi,fbali,alphai,betai,
!$omp&    fxi,fyi,fzi,pdvi,dqi,rmui,
!$omp&    moreneig,nstart,nneigh,listnay,distmas,knay,k,
!$omp&    WdW,emgrW,gradpij,pro2j,
!$omp&    projv,alfabar,betabar,fbalbar,vsbar,robar,divvel,PIij,artvij)

      call initgetneigh

!$omp do schedule(dynamic,32)
      do jclump=1,nactive_c

         iclump=iactive_c(jclump)
         call getparticles(iclump,maxclumpsize,
     &                     ntot_a,ntot_i,listpt_a,listpt_i)
!
! initialize local arrays
!
         do j=1,ntot_a
            pdvi (j)=zero
            dqi  (j)=zero
            rmui (j)=zero
            fxi  (j)=zero
            fyi  (j)=zero
#if NDIM==3
            fzi  (j)=zero
#endif
         enddo
         moreneig=1
         nstart=0
         do while(moreneig.ne.0)
            call getneigh(ibctype,0,iclump,ntot_a,ntot_i,
     &                listpt_a,listpt_i,nstart,
     &                nneigh,listnay,distmas,moreneig)
            call SPHkern(2,nneigh,ntot_a,distmas,WdW)
            do j=1,ntot_a
               i=listpt_a(j)
               if(i.le.npart_sph)then      !avoid out of bounds array 
                                           !accesses for nbody particles
                  rhoi   =rho(i)*cosmo_rho
                  pro2i  =thermo(i_pres ,i)/(rhoi*cosmo_rho)**2
                  vsoundi=thermo(i_sound,i)
                  fbali  =thermo(i_bal  ,i)
                  alphai =alpha(i)
                  betai  =viscratio*alpha(i)
               endif
! 
! loop over neighbor list
!
               do k=1,nneigh(j)        !iboth=0 in getneigh call above, 
                  knay=listnay(k,j)    !means no nbody neighbors
!
! pressure gradient term and pdv
!
                  emgrW=distmas(ims,k,j)*WdW(2,k,j)
                  pro2j=thermo(i_pres,knay)/(rho(knay)*cosmo_rho)**2
                  gradpij= press_on*(pro2i + pro2j)*emgrW*cosmo_a4i
                  fxi(j)=fxi(j)-gradpij*distmas(idx,k,j)
                  fyi(j)=fyi(j)-gradpij*distmas(idy,k,j)

                  projv= (vel(1,i) - vel(1,knay))*distmas(idx,k,j)  
     &                  +(vel(2,i) - vel(2,knay))*distmas(idy,k,j) 
#if NDIM==3
     &                  +(vel(3,i) - vel(3,knay))*distmas(idz,k,j)
                  fzi(j)=fzi(j)-gradpij*distmas(idz,k,j)
#endif
                  projv=projv * cosmo_a2
                  pdvi(j) = pdvi(j) + emgrW*projv*cosmo_a3i

                  if(projv.lt.zero)then
!
! artificial viscosity
!
                     alfabar=half*( alphai +           alpha(knay))
                     betabar=half*(  betai + viscratio*alpha(knay))
                     fbalbar=half*(  fbali +  thermo(  i_bal,knay))
                     vsbar  =half*(vsoundi +  thermo(i_sound,knay))
                     robar  =half*(   rhoi +  rho(knay))*cosmo_rho
                     divvel=projv*sqrt(distmas(ih2,k,j))*cosmo_a1i/
     &                       (distmas(ir2,k,j)+epsil*distmas(ih2,k,j))
                     divvel=fbalbar*divvel
                     PIij=divvel*(betabar*divvel-alfabar*vsbar)/robar
                     artvij=visc_on*emgrW*PIij*cosmo_a4i
                     rmui(j)=max(rmui(j),-divvel)
                     dqi (j)=dqi(j) + artvij*projv*cosmo_a1
                     fxi (j)=fxi(j) - artvij*distmas(idx,k,j)
                     fyi (j)=fyi(j) - artvij*distmas(idy,k,j)
#if NDIM==3
                     fzi (j)=fzi(j) - artvij*distmas(idz,k,j)
#endif
                  endif
               enddo
            enddo
         enddo
         do j=1,ntot_a
            i=listpt_a(j)
            if(i.le.npart_sph)then           !avoid assigning hydro
               pdv(i)   = press_on*pdvi(j)   !quantities to nbody 
               dq (i)   = visc_on *dqi (j)   !particles (i.e. out of
               rmu(i)   = visc_on *rmui(j)   !bounds array access)
               acc(1,i) = acc(1,i) + fxi(j)
               acc(2,i) = acc(2,i) + fyi(j)
#if NDIM==3
               acc(3,i) = acc(3,i) + fzi(j)
#endif
            endif
         enddo         

      enddo  !loop over clumps
!$omp end parallel

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_sphaccel)=timing(itme_sphaccel) + tme2-tme1
         ncalls(itme_sphaccel)=ncalls(itme_sphaccel) + 1
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine sphaccel
!--------------------------------------------------------------------
      subroutine hdot(nactive_s,iactive_s,nactive_n,iactive_n,
     &      itstepbin,npart,npart_sph,dtincr,h,divv,dhdt)
!****************************************************************
!                                                               *
! Compute derivative of the smoothing length, enforcing         *
! several auxilliary conditions:                                *
!                                                               *
!   1) Enforce finite range of neighbors but limit the          *
!   maximum relative variation of h, and                        *
!                                                               *
!   2) Enforce absolute limits on size of h (via hmin and hmax) *
!                                                               *
!   3) Handle specially the pathological case of a particle     *
!   getting isolated with very few neighbors. `Very few' in     *
!   this code is set to 5. Your mileage may vary. If this case  *
!   is not specially handled, then pathologies may occur. Two   *
!   possible examples of h<hmax and nneigh too small:           *
!                 1) rapidly flying through a group, perhaps    *
!                 getting 0 neighbors on the way back out       *
!                 2) floating away from the main group          *
!                                                               *
!   Our solution is to force dhdt to be positive valued and     *
!   at least 5% of the maximum allowed change per timestep.     *
!   The 5% should allow the timestep limitation from h to be    *
!   comfortably less strict than from other variables. It may   *
!   be possible to increase this 5% value to something somewhat *
!   larger, but in no case should it exceed about 25%. This     *
!   ensures that dhdt does not set the timestep (see subr       *
!   hchange). If it does, then the timestep may get shorter     *
!   each time the particle is active, and integration will      *
!   stop dead in its tracks due to this artifical `bumper'      *
!   condition, not to something physical happening. That is     *
!   bad and will make you sad and mad and will not make you     *
!   glad.                                                       *
!                                                               *
! This routine can be adapted to varying nbody smoothing        *
! lengths by changing nactive_s and iactive_s to nactive and    *
! iactive. There are physical reasons not to do this however,   *
! having to do with how and how well your simulation remains    *
! `collisionless', and what happens when it isn't.              *
!                                                               *
!****************************************************************
      integer, intent(in):: npart,npart_sph
      integer, intent(in):: nactive_s,iactive_s(npart_sph)
      integer, intent(in):: nactive_n,iactive_n(max(1,npart-npart_sph))
      integer, intent(in):: itstepbin(npart)
      double precision, intent(in):: dtincr
      double precision, intent(in):: h(npart),divv(npart_sph)
      double precision, intent(out)::dhdt(npart_sph)

      integer i,j,nbotlim,ntoplim,nrange
      double precision dhfrac,dh,dhs,dhl,buffer,bufinv

      double precision tme1,tme2
      character(len=20), parameter:: where =('hdot')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

      nrange=(neimax-neimin)/2   !gives a 10 neighbor cushion
      dhfrac=0.05d0*dhmax

!
! Cosmological expansion. As divv is in physical coordinates and h in comoving,
! we need to convert divv to comoving. For more details, see the header of 
! this file.
!
!      if(iexpand.eq.1)then
!         cosmo_a1=ascale
!      else
!         cosmo_a1=one
!      endif

!$omp parallel do default(none) 
!$omp&  firstprivate(hmin,hmax,dhfrac,nrange)
!$omp&  private(i,j,dh,dhs,dhl,ntoplim,nbotlim,buffer,bufinv)
!$omp&  shared(nneightot,nactive_s,iactive_s,itstepbin,dtincr)
!$omp&  shared(h,dhdt,divv)
      do j=1,nactive_s
         i=iactive_s(j)

         dh =diminv*h(i)*divv(i)
         dhs=-two*abs(dh)
         dhl=-dhs
         ntoplim=max(neimax-nneightot(i),-10)
         nbotlim=max(nneightot(i)-neimin,-10)

         if(ntoplim.lt.nrange)then
            buffer=exp(dble(ntoplim))
            bufinv=one/buffer
            dh=(buffer*dh+bufinv*dhs)/(buffer+bufinv)
         elseif(nbotlim.lt.nrange) then
            buffer=exp(dble(nbotlim))
            bufinv=one/buffer
            dh=(buffer*dh+bufinv*dhl)/(buffer+bufinv)
         endif
         dhdt(i)=dh

         if(nneightot(i).le.5)then
            dhdt(i)= max(dh,dhfrac*h(i)/(dtincr*dble(itstepbin(i))))
         endif

         if((hmin-h(i)).ge.zero)dhdt(i)=max(zero,dhdt(i))
         if((h(i)-hmax).ge.zero)dhdt(i)=min(zero,dhdt(i))

      enddo

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_hdot    )=timing(itme_hdot    ) + tme2-tme1
         ncalls(itme_hdot    )=ncalls(itme_hdot    ) + 1
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      if(idbg.ge.3)print *,'done ',where
      return
      end subroutine hdot
!--------------------------------------------------------------------
      subroutine viscdot(nactive_s,iactive_s,npart,npart_sph,
     &                   h,alpha,divv,alphadot)
!****************************************************************
!                                                               *
! compute viscosity change of each particle.                    *
!  Morris and Monaghan 1997 J. Comp. Phys., 136, 41             *
!  Rosswog et al. 2000 A&A 360 171-184                          *
!  Price 2004 PhD thesis.                                       *
!                                                               *
! c1 is a parameter which scales the decay length, and as MM97  *
! show includes the post shock Mach number. We allow the        *
! decay length directly using the vlength input parameter,      *
! in combination with the Mach number as defined by MM97.       *
!                                                               *
! Note that for general equations of state, where gamma is      *
! a state variable rather than constant, you need to set c1     *
! in the loop itself.                                           *
!                                                               *
! For isothermal eos's we set c1 directly to avoid messes       *
! with what gamma-1 actually ought to do.                       *
!                                                               *
! As mentioned slightly in MM97 and made more explicit in       *
! Price (2004), the source term needs to be multiplied by       *
! an eos dependent factor. That factor is:                      *
!                                                               *
! Sfac= ln{(5/3+1)/(5/3-1)]/ln[(gamma+1)/gamma-1)]              *
!                                                               *
! again, as with c1, the source factor will be a function of    *
! the equation of state, through the gamma value. If it varies  *
! then you need to redefine it below.                           *
!                                                               *
!****************************************************************
      integer, intent(in):: npart,npart_sph
      integer, intent(in):: nactive_s,iactive_s(npart_sph)
      double precision, intent(in):: h(npart),divv(npart_sph)
      double precision, intent(in):: alpha(npart_sph)
      double precision, intent(out):: alphadot(npart_sph)

      integer i,j
      double precision fbali,vsoundi,c1,Sfac,alffac
      double precision tme1,tme2
      character(len=20), parameter:: where =('viscdot')
      if(idbg.ge.2)print *,where

      if(itme.ge.1)call cpu_time(tme1)

      if(ivtime.eq.0)then

!$omp parallel do default(none) private(i,j) shared(nactive_s,iactive_s)
!$omp& shared (alphadot)
         do j=1,nactive_s
            i=iactive_s(j)
            alphadot(i)=zero
         enddo
      else

         if(ieos.eq.id_euleriso .or. ieos.eq.id_lagrgiso)then
            c1  =0.1d0
            Sfac=ONE
         else
            c1= sqrt(half*(gamma-ONE)/gamma)/vlength
            Sfac= log(atethd/twothd)/log((gamma+ONE)/(gamma-ONE)) 
         endif

!$omp parallel do default(none) private(i,j,fbali,vsoundi,alffac)
!$omp&  firstprivate(alphamax,alphastar,c1,Sfac)
!$omp&  shared(nactive_s,iactive_s)
!$omp&  shared(thermo,alpha,h,divv,alphadot)
         do j=1,nactive_s
            i=iactive_s(j)

            fbali  =thermo(i_bal  ,i)
            vsoundi=thermo(i_sound,i)
            alffac =max(alphamax-alpha(i),alphastar)
            alphadot(i)=Sfac*fbali*max(-divv(i)*alffac,zero) -
     &          c1*(max(zero,alpha(i)-alphastar))*vsoundi/h(i)
         enddo
      endif

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_alphadot)=timing(itme_alphadot) + tme2-tme1
         ncalls(itme_alphadot)=ncalls(itme_alphadot) + 1
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      if(idbg.ge.3)print *,'done ',where
      return
      end subroutine viscdot
!--------------------------------------------------------------------
      subroutine cour_ts(nactive_s,iactive_s,itstepbin,dtincr,
     &                       h,alpha, dtrat)
!****************************************************************
!                                                               *
! CFL present to next dt ratio (sph particles only)             *
!                                                               *
! Note that nactive_s could actually be nactivef_s in the       *
!  calling program. This is not a mistake. It is a swindle      *
!  to allow using one or two step integrators transparently.    *
!                                                               *
!****************************************************************

      integer, intent(in):: nactive_s, iactive_s(nactive_s)
      integer, intent(in):: itstepbin(idim)
      double precision, intent(in ):: dtincr
      double precision, intent(in ):: h(idim),alpha(isphdim)
      double precision, intent(out):: dtrat(idim)

      integer i,j
      double precision viscratio,aux1,aux2,beta,denom,dt_last
      double precision dtrat_cfl

      viscratio=betastar/alphastar

!$omp parallel do if(nactive_s.gt.1024) default(none)
!$omp&  firstprivate(viscratio,cfl)
!$omp&  shared(nactive_s,iactive_s,itstepbin,dtincr)
!$omp&  shared(dtrat,thermo,h,alpha,rmu)
!$omp&  private(i,j,aux1,aux2,beta,denom,dtrat_cfl,dt_last)
      do j=1,nactive_s
         i=iactive_s(j)
         
         aux1  = alpha(i)*thermo(i_sound,i)
         beta  = viscratio*alpha(i)
         aux2  = beta*rmu(i) !* cosmo_a1
         denom = thermo(i_sound,i) + 1.2d0*(aux1 + aux2)
         dt_last=dtincr*dble(itstepbin(i))
         dtrat_cfl = cfl*h(i)/(denom*dt_last)
         dtrat(i)=min(dtrat(i),dtrat_cfl)
      enddo

      return
      end subroutine cour_ts
!--------------------------------------------------------------------
      end module sph
