      module treesort
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!*********************************************************************
!  This file contains the VINE module to do a parallel quick sort    *
!  on a list of particles.                                           *
!                                                                    *
!  Use the parallel quicksort algorithm as discussed in              *
!     Sanders and Hansch (1997) 4th International Symposium          *
!             on Solving Irregularly Structured Problems             *
!             in Parallel. Lecture Notes in Computer Science         *
!             (LNCS), Number 1253, p13-24 Springer                   *
!                                                                    *
!  The quicksort algorithm itself was invented by C. A. R. Hoare,    *
!  who also is responsible for the following pithy quote:            *
!                                                                    *
!  "I conclude that there are two ways of constructing a software    *
!   design: One way is to make it so simple that there are           *
!   obviously no deficiencies and the other way is to make it so     *
!   complicated that there are no obvious deficiencies"              *
!                                                                    *
!  Users of this module must decide for themselves which design      *
!  strategy it follows. Unfortunately, after having written the      *
!  code and later attempted to debug it, my bet is on the second :/  *
!                                                                    *
!  Overview of the method                                            *
!     1) Divide the array into nsegments segments of equal length    *
!        where nsegments=number of cpus available                    *
!     2) Do quicksort on each segment                                *
!     3) Find npartitions-1 pivots as the median of the              *
!         (medians,quartiles,whatever) of each segment, where        *
!         npartitions can take the values 2, 3 or 4 depending        *
!         on how many cpu segments are in the current set.           *
!     4) Divide segments into npartitions partitions. Associate      *
!        one cpu from each partition with corresponding cpus in      *
!        the other sets. For example, with two partitions you        *
!        have cpu pairs.                                             *
!     5) For each association (e.g. for nsegments=8 pairs are:       *
!          1/5, 2/6 etc) do:                                         *
!          send >pivot part of lower cpu segment to upper cpu        *
!          send <pivot part of upper cpu segment to lower cpu        *
!     6) Merge together                                              *
!     7) Now we have npartitions sets of numcpu/npartitions sorted   *
!        lists. The lower sets have all elements below the           *
!        corresponding pivots. The upper sets have all elements      *
!        above the pivot.                                            *
!     8) Do steps 3--7 again for each set recursively until we have  *
!        ncpu sets of 1 cpu sorted lists. For example, for the       *
!        8cpu example in 5) we now have sets 1-4 and 5-8 cpus.       *
!        Pair 1/3 and 2/4, 5/7 and 6/8. Gives 4 sets of ncpu/4       *
!        sorted lists. Repeat with 1/2, 3/4, 5/6, 7/8. Then done.    *
!                                                                    *
!  Nomenclature                                                      *
!     Segment: a grouping of data values assigned to one cpu.        *
!     Set    : a group of segments with possible array values        *
!              in the same range (i.e. initially -inf to +inf,       *
!              thereafter -inf to pivot etc)                         *
!     Partition: A division increment of a set. One partition        * 
!              contains or will contain a group of segments to       *
!              which data from all other partitions is sent if it    *
!              satisfies the condition of being on the same side     *
!              of some pivot value. Partitions in one merge cycle    *
!              are promoted to be sets in the next cycle.            *
!     Associate: a segment that is part of a grouping of segments    *
!              that are going to exchange data.                      *
!              There is a maximum of one associate per partition.    *
!              It is possible to have one more associate than        *
!              partitions if there is an unbalanced top segment,     *
!              That would have no parter otherwise. For example,     *
!              with 7 segments, we separate into 3 partitions of     *
!              two segments each, except the first association.      *
!              Specifically, the associations are 1/4/7, 2/5 and 3/6 *
!                                                                    *
!  Implementation notes:                                             *
!     1) S&H claim that their best version (of about 70 or so)       *
!        produces a speedup of about 780/1000 on random data.        *
!        The data we expect this routine to see are very definitely  *
!        not random, so the speedup we get is much lower.            *
!                                                                    *
!     2) This version is generalized to allow for non-power of two   *
!        processor counts. This means that some processors become    *
!        idle part way through the sorting process, since they       *
!        cannot be associated with another processor(s).             *
!                                                                    *
!     3) S&H use multiple pivots, (e.g. quartiles) to partition      *
!        the data. This does not seem to buy very much for our case, *
!        presumably because the data are typically not randomly      *
!        distributed, so that quartiles produce values that are      *
!        off by quite a bit, fairly easily. At least it doesn't      *
!        appear to degrade performance much, if at all, so I've      *
!        left it on. It should be possible to change the value       *
!        of maxpartitions below, to 2 or 3 (but not >4), to          *
!        return to smaller number of pivots, but this particular     *
!        version has only been tested with maxpartitions=4.          *
!                                                                    *
!     4) All subroutines in this module are 'contain'ed in the       *
!        main, public pquick routine. This way we don't have to      *
!        give arguments to many of the calls. For reasons I don't    *
!        fully understand, the containment process appears only      *
!        to work for openmp SHARED variables on some compilers.      *
!        PRIVATE variables must still be passed directly.            *
!                                                                    *
!     5) In general, this module seems to be very tough for openmp   *
!        to get right.                                               *
!                                                                    *
!     6) The fact that lots of variables are `save'd is due to a     *
!        bug in the intel fortran compiler: it chokes on this code   *
!        if you don't have the saves in. This workaround is from     *
!        an intel support person. Someday they may even fix the      *
!        bug itself (variables go out of scope in contained routines *
!        when omp is on)                                             *
!                                                                    *
!  Future optimizations:                                             *
!    1) Do all ndim (2 or 3) sorts at one time. This will improve    *
!       parallel performance by making the effective cpu count       *
!       ndim times smaller.                                          *
!    2) Randomize order of input data. Not clear this will be        *
!       helpful. If so, it would improve somewhat both the serial    *
!       and parallel performance by keeping partitions more          *
!       equally sized.                                               *
!                                                                    *
!  Written by Andy Nelson                                            *
!*********************************************************************
      use dimens
      use constants

      implicit none

      public pquick
      private

! Set the number of maximum number of partitions for 
! the merge cycles in the parallel part of the sort.
! npartitions=2 for standard single pivot setups. 
! npartitions=3 for thirds   pivots.
! npartitions=4 for quartile pivots.
!
      integer, parameter:: maxpartitions=4

      contains
      subroutine pquick(n,idir,array,indx,iwork,arr2sort,work)
!*********************************************************************
!                                                                    *
!  Do a quicksort on variables in array `arr', defined below         *
!                                                                    *
!  Input variables:                                                  *
!                                                                    *
!  n       : length of the sorting arrays                            *
!  idir    : index of the array to be sorted (i.e. direction.x,y,z)  *
!  arr     : multidimensional array where unsorted values are found  *
!  indx    : a pointer into which the sorted indices of arr          *
!            are stored                                              *
!  arr2sort: real work array. Returns values in sorted order.        *
!  iwork   : integer work array.                                     *
!   work   : real work array.                                        *
!                                                                    *
!    arr2sort is sorted too even though it is never used (in the     *
!    caller) because the memory access speeds are better for the     *
!    continuing sorting process if the partially sorted values are   *
!    nearby in memory, as well as the indices.                       *
!                                                                    *
!  This sort can be transformed into a sort of an array              *
!  with holes by adding a `list' argument to the arrays passed       *
!  above and then replacing the first loop with:                     *
!                                                                    *
!     do i=1,n                                                       *
!        indx(i)=i                                                   *
!        arr2sort(i)=arr(idir,list(i))                               *
!     enddo                                                          *
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************
      integer         , intent(in) :: n,idir
      double precision, intent(in) :: array(ndim3,n)

      integer         , intent(out), target :: indx    (ndim*n)
      integer         , intent(out), target :: iwork   (ndim*n)
      double precision, intent(out), target :: arr2sort(ndim*n)
      double precision, intent(out), target :: work    (ndim*n)
      integer         , save, pointer, dimension(:) :: indxsrc, indxdest
      double precision, save, pointer, dimension(:) :: arrsrc , arrdest
!
! left and right limits of array elements touchable by each cpu
! i.e. the left and right limits of the set.
!
      integer, save:: lr_limits(2,maxcpus)
!
! lower and upper (1,2) limits of the elements in
! the partitioned segments of the sort array.
!
      integer, save:: partition_lims(2,maxpartitions,maxcpus)
      integer, save:: my_lims(2,maxpartitions+1)
!
! index of starting/ending destination array 
! elements into which arrays are merged
!
      integer, save:: ioffset(2,maxpartitions)
!
! array pivot values
!
      double precision, save:: pivot(maxpartitions-1)
!
! which segments in which partitions, will my_seg pass data to and from:
!
      integer, save:: my_associates(maxpartitions+1,maxcpus)

      integer, save:: npasses,nperproc,i
      integer, save:: nsegments,nsegs_per_set,nsegs_in_myset
      integer, save:: npartitions,nsegs_per_partition,my_partition
      integer, save:: nseg_l,nseg_r,iseg,my_seg
      integer, save:: mb,mt

      logical, save:: i_am_active

      character(len=20), parameter::where='pquick'
      if(idbg.ge.4)print *,where

!$omp parallel default(none) 
!$omp& shared(indx,arr2sort)
!$omp& shared(arrsrc,arrdest,indxsrc,indxdest)
!$omp& shared(idir,n,array,iwork,work,lr_limits)
!$omp& private(npasses,nsegments,nperproc,i)
!$omp& private(nsegs_per_set,nsegs_in_myset,nsegs_per_partition)
!$omp& private(npartitions,nseg_l,nseg_r)
!$omp& private(my_seg,my_partition,my_associates,i_am_active)
!$omp& private(partition_lims,my_lims,ioffset)
!$omp& private(pivot,mb,mt)

      call numthreads(nsegments)
      call get_mycpu(my_seg)

      if(nsegments.gt.maxcpus)then
         stop'pquick:maxcpus is dimensioned too small-reset larger'
      endif
!
! Do a workless run through to determine 
! how to point at the src and dest arrays 
!
      npasses=0
      nseg_l=1 ; nseg_r=nsegments
      nsegs_in_myset=nsegments ; nsegs_per_set=nsegments
      do while(nsegs_per_set.gt.1)
         call setpartitions(nseg_l,nseg_r,nsegs_in_myset,my_seg,
     &                      npartitions,nsegs_per_partition,
     &                      my_partition,my_associates)
         nseg_r=nseg_l + nsegs_per_partition*my_partition     - 1
         nseg_l=nseg_l + nsegs_per_partition*(my_partition-1)
         nsegs_in_myset = nseg_r - nseg_l + 1
         nsegs_per_set=nsegs_per_partition

         npasses=npasses+1
      enddo

!$omp single
      if(mod(npasses,2).eq.0)then
         indxsrc  => indx  ; arrsrc  => arr2sort
         indxdest => iwork ; arrdest => work
      else
         indxsrc  => iwork ; arrsrc  => work
         indxdest => indx  ; arrdest => arr2sort
      endif
!$omp end single

      nperproc=n/nsegments
      lr_limits(1,my_seg) = 1 + nperproc*(my_seg-1)
      lr_limits(2,my_seg) =     nperproc*my_seg
      if(my_seg.eq.nsegments)lr_limits(2,my_seg) = n
!
! Copy data for my_seg into sorting array
!
      do i=lr_limits(1,my_seg),lr_limits(2,my_seg)
         indxsrc(i)=i
         arrsrc (i)=array(idir,i)
      enddo
!
! Sort local data first
!
      call quick(my_seg)

!$omp barrier
!====================================================
!====================================================
! wait for everyone to finish their segment, then if
! nsegments>1, start in on a series of merge cycles.
!====================================================
!====================================================

!================================================================
! Start with one set of all segments that encompasses all the
! cpus. recursively divide each set into npartitions partitions
! each with nsegs_per_partition segments. Determine pivot 
! values that separate the data on each segment into npartitions,
! then migrate these data to the appropriate partner cpu in
! the corresponding partition. Disable the top cpus when sets 
! have a cpu count such that mod(nsegs_per_set,npartitions).ne.0
!================================================================

      nseg_l=1 ; nseg_r=nsegments

      nsegs_per_set =nsegments
      nsegs_in_myset=nsegments
      i_am_active=.true.

      do while(npasses.gt.0) 

         if(i_am_active)then
!
! Partition this set of cpus.
!
            call setpartitions(nseg_l,nseg_r,nsegs_in_myset,my_seg,
     &                         npartitions,nsegs_per_partition,
     &                         my_partition,my_associates)
!
! Get pivots for this set of cpus
!
            call getpiv(nseg_l,nseg_r,npartitions,pivot)
!
! For all segments in the source array for this set, find the
! limits of the array sections on each side of the partitioning 
! elements (partition_lims). Break out merge limits for my_seg 
! and its associates (my_lims)
!
            call findsep(my_seg,my_partition,my_associates,npartitions,
     &                   nseg_l,nseg_r,pivot,partition_lims,my_lims)
!
! Get starting/ending points for all cpus in my segment for
! the destination array. Break out merge limits for my_seg (mb,mt)
!
            call getoffset(my_seg,my_partition,my_associates,
     &                     npartitions,nsegs_per_partition,
     &                     nseg_l,nseg_r,partition_lims,
     &                     ioffset,mb,mt)
!
! Merge segments from different partitions into a segment in
! my_partition. Disable merging for unbalanced top cpus.
!
            if(my_partition.le.npartitions)then
               call mergelists(my_lims,mb,mt)
            endif

         endif  !idle cpu bypass 

!$omp barrier
!===========================================
!===========================================
! wait for everyone to finish their segment,
! then reset bookkeeping for the next round.
!===========================================
!===========================================

         if(my_partition.le.npartitions)then

            nseg_r=nseg_l + nsegs_per_partition*my_partition     - 1
            nseg_l=nseg_l + nsegs_per_partition*(my_partition-1)
            nsegs_in_myset = nseg_r - nseg_l + 1

            lr_limits(1,my_seg)=ioffset(1,my_partition)
            lr_limits(2,my_seg)=ioffset(2,my_partition)

         else
            i_am_active=.false.
            nseg_l=0 ; nseg_r=0
            lr_limits(1,my_seg)=1
            lr_limits(2,my_seg)=0
         endif

!$omp single
         nullify(indxsrc ) ;nullify(arrsrc )
         nullify(indxdest) ;nullify(arrdest)

         if(mod(npasses,2).eq.1)then
            indxsrc  => indx  ; arrsrc  => arr2sort
            indxdest => iwork ; arrdest => work
         else
            indxsrc  => iwork ; arrsrc  => work
            indxdest => indx  ; arrdest => arr2sort
         endif

!$omp end single

      npasses=npasses-1
      nsegs_per_set=nsegs_per_partition
      enddo
!$omp end parallel

      nullify(indxsrc ) ;nullify(arrsrc )
      nullify(indxdest) ;nullify(arrdest)

      return

      contains
!-----------------------------------------------------------------------
      subroutine quick(my_seg)

      integer, intent(in) :: my_seg

      integer, parameter::minsubarray=5,maxstack=500
      integer istack,lr(2,maxstack)
      integer i1,i2,i3,ipivot
      integer i,j,il,ir,itmp,indxtest
      double precision piv,stmp,testval

      istack=1
      lr(1,istack)=lr_limits(1,my_seg)
      lr(2,istack)=lr_limits(2,my_seg)

      do while(istack.gt.0)

         if(lr(2,istack)-lr(1,istack) .gt. minsubarray)then

! quicksort iteration, with median of 3 pivot selection

            i1=lr(1,istack)
            i2=lr(2,istack)
            i3=(i1+i2)/2

            if(arrsrc(i1).gt.arrsrc(i2))then
               if(arrsrc(i1).gt.arrsrc(i3))then
                  if(arrsrc(i2).gt.arrsrc(i3))then
                    ipivot=i2                  ! i3 < i2 < i1
                  else
                    ipivot=i3                  ! i2 < i3 < i1
                  endif
               else
                  ipivot=i1                    ! i2 < i1 < i3
               endif
            else
               if(arrsrc(i1).lt.arrsrc(i3))then
                  if(arrsrc(i2).lt.arrsrc(i3))then
                     ipivot=i2                  ! i1 < i2 < i3
                  else
                     ipivot=i3                  ! i1 < i3 < i2
                  endif
               else
                  ipivot=i1                     ! i3 < i1 < i2
               endif
            endif
            piv=arrsrc(ipivot)

            il=lr(1,istack)-1
            ir=lr(2,istack)+1

            do while(il.lt.ir)

10             continue
                  il=il+1
               if(arrsrc(il).lt.piv)goto 10
20             continue
                  ir=ir-1
               if(arrsrc(ir).gt.piv)goto 20

               if(il.lt.ir)then
                  itmp=indxsrc(ir)        ; stmp=arrsrc(ir)
                  indxsrc(ir)=indxsrc(il) ; arrsrc(ir)=arrsrc(il)
                  indxsrc(il)=itmp        ; arrsrc(il)=stmp
               endif
            enddo
            if(ir.eq.il)il=il+1

            lr(1,istack  )=lr(1,istack)
            lr(2,istack+1)=lr(2,istack)
            lr(2,istack  )=ir
            lr(1,istack+1)=il
            istack=istack+1

         else                               ! insertion sort
!
! Insertion sort. Assume leftmost is in correct position 
! (the bottom) then put array elements on top of it and let 
! them sink downwards to proper position (by consecutively 
! switching elements that are out of position all the way
! down) in the partially sorted list. Do again for 3rd,
! 4th, Nth element, until done.
!
            do i=lr(1,istack)+1,lr(2,istack)
               indxtest=indxsrc(i)
               testval =arrsrc (i)
               do j=i-1,lr(1,istack),-1
                  if(testval .ge. arrsrc(j))exit
                  indxsrc(j+1)=indxsrc(j)
                  arrsrc (j+1)=arrsrc (j)
               enddo
               indxsrc(j+1)=indxtest
               arrsrc (j+1)=testval
            enddo

            istack=istack-1

         endif  !qsort/insertion sort
      enddo

      return
      end subroutine quick
!-----------------------------------------------------------------------
      subroutine setpartitions(nseg_l,nseg_r,nsegs_in_myset,my_seg,
     &                         npartitions,nsegs_per_partition,
     &                         my_partition,my_associates)
!
! Set the number of partitions, find out which one I'm in,
! who my associates are, and disable my attempt to merge
! with others if I'm above the top partition because my 
! data gets handled by lower partition members.
!
      integer, intent(in):: nseg_l,nseg_r,my_seg,nsegs_in_myset

      integer, intent(out)::npartitions,nsegs_per_partition
      integer, intent(out)::my_partition
      integer, intent(out)::my_associates(maxpartitions+1,maxcpus)

      integer iseg0,iseg,np,n_associates
!
! Set the large number of partitions possible, 
! subject to conditions 
!     1) there are no more than maxpartitions
!     2) there are few enough partitions to associate
!        all segments with at least one other
!
      npartitions=min(nsegs_in_myset,maxpartitions)
      nsegs_per_partition=nsegs_in_myset/npartitions

      if(nsegs_in_myset.gt.npartitions)then

         if(nsegs_per_partition*npartitions.lt.nsegs_in_myset)then
            npartitions=npartitions-1
            nsegs_per_partition=nsegs_in_myset/npartitions
         endif

         do while(2*npartitions.gt.nsegs_in_myset)
            npartitions=npartitions-1
            nsegs_per_partition=nsegs_in_myset/npartitions
         enddo
      endif
!
! Determine my partition
!
      my_partition=1
      do while(my_seg.ge.nseg_l+my_partition*nsegs_per_partition)
         my_partition=my_partition+1
      enddo
!
! Determine the associations between all cpus in all partitions
!
      do iseg=nseg_l,nseg_r
         iseg0=iseg
         do while(iseg0.ge.nseg_l)
            iseg0=iseg0-nsegs_per_partition
         enddo
         iseg0=iseg0+nsegs_per_partition

         do np=1,npartitions+1
            my_associates(np,iseg)=iseg0 + (np-1)*nsegs_per_partition
         enddo
      enddo

      return
      end subroutine setpartitions
!-----------------------------------------------------------------------
      subroutine getpiv(nseg_l,nseg_r,npartitions,pivot)
!
! Get pivot candidates, then sort them using an insertion sort 
! making sure not to look for candidates in segments with no data 
!
      integer         , intent(in) :: nseg_l,nseg_r,npartitions
      double precision, intent(out):: pivot(maxpartitions-1)

      integer np,iseg,ileft,irght,num,i2
      integer i,j,ilow,med

      double precision testval
      double precision pivarray(maxcpus,maxpartitions-1)

      do np=1,npartitions-1
         do iseg=nseg_l,nseg_r
            ileft=lr_limits(1,iseg)  ; irght=lr_limits(2,iseg)
            num=irght-ileft+1
            i2 =min(irght,ileft+ (np*num)/npartitions) 
            pivarray(iseg,np)= -huge(1d0)
            if(num.gt.0)pivarray(iseg,np)= arrsrc(i2) 
         enddo
      enddo

      pivot=0d0

      do np=1,npartitions-1
         do i=nseg_l+1,nseg_r
            testval=pivarray(i,np)
            do j=i-1,nseg_l,-1
               if(testval .ge. pivarray(j,np))exit
               pivarray(j+1,np)=pivarray(j,np)
            enddo
            pivarray(j+1,np)=testval
         enddo
         ilow=nseg_l
         do i=nseg_r,nseg_l,-1
            if(pivarray(i,np).gt.-half*huge(1d0))ilow=i
         enddo
         med=min(nseg_r,ilow + (nseg_r-ilow)/2)

         pivot(np)=pivarray(med,np)
      enddo

      return
      end subroutine getpiv
!-----------------------------------------------------------------------
      subroutine findsep(my_seg,my_partition,my_associates,npartitions,
     &                   nseg_l,nseg_r,pivot,partition_lims,my_lims)
!
! Find separator elements for each segment and assign limits
! to each component of the pair. Need lower and upper (1,2)
! bounds for components to be sent to each partition.
! Break out specific merge limits for my_seg. (my_lims)
!
      integer, intent(in) :: my_seg,my_partition,npartitions
      integer, intent(in) :: my_associates(maxpartitions+1,maxcpus)
      integer, intent(in) :: nseg_l,nseg_r
      double precision, intent(in) :: pivot(maxpartitions-1)

      integer, intent(out):: partition_lims(2,maxpartitions,maxcpus)
      integer, intent(out):: my_lims(2,maxpartitions+1)

      integer iseg,np,num,ilbnd,irbnd,idiff,mid

      do iseg=nseg_l,nseg_r

         num=lr_limits(2,iseg)-lr_limits(1,iseg)+1

         if(num.eq.0)then
!
! segment has no elements, so use full limits which are set to do nothing
!
            do np=1,npartitions
               partition_lims(1,np,iseg)=lr_limits(1,iseg)
               partition_lims(2,np,iseg)=lr_limits(2,iseg)
            enddo

         else
!
! segment has elements, so now find separators between partitions
!
            partition_lims(1,1          ,iseg)=lr_limits(1,iseg)    !extremes
            partition_lims(2,npartitions,iseg)=lr_limits(2,iseg)

            do np=1,npartitions-1

               ilbnd=lr_limits(1,iseg)
               irbnd=lr_limits(2,iseg)

               if    (arrsrc(irbnd).lt.pivot(np))then
!
! all elements .lt. pivot, so set limit to right bound
!
                  partition_lims(2,np  ,iseg)=irbnd
                  partition_lims(1,np+1,iseg)=irbnd+1

               elseif(arrsrc(ilbnd).ge.pivot(np))then
!
! all elements .gt. pivot so set limit to left bound
!
                  partition_lims(2,np  ,iseg)=ilbnd-1
                  partition_lims(1,np+1,iseg)=ilbnd  

               else
!
! bisection search for separating element
!
                  idiff=irbnd-ilbnd
                  mid=ilbnd+idiff/2
                  do while(idiff.gt.0)
                     if(arrsrc(mid).gt.pivot(np))then
                        irbnd=mid
                     else
                        ilbnd=mid
                     endif
                     idiff=(irbnd-ilbnd)/2
                     mid=ilbnd+idiff
                  enddo
                  if(arrsrc(mid).lt.pivot(np))mid=irbnd
                  partition_lims(2,np  ,iseg)=mid-1
                  partition_lims(1,np+1,iseg)=mid

               endif
            enddo
         endif

      enddo
!
! Merge limits for my_seg, including my associates
!
      my_lims(1,:)=1 ; my_lims(2,:)=0
      if(my_partition.le.npartitions)then

         do np=1,npartitions+1
            iseg=my_associates(np,my_seg)
            if(iseg.le.nseg_r)then
               my_lims(1,np)=partition_lims(1,my_partition,iseg) 
               my_lims(2,np)=partition_lims(2,my_partition,iseg) 
            endif
         enddo

      endif

      return
      end subroutine findsep
!-----------------------------------------------------------------------
      subroutine getoffset(my_seg,my_partition,my_associates,
     &                     npartitions,nsegs_per_partition,
     &                     nseg_l,nseg_r,partition_lims,
     &                     ioffset,mb,mt)
!
! Find offset locations for beginning of each cpu's new array segment
! in the destination arrays. (ioffset). Break out for my_seg (mb,mt).
!
      integer, intent(in) :: my_seg,my_partition,npartitions
      integer, intent(in) :: my_associates(maxpartitions+1,maxcpus)
      integer, intent(in) :: nsegs_per_partition,nseg_l,nseg_r
      integer, intent(in) :: partition_lims(2,maxpartitions,maxcpus)
      integer, intent(out):: ioffset(2,maxpartitions),mb,mt

      integer ip,np,num,nsofar,iseg,jseg

      nsofar=0
      do np=1,npartitions
!
! Offset from all lower segments of current partition
!
         do iseg=nseg_l,my_associates(1,my_seg)-1
            do ip=1,npartitions+1
               jseg=my_associates(ip,iseg)
               if(jseg.le.nseg_r)then
                  num=partition_lims(2,np,jseg) -
     &                       partition_lims(1,np,jseg)+1
                  nsofar=nsofar+num
               endif
            enddo
         enddo
!
! Offsets for this segment of current partition
!
         ioffset(1,np)=lr_limits(1,nseg_l) + nsofar
         do ip=1,npartitions+1
            iseg=my_associates(ip,my_seg)
            if(iseg.le.nseg_r)then
               num=partition_lims(2,np,iseg)-partition_lims(1,np,iseg)+1
               nsofar=nsofar+num
            endif
         enddo
         ioffset(2,np)=lr_limits(1,nseg_l) + nsofar - 1
!
! Offset from all higher segments of current partition. Need
! this to get initial offset for the next partition.
!
         do iseg=my_associates(1,my_seg)+1,nseg_l+nsegs_per_partition-1
            do ip=1,npartitions+1
               jseg=my_associates(ip,iseg)
               if(jseg.le.nseg_r)then
                  num=partition_lims(2,np,jseg) -
     &                       partition_lims(1,np,jseg)+1
                  nsofar=nsofar+num
               endif
            enddo
         enddo

      enddo
!
! Destination array limits:
!
      mb=1 ; mt=0
      if(my_partition.le.npartitions)then
         mb=ioffset(1,my_partition)
         mt=ioffset(2,my_partition)
      endif

      return
      end subroutine getoffset
!-----------------------------------------------------------------------
      subroutine mergelists(my_lims,mb,mt)

      integer, intent(in) :: my_lims(2,maxpartitions+1)
      integer, intent(in) :: mb,mt

      integer ib,it,jb,jt,kb,kt,lb,lt


      ib=my_lims(1,1) ; it=my_lims(2,1)
      jb=my_lims(1,2) ; jt=my_lims(2,2)
      kb=my_lims(1,3) ; kt=my_lims(2,3)
      lb=my_lims(1,4) ; lt=my_lims(2,4)

      if(lt.ge.lb)then
         call merge4(ib,it,jb,jt,kb,kt,lb,lt,mb,mt)
      else
         if(kt.ge.kb)then
            call merge3(ib,it,jb,jt,kb,kt,mb,mt)
         else
            call merge2(ib,it,jb,jt,mb,mt)
         endif
      endif

      return
      end subroutine mergelists
!-----------------------------------------------------------------------
      subroutine merge4(ib,it,jb,jt,kb,kt,lb,lt,mb,mt)
!
! merge four sorted subarrays into output arrays
!
!     ib,it : Top and bottom indices of first  segment of source array
!     jb,jt : Top and bottom indices of second segment of source array
!     kb,kt : Top and bottom indices of third  segment of source array
!     lb,lt : Top and bottom indices of fourth segment of source array
!     mb,mt : Top and bottom indices of destination arrays

      integer, intent(in) :: ib,jb,kb,lb,mb
      integer, intent(in) :: it,jt,kt,lt,mt

      integer i,j,k,l,m,iminloc

      double precision try(4)

      try(1)=arrsrc(ib) ; try(2)=arrsrc(jb)
      try(3)=arrsrc(kb) ; try(4)=arrsrc(lb)

      i=ib ; j=jb ; k=kb ; l=lb ; m=mb
!
! all four arrays active
!
      do while(i.le.it .and. j.le.jt .and. k.le.kt .and. l.le.lt)

         iminloc=minloc(array=try,dim=1)

         select case(iminloc)
         case(1)
            indxdest(m)=indxsrc(i) ; arrdest (m)=arrsrc (i)
            i=i+1
            try(1)=arrsrc(min(i,it))
         case(2)
            indxdest(m)=indxsrc(j) ; arrdest (m)=arrsrc (j)
            j=j+1
            try(2)=arrsrc(min(j,jt))
         case(3)
            indxdest(m)=indxsrc(k) ; arrdest (m)=arrsrc (k)
            k=k+1
            try(3)=arrsrc(min(k,kt))
         case(4)
            indxdest(m)=indxsrc(l) ; arrdest (m)=arrsrc (l)
            l=l+1
            try(4)=arrsrc(min(l,lt))
         end select
         m=m+1
      enddo
!
! Finished one array. Three left.
!
      if(i.gt.it)then
         call merge3(j,jt,k,kt,l,lt,m,mt)
      elseif(j.gt.jt)then
         call merge3(i,it,k,kt,l,lt,m,mt)
      elseif(k.gt.kt)then
         call merge3(i,it,j,jt,l,lt,m,mt)
      else
         call merge3(i,it,j,jt,k,kt,m,mt)
      endif

      return
      end subroutine merge4
!-----------------------------------------------------------------------
      subroutine merge3(ib,it,jb,jt,kb,kt,mb,mt)
!
! merge three sorted subarrays into output arrays
!
!     ib,it : Top and bottom indices of first  segment of source array
!     jb,jt : Top and bottom indices of second segment of source array
!     kb,kt : Top and bottom indices of third  segment of source array
!     mb,mt : Top and bottom indices of destination arrays
!
      integer, intent(in) :: ib,jb,kb,mb
      integer, intent(in) :: it,jt,kt,mt

      integer i,j,k,m,iminloc

      double precision try(3)

      try(1)=arrsrc(ib)
      try(2)=arrsrc(jb)
      try(3)=arrsrc(kb)

      i=ib ; j=jb ; k=kb ; m=mb
!
! all three arrays active
!
      do while(i.le.it .and. j.le.jt .and. k.le.kt)

         iminloc=minloc(array=try,dim=1)

         select case(iminloc)
         case(1)
            indxdest(m)=indxsrc(i) ; arrdest (m)=arrsrc (i)
            i=i+1
            try(1)=arrsrc(min(i,it))
         case(2)
            indxdest(m)=indxsrc(j) ; arrdest (m)=arrsrc (j)
            j=j+1
            try(2)=arrsrc(min(j,jt))
         case(3)
            indxdest(m)=indxsrc(k) ; arrdest (m)=arrsrc (k)
            k=k+1
            try(3)=arrsrc(min(k,kt))
         end select
         m=m+1
      enddo
!
! finished one array. Two left.
!
      if(i.gt.it)then
         call merge2(j,jt,k,kt,m,mt)
      elseif(j.gt.jt)then
         call merge2(i,it,k,kt,m,mt)
      else
         call merge2(i,it,j,jt,m,mt)
      endif

      return
      end subroutine merge3
!-----------------------------------------------------------------------
      subroutine merge2(ib,it,jb,jt,mb,mt)
!
! merge two sorted segments of source arrays into output arrays
!
!     ib,it : Top and bottom indices of first  segment of source array
!     jb,jt : Top and bottom indices of second segment of source array
!     mb,mt : Top and bottom indices of destination arrays
!
      integer, intent(in) :: ib,it,jb,jt,mb,mt

      integer i,j,m,ii,jj

      i=ib ; j=jb ; m=mb 

      do while(i.le.it .and. j.le.jt)
         if(arrsrc(i).lt.arrsrc(j))then
            indxdest(m)=indxsrc(i) ; arrdest (m)=arrsrc (i)
            i=i+1
         else
            indxdest(m)=indxsrc(j) ; arrdest (m)=arrsrc (j)
            j=j+1
         endif
         m=m+1
      enddo

      if(i.le.it)then
         do ii=i,it
            indxdest(m)=indxsrc(ii) ; arrdest (m)=arrsrc (ii)
            m=m+1
         enddo
      else
         do jj=j,jt
            indxdest(m)=indxsrc(jj) ; arrdest (m)=arrsrc (jj)
            m=m+1
         enddo
      endif

      return
      end subroutine merge2
!-----------------------------------------------------------------------
      end subroutine pquick
!-----------------------------------------------------------------------
      end module treesort
