      module vineutils
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!****************************************************************
!                                                               *
!  This file contains the VINE module contains some useful      * 
!  utilities which, while they are not required for correct     *
!  operation of the code, make things go faster and avoid       *
!  some problems that particles simulations may develop         *
!  when they are evolved for soem time.                         *
!                                                               *
!  The routines contained here are:                             *
!                                                               *
!    grabmem       : Grab the best memory for the tree. See     *
!                    comments in the header of that routine.    *
!    orderparticles: reorder the particles for speed            *
!                    for speed and parallel efficiency          *
!                    of the code as it runs.                    *
!    restoreorder  : restore the original order. Not well tested*
!    findmult      : Finds multiple particles in the particle   *
!                    distribution, defined as particles         *
!                    that have identical positions and          *
!                    velocities.                                *
!                                                               *
!  Written by Andy Nelson                                       *
!****************************************************************
      use dimens
      use partvars
      use integvars
      use indivtsvars
      use miscvars, only: igrape,isoft,eps,sepmax,bsepmax,boxlims
      use miscvars, only: ibctype,imac,igrav,hclmp_is_exact
      use miscvars, only: maxmult,nmulttot,treeacc
      use miscvars, only: vstep,vkernmax,dvkern,kern_incr,neigdistrib
      use sphvars
      use clmpvars
      use treevars
      use tree    , only: getallparticles,getparticles
      use tree    , only: initgetneigh,getallneigh
      use treemake, only: mtree,rtree,resethnode
      use stubs
      use perform
      use constants

#if SELFFORCE
      use forcelaw, only: initpole
#endif
#if GRAPE==6
      use grape
      implicit none
#else
      implicit none
      double precision g6_oldacc,g6_oldjrk,g6_oldphi
#endif

      private
      public grabmem,orderparticles,restoreorder,findmult


      contains
!----------------------------------------------------------------------
      subroutine grabmem(npart)
!***********************************************************************
!                                                                      *
! Subroutine to allocate memory for most arrays. This is a good thing  *
!  to do for several reasons:                                          *
!     o on some machines (like SGI where it was originally written),   *
!        memory allocation proceeds page by page on an as needed       *
!        basis and is rather slow. Lets get it done all at once.       *
!     o What is more important is that the page sizes that you get     *
!        are different, and for this code, page size matters a *LOT*,  *
!        with large pages being much faster (factors of more than 3)   *
!        than small page sizes, though there are now a number of       *
!        optimizations that mitigate this. Therefore, if we allocate   *
!        the memory first that, when used, gives code that is the      *
!        most sensitive to page size, then we will be most likely to   *
!        get large pages on this memory, and the code will run fast.   *
!     o Also very important on large machines that have NUMA           *
!        (Non Uniform Memory Access), is getting the arrays distributed*
!        across all of the nodes in some sensible fashion. Here we     *
!        do a parallel loop, so that a very simple first touch policy  *
!        will get a fairly uniform distribution.                       *
!                                                                      *
!                                                                      *
! Implementation note: Someday, if and when it becomes advisable to    *
!     redo the array stuff in vine to be all `allocatable' rather than *
!     set by the idim parameter at compile time, this routine would    *
!     be the obvious place to do all of the allocations.               *
!                                                                      *
!                                                                      *
!  Subroutine by A. Nelson                                             *
!                                                                      *
!***********************************************************************
      integer, intent(in):: npart
      integer i

      print *,'Grabbing the best memory...'


!$omp parallel default(none)
!$omp& private(i)
!$omp& shared(npart)
!$omp& shared(link,treept,lstbunches,amag,clmpsize0,icount)
!$omp& shared(lstclumps,linksphtotree,linktreetosph,next)
!
! Variables used very often in the tree traversal and build
!
!$omp do
      do i=1,2*npart
         link(idau,i)=0       ; link(isib,i)=0
         linksphtotree(i)=0   ; linktreetosph(i)=0   !important for
         next         (i)=0                          !treebuild
         treept(ims,i)=zero
         treept(1  ,i)=zero
         treept(2  ,i)=zero
#if NDIM==3
         treept(3  ,i)=zero
#endif
      enddo
!$omp enddo nowait

!$omp do
      do i=1,npart
         amag      (i)=zero
         clmpsize0 (i)=zero
         lstclumps (i)=0
         lstbunches(i)=0
      enddo
!$omp enddo nowait
!$omp do 
      do i=1,ndim*npart
         icount(i)=0
      enddo
!$omp end parallel

      if(igrav.eq.1)call initpole(npart)

      return
!
! The below initializations don't appear to produce
! any benefit and often instead decrease performance
! somewhat. Therefore they are disabled. Your mileage
! may vary.
!

!$omp parallel default(none)
!$omp& private(i)
!$omp& shared(npart,npart_sph)
!$omp& shared(ipindx,posmas,dumposmas,vel,dumvel,acc1,acc2,poten)
!$omp& shared(u,h,alpha,rho,divv,thermo,dumu,dumh,dumalf,oldpot)
!$omp& shared(dudt1,dhdt1,dalfdt1,dudt2,dhdt2,dalfdt2,velhaf)
!$omp& shared(rhosave,divvsave,rhohalf,divvhalf,material)
!$omp& shared(iactive,iactive_s,iactive_n,it0,it1)
!$omp& shared(iactive_c,itstepbin,deltat,dtrat)

!
! Variables used very often the SPH calculations
!
!$omp do
      do i=1,npart_sph
         u             (i)=zero
         h             (i)=zero
         alpha         (i)=zero
         rho           (i)=zero
         divv          (i)=zero
         thermo(i_bal  ,i)=zero
         thermo(i_pres ,i)=zero
         thermo(i_sound,i)=zero
      enddo
!$omp enddo nowait
!$omp do
      do i=1,npart
         ipindx    (i)=0
         poten (    i)=zero
         posmas(ims,i)=zero
         posmas(1  ,i)=zero
         posmas(2  ,i)=zero
         vel   (1  ,i)=zero
         vel   (2  ,i)=zero
#if NDIM==3
         posmas(3  ,i)=zero
         vel   (3  ,i)=zero
#endif
      enddo
!$omp enddo nowait
!
! Variables used in updates, both of tree info and of 
! time updatesin the integrator
!
!$omp do
      do i=1,npart_sph
         dumu  (i)=zero
         dumh  (i)=zero
         dumalf(i)=zero
      enddo
!$omp enddo nowait
!$omp do
      do i=1,npart
         iactive  (i)=0
         iactive_s(i)=0
         iactive_c(i)=0
         itstepbin(i)=0
         deltat   (i)=zero
         dumposmas(ims,i)=zero
         dumposmas(1  ,i)=zero
         dumposmas(2  ,i)=zero
         dumvel   (1  ,i)=zero
         dumvel   (2  ,i)=zero
#if NDIM==3
         dumposmas(3  ,i)=zero
         dumvel   (3  ,i)=zero
#endif
      enddo
!$omp enddo nowait

!$omp do
      do i=1,npart
         acc1 (1  ,i)=zero
         acc1 (2  ,i)=zero
#if NDIM==3
         acc1 (3  ,i)=zero
#endif
#ifdef RK
         velhaf(1,i)=zero
         velhaf(2,i)=zero
         acc2  (1,i)=zero
         acc2  (2,i)=zero
#if NDIM==3
         acc2  (3,i)=zero
         velhaf(3,i)=zero
#endif
#endif
      enddo
!$omp enddo nowait
!$omp do
      do i=1,npart_sph
         dudt1  (i)=zero
         dhdt1  (i)=zero
         dalfdt1(i)=zero
#ifdef RK
         dudt2  (i)=zero
         dhdt2  (i)=zero
         dalfdt2(i)=zero
#endif
       enddo
!$omp enddo nowait

!
! Miscellaneous stuff
!
!$omp do
      do i=1,npart
         iactive_n(i)=0   !not used in current code
         it0      (i)=0
         it1      (i)=0
         dtrat    (i)=zero
      enddo
!$omp do
      do i=1,npart_sph
         material(i)=0
         oldpot  (i)=zero
         rhosave (i)=zero
         divvsave(i)=zero
#ifdef RK
         rhohalf (i)=zero
         divvhalf(i)=zero
#endif
      enddo
!$omp end parallel

      return
      end subroutine grabmem
!----------------------------------------------------------------------
      subroutine orderparticles
!*********************************************************************
!                                                                    *
! Reorder the particles in the main code, using the current          *
! tree ordering. This technique speeds up the tree build by a        *
! factor of about 2.5-3 or so on machines where pagesize is not      *
! large (about 10-20% where it is large), and probably other parts   *
! of the main code as well, though that has not been specifically    *
! tested. In addition, it speeds up the SPH calculations and makes   *
! them more parallel at high processor counts. Due to better         *
! localization of memory accesses among processors.                  *
!                                                                    *
! This routine does not need to be called terribly often, or even    *
! at all, though speed will suffer some if you never call it.        *
!                                                                    *
! Note: A call to mtree concludes this routine because               *
! linksphtotree/treetosph get scrambled by the reordering and        *
! some of the other tree arrays are used as dummy temp storage       *
! here.                                                              *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************

      integer i,j,jj,kk,it,ntot,nact
      logical, save :: first = .true.

      double precision tme1,tme2
      character(len=20), parameter:: where=('orderparticles')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)
!
! determine tree ordering in sph counting. Use lstclumps
!  as a work array for storing the ordered list of particles
!  in the tree
!
      call getallparticles(nroot,npart,ntot,nact,lstclumps)
      jj=0
      kk=npart_sph
      do j=1,npart
         it=linktreetosph(lstclumps(j))
         if(it.le.npart_sph)then
            jj=jj+1
            isort(jj)=it
         else
            kk=kk+1
            isort(kk)=it
         endif
      enddo

!$omp parallel default(none) 
!$omp& shared(first)
!$omp& shared(npart,npart_sph,isort,temparr,itemparr)
!$omp& shared(ipindx,itstepbin,it0,it1,material)
!$omp& shared(posmas,vel,h,u,alpha,poten,amag)
!$omp& shared(oldpot)
!$omp& shared(deltat,acc1,dhdt1,dalfdt1,dudt1)
!$omp& shared(rho,divv,rhosave,divvsave,thermo)
!$omp& shared(velhaf,rhohalf,divvhalf,acc2,dhdt2,dudt2,dalfdt2)
!$omp& shared(g6_oldphi,g6_oldacc,g6_oldjrk)
!$omp& private(j,i) 
!
!======first reordering conversion=======
!
!$omp do
      do j=1,npart
         itemparr(1,j)=ipindx(j)
         itemparr(2,j)=it0(j)
         itemparr(3,j)=it1(j)
         itemparr(4,j)=itstepbin(j)
      enddo
!$omp do
      do j=1,npart
         i=isort(j)
         ipindx   (j)=itemparr(1,i)
         it0      (j)=itemparr(2,i)
         it1      (j)=itemparr(3,i)
         itstepbin(j)=itemparr(4,i)
      enddo
!
!======next reordering conversion=======
!
!$omp do
      do j=1,npart
         temparr(ndim2,j)=h           (j) 
         temparr(ndim1,j)=posmas(ndim1,j)
         temparr(1    ,j)=posmas(1    ,j)
         temparr(2    ,j)=posmas(2    ,j)
#if NDIM==3
         temparr(3    ,j)=posmas(3    ,j)
#endif
      enddo
!$omp do
      do j=1,npart
         i=isort(j)
         h           (j)=temparr(ndim2,i)
         posmas(ndim1,j)=temparr(ndim1,i)
         posmas(1    ,j)=temparr(1    ,i)
         posmas(2    ,j)=temparr(2    ,i)
#if NDIM==3
         posmas(3    ,j)=temparr(3    ,i)
#endif
      enddo
!
!======next reordering conversion=======
!
!$omp do
      do j=1,npart
         temparr(ndim2,j)=amag  (j)
         temparr(ndim1,j)=poten (j)
         temparr(1    ,j)=vel (1,j)
         temparr(2    ,j)=vel (2,j)
#if NDIM==3
         temparr(3    ,j)=vel (3,j)
#endif
      enddo

!$omp do
      do j=1,npart
         i=isort(j)
         amag  (j)=temparr(ndim2,i)
         poten (j)=temparr(ndim1,i)
         vel (1,j)=temparr(1    ,i)
         vel (2,j)=temparr(2    ,i)
#if NDIM==3
         vel (3,j)=temparr(3    ,i)
#endif
      enddo
!
!======next reordering conversion=======
!
!$omp do
      do j=1,npart_sph
         itemparr(1,i)=material(j)
         temparr (1,j)=u    (j)
         temparr (2,j)=alpha(j)
      enddo

!$omp do
      do j=1,npart_sph
         i=isort(j)
         material(j)=itemparr(1,i)
         u       (j)=temparr (1,i)
         alpha   (j)=temparr (2,i)
      enddo

!================================================
! The variables below have not been defined yet
! if we are reordering for the first time.
! (Assuming the first call of this routine is in
! the `first' region of the onestep or twostep
! routines)
! We have to do this really huge if block because
! we are in the middle of a parallel region here
! and can't just jump out of it arbitrarily.
!================================================

      if(.not.first)then   !(skip the rest of the reorderings)
!
!======next reordering conversion=======
!
!$omp do
         do j=1,npart
            temparr(ndim2,j)=oldpot (j)
            temparr(ndim1,j)=deltat (j)
            temparr(1    ,j)=acc1 (1,j)
            temparr(2    ,j)=acc1 (2,j)
#if NDIM==3
            temparr(3    ,j)=acc1 (3,j)
#endif
         enddo

!$omp do
         do j=1,npart
            i=isort(j)
            oldpot (j)=temparr(ndim2,i)
            deltat (j)=temparr(ndim1,i)
            acc1 (1,j)=temparr(1    ,i)
            acc1 (2,j)=temparr(2    ,i)
#if NDIM==3
            acc1 (3,j)=temparr(3    ,i)
#endif
         enddo
!
!======next reordering conversion=======
!
#if GRAPE==6
!$omp do
         do j=1,npart
            temparr(ndim2,j)=g6_oldphi(  j)
            temparr(1    ,j)=g6_oldacc(1,j)
            temparr(2    ,j)=g6_oldacc(2,j)
#if NDIM==3
            temparr(3    ,j)=g6_oldacc(3,j)
#endif
         enddo

!$omp do
         do j=1,npart
            i=isort(j)
            g6_oldphi(  j)=temparr(ndim2,i)
            g6_oldacc(1,j)=temparr(1    ,i)
            g6_oldacc(2,j)=temparr(2    ,i)
#if NDIM==3
            g6_oldacc(3,j)=temparr(3    ,i)
#endif
         enddo
!
!======next reordering conversion=======
!
!$omp do
         do j=1,npart
            temparr(1,j)=g6_oldjrk(1,j)
            temparr(2,j)=g6_oldjrk(2,j)
#if NDIM==3
            temparr(3,j)=g6_oldjrk(3,j)
#endif
         enddo

!$omp do
         do j=1,npart
            i=isort(j)
            g6_oldjrk(1,j)=temparr(1,i)
            g6_oldjrk(2,j)=temparr(2,i)
#if NDIM==3
            g6_oldjrk(3,j)=temparr(3,i)
#endif
         enddo
#endif 
!GRAPE
!
!======next reordering conversion=======
!
!$omp do
         do j=1,npart_sph
            temparr(1    ,j)=thermo (i_bal,j)  !need to remember inactives
            temparr(2    ,j)=dalfdt1(      j)
            temparr(ndim1,j)=dudt1  (      j)
            temparr(ndim2,j)=dhdt1  (      j)
         enddo
!$omp do
         do j=1,npart_sph
            i=isort(j)
            thermo (i_bal,j)=temparr(1    ,i)
            dalfdt1(      j)=temparr(2    ,i)
            dudt1  (      j)=temparr(ndim1,i)
            dhdt1  (      j)=temparr(ndim2,i)
         enddo
!
!======next reordering conversion=======
!
!$omp do
         do j=1,npart_sph
            temparr(1    ,j)=divv    (j)
            temparr(2    ,j)=divvsave(j)
            temparr(ndim1,j)=rho     (j)
            temparr(ndim2,j)=rhosave (j)
         enddo
!$omp do
         do j=1,npart_sph
            i=isort(j)
            divv    (j)=temparr(1    ,i)
            divvsave(j)=temparr(2    ,i)
            rho     (j)=temparr(ndim1,i)
            rhosave (j)=temparr(ndim2,i)
         enddo
!
!======next reordering conversion=======
!
#if RK
!$omp do
         do j=1,npart
            temparr(1    ,j)=velhaf(1,j)
            temparr(2    ,j)=velhaf(2,j)
            temparr(ndim1,j)=acc2  (1,j)
            temparr(ndim2,j)=acc2  (2,j)
#if NDIM==3
            temparr(3    ,j)=velhaf(3,j)
            temparr(ndim3,j)=acc2  (3,j)
#endif
          enddo
!$omp do
         do j=1,npart
            i=isort(j)
            velhaf(1,j)=temparr(1    ,i)
            velhaf(2,j)=temparr(2    ,i)
            acc2  (1,j)=temparr(ndim1,i)
            acc2  (2,j)=temparr(ndim2,i)
#if NDIM==3
            acc2  (3,j)=temparr(ndim3,i)
            velhaf(3,j)=temparr(3    ,i)
#endif
         enddo
!
!======next reordering conversion=======
!
!$omp do
         do j=1,npart_sph
            temparr(1    ,j)=rhohalf (j)
            temparr(2    ,j)=divvhalf(j)
            temparr(ndim1,j)=dhdt2   (j)
            temparr(ndim2,j)=dudt2   (j)
            temparr(ndim3,j)=dalfdt2 (j)
         enddo
!$omp do
         do j=1,npart_sph
            i=isort(j)
            rhohalf (j)=temparr(1    ,i)
            divvhalf(j)=temparr(2    ,i)
            dhdt2   (j)=temparr(ndim1,i)
            dudt2   (j)=temparr(ndim2,i)
            dalfdt2 (j)=temparr(ndim3,i)
         enddo
#endif
!RK

      endif    !first time

!$omp end parallel

      if(itme.ge.1)then
         call cpu_time(tme2)
         if(itme.ge.2)then
            print *,perfnames(itme_reorder),xcpus_inv*(tme2-tme1)
         endif
         timing(itme_reorder)=timing(itme_reorder)+tme2-tme1
         ncalls(itme_reorder)=ncalls(itme_reorder)+1
         timing(itme_gravity)=timing(itme_gravity)+tme2-tme1
!         ncalls(itme_gravity) increment only in gravity subroutine
      endif

      call mtree(isoft,igrape,npart,npart_sph,sepmax,bsepmax,eps,
     &           boxlims,posmas,h,numclumps,lstclumps)

      first=.false.

      return
      end subroutine orderparticles
!----------------------------------------------------------------------
      subroutine restoreorder
!*********************************************************************
!                                                                    *
!  Restore original ordering for saved variables.                    *
!  See also subr. orderparticles, for the way the order              *
!  gets tangled in the first place.                                  *
!                                                                    *
!  You must call mtree after calling this routine if you             *
!  want to use the tree again, since we stomp on some of             *
!  the tree arrays here.                                             *
!                                                                    *
!  Be aware that this routine does not account for particles         *
!  that have been created or removed. That means that it             *
!  could write over the top of some real, newly created particles    *
!  or leave holes where old, previously removed particles were       *
!  found originally.                                                 *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer i,j

!$omp parallel default(none) 
!$omp& shared(npart,npart_sph,isort,temparr,itemparr)
!$omp& shared(ipindx,itstepbin,material)
!$omp& shared(posmas,vel,poten,h,alpha,u,rho)
!$omp& shared(velhaf,rhohalf,divvhalf)
!$omp& private(j,i) 

!$omp do
      do j=1,npart
         isort(ipindx(j))=j
      enddo
!
!======first reordering conversion=======
!
!$omp do
      do j=1,npart
         temparr(ndim3,j)=poten(j)
         temparr(ndim2,j)=h    (j)
         temparr(ndim1,j)=posmas(ndim1,j)
         temparr(1    ,j)=posmas(1    ,j)
         temparr(2    ,j)=posmas(2    ,j)
#if NDIM==3
         temparr(3    ,j)=posmas(3    ,j)
#endif
      enddo
!$omp do
      do j=1,npart
         i=isort(j)
         poten       (j)=temparr(ndim3,i)
         h           (j)=temparr(ndim2,i)
         posmas(ndim1,j)=temparr(ndim1,i)
         posmas(1    ,j)=temparr(1    ,i)
         posmas(2    ,j)=temparr(2    ,i)
#if NDIM==3
         posmas(3    ,j)=temparr(3    ,i)
#endif
      enddo
!
!======next reordering conversion=======
!
!$omp do
      do j=1,npart
         temparr(1    ,j)=vel  (1,j)
         temparr(2    ,j)=vel  (2,j)
#if NDIM==3
         temparr(3    ,j)=vel  (3,j)
#endif
      enddo

!$omp do
      do j=1,npart
         i=isort(j)
         vel  (1,j)=temparr(1    ,i)
         vel  (2,j)=temparr(2    ,i)
#if NDIM==3
         vel  (3,j)=temparr(3    ,i)
#endif
      enddo
!
!======next reordering conversion=======
!
!$omp do
      do j=1,npart_sph
         itemparr(1    ,j)=material(j)
         temparr (1    ,j)=alpha   (j)
         temparr (ndim1,j)=u       (j)
         temparr (ndim3,j)=rho     (j)
      enddo
!$omp do
      do j=1,npart_sph
         i=isort(j)
         material(j)=itemparr(1    ,i)
         alpha   (j)=temparr (1    ,i)
         u       (j)=temparr (ndim1,i)
         rho     (j)=temparr (ndim3,i)
      enddo
!
!======next reordering conversion=======
!
!$omp do
      do j=1,npart
         itemparr(1,j)=ipindx(j)
         itemparr(2,j)=itstepbin(j)
      enddo
!$omp do
      do j=1,npart
         i=isort(j)
         ipindx   (j)=itemparr(1,i)
         itstepbin(j)=itemparr(2,i)
      enddo

!$omp end parallel

      return
      end subroutine restoreorder
!----------------------------------------------------------------------
      subroutine findmult
!****************************************************************
!                                                               *
!  This routine finds particles which have `paired' in the      *
!  sense that their positions and velocities are identical.     *
!                                                               *
!  Subroutine by A. Nelson.                                     *
!                                                               *
!****************************************************************
      integer listpt_a(maxclumpsize), ntot_a
      integer listpt_i(maxclumpsize), ntot_i
      double precision distmas(ndim3,maxbite,maxclumpsize)
      integer nneigh(maxclumpsize),listnay(maxbite,maxclumpsize)

      integer jclump,iclump,j,i,k,knay
      integer nstart,moreneig
      
      logical iammult
      integer multip,multiplicity(maxclumpsize)
      integer kernv,neighdist(kern_incr)
      double precision vkern
      double precision dvx,dvy,dvz

      double precision avevel(ndim,maxclumpsize)
      double precision rmsvel(ndim,maxclumpsize)

      integer imac0,ife,irev
      logical:: first=.true.
      double precision ta
      integer na

      double precision tme1,tme2
      character(len=20), parameter:: where =('findmult')
      if(idbg.ge.2)print *,where 
      if(itme.ge.1)call cpu_time(tme1)

!
! update the tree information
!
      imac0=imac ; ta=treeacc
      if(first .and. imac.eq.id_gadget)then
         first=.false.
         imac0=-id_angle              
         ta=HALF
      endif
      irev=1
      ife=1

      nactive=npart
      nactive_s=npart_sph

      do j=1,nactive
         iactive(j)=j
      enddo
      do j=1,nactive_s
         iactive_s(j)=j
      enddo

      call rtree(igrav,imac0,isoft,igrape,ife,irev,npart,
     &           nactive,iactive,nactive_s,iactive_s,eps,ta,
     &           boxlims,posmas,h,numclumps,lstclumps,
     &           hclmp_is_exact)
!
! For some softening options (eg. fixed spline), we need to
! reset the smoothing lengths in the tree, so that the SPH
! particles get correct smoothing.  
!
      if(isoft.ne.id_varkern)then
         call resethnode(irev,npart,h,numclumps,lstclumps)
      endif

      do j=1,kern_incr
         vstep(j)=dble(j)*dvkern
      enddo

      neigdistrib=0
      nmulttot=0
 
!$omp parallel default(none)
!$omp&  firstprivate(ibctype)
!$omp&  shared(posmas,vel,h)
!$omp&  shared(numclumps,lstclumps,neigdistrib,nmulttot)
!$omp&  private(jclump,iclump,ntot_a,listpt_a,ntot_i,listpt_i,
!$omp&      iammult,multip,multiplicity,
!$omp&      i,j,k,knay,neighdist,moreneig,
!$omp&      nstart,nneigh,listnay,distmas,vkern,kernv,
!$omp&      dvx,dvy,dvz,rmsvel,avevel)

      call initgetneigh

      do j=1,kern_incr
         neighdist(j)=0
      enddo

!$omp do
      do jclump=1,numclumps

         iclump=lstclumps(jclump)
         call getparticles(iclump,maxclumpsize,
     &                     ntot_a,ntot_i,listpt_a,listpt_i)

         do j=1,ntot_i
            i=listpt_i(j)
            ntot_a=ntot_a+1
            listpt_a(ntot_a)=i
         enddo
         ntot_i=0

         do j=1,ntot_a
            i=listpt_a(j)
            multiplicity(j)=1
            avevel(:,j)=zero
            rmsvel(:,j)=zero
         enddo

         moreneig=1
         nstart=0
         do while(moreneig.ne.0)
            call getallneigh(ibctype,iclump,ntot_a,ntot_i,
     &                listpt_a,listpt_i,nstart,
     &                nneigh,listnay,distmas,moreneig)

            do j=1,ntot_a
               i=listpt_a(j)

               do k=1,nneigh(j)
                  knay=listnay(k,j)
                  iammult=.false.

!                  dvx=vel(1,i)-vel(1,knay)
!                  dvy=vel(2,i)-vel(2,knay)

                  vkern=sqrt(distmas(ir2,k,j)/distmas(ih2,k,j))
                  kernv=max(1,min(kern_incr,int(vkern/dvkern)))

                  neighdist(kernv)=neighdist(kernv) + 1

                  if(vkern.lt.1d-3)iammult=.true.

!                  if(distmas(ir2,k,j)/distmas(ih2,k,j).lt.1d-5 .and.
!     &               abs(dvx).lt.small .and.
!     &               abs(dvy).lt.small)iammult=.true.
!
!#if NDIM==3
!                  dvz=vel(3,i)-vel(3,knay)
!                  if(abs(dvz             ).lt.small .and.
!                     iammult)iammult=.true.
!#endif

                  if(iammult)then
                     multiplicity(j)=multiplicity(j)+1
                     !store neighbor identities to clean later
                  endif
               enddo
            enddo
         enddo

!
! save quantities for later.
!
!$omp critical
         do j=1,ntot_a
            i=listpt_a(j)
            multip=min(maxmult,multiplicity(j))
            nmulttot(multip)=nmulttot(multip)+1
         enddo
! 
! fix multiples by bumping them in opposite directions.
! Implement someday if its important.
!
!         do j=1,ntot_a
!            if(multiplicity(j).gt.1)then
!
!
!            endif
!         enddo

!$omp end critical


      enddo        !jclump loop
!
! Store total distribution of all neighbors
!
!$omp critical
      do j=1,kern_incr
         neigdistrib(j)=neigdistrib(j) + neighdist(j)
      enddo
!$omp end critical

!$omp end parallel

! 
! fix multiple counting
!
      do j=1,maxmult
         nmulttot(j)=nmulttot(j)/j
      enddo
      do j=1,kern_incr
         neigdistrib(j)=neigdistrib(j)/2
      enddo

      if(idbg.ge.1)then
         iammult=.false.
         do j=2,maxmult
            if(nmulttot(j).gt.0)iammult=.true.
         enddo
         if(iammult)then
            print *,'Maximum neighbor separation set to: ',vkernmax
            print *,' '
            print *,'found the following '
            print *,'multiple particle occurences'
            print *,' '
            print *,'Multiplicity | Number of Occurences'
            do j=2,maxmult
               write(*,40)j,nmulttot(j)
40             format(i8,'     |   ',i8)
            enddo
         endif
      endif
      print *,'       WARNING WARNING WARNING '
      print *,'findmult only detects multiple particles '
      print *,'      now, it does not repair them!!!'


      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_findmult)=timing(itme_findmult) + tme2-tme1
         ncalls(itme_findmult)=ncalls(itme_findmult) + 1
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine findmult
!----------------------------------------------------------------------
      end module vineutils
